Package: distionary
Title: Create and Evaluate Probability Distributions
Version: 0.1.1
Authors@R: c(
    person("Vincenzo", "Coia", , "vincenzo.coia@gmail.com", 
           role = c("aut", "cre", "cph")),
    person("Amogh", "Joshi", role = "ctb"),
    person("Shuyi", "Tan", role = "ctb"),
    person("Zhipeng", "Zhu", role = "ctb"),
    person("olivroy", role = "ctb", comment = "GitHub contributor"))
Description: Create and evaluate probability distribution objects from a
    variety of families or define custom distributions. Automatically compute
    distributional properties, even when they have not been specified.
    This package supports statistical modeling and simulations, and forms
    the core of the probaverse suite of R packages.
License: MIT + file LICENSE
Suggests: covr, knitr, rmarkdown, testthat (>= 3.0.0), tibble
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: checkmate, rlang, stats, vctrs
VignetteBuilder: knitr
URL: https://distionary.probaverse.com/,
        https://github.com/probaverse/distionary
BugReports: https://github.com/probaverse/distionary/issues
Depends: R (>= 2.10)
NeedsCompilation: no
Packaged: 2026-04-27 07:30:28 UTC; VCoia
Author: Vincenzo Coia [aut, cre, cph],
  Amogh Joshi [ctb],
  Shuyi Tan [ctb],
  Zhipeng Zhu [ctb],
  olivroy [ctb] (GitHub contributor)
Maintainer: Vincenzo Coia <vincenzo.coia@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-27 09:00:02 UTC
Built: R 4.5.2; ; 2026-05-01 17:31:29 UTC; unix

Package: distionary
Title: Create and Evaluate Probability Distributions
Version: 0.1.1
Authors@R: c(
    person("Vincenzo", "Coia", , "vincenzo.coia@gmail.com", 
           role = c("aut", "cre", "cph")),
    person("Amogh", "Joshi", role = "ctb"),
    person("Shuyi", "Tan", role = "ctb"),
    person("Zhipeng", "Zhu", role = "ctb"),
    person("olivroy", role = "ctb", comment = "GitHub contributor"))
Description: Create and evaluate probability distribution objects from a
    variety of families or define custom distributions. Automatically compute
    distributional properties, even when they have not been specified.
    This package supports statistical modeling and simulations, and forms
    the core of the probaverse suite of R packages.
License: MIT + file LICENSE
Suggests: covr, knitr, rmarkdown, testthat (>= 3.0.0), tibble
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: checkmate, rlang, stats, vctrs
VignetteBuilder: knitr
URL: https://distionary.probaverse.com/,
        https://github.com/probaverse/distionary
BugReports: https://github.com/probaverse/distionary/issues
Depends: R (>= 2.10)
NeedsCompilation: no
Packaged: 2026-04-27 07:30:28 UTC; VCoia
Author: Vincenzo Coia [aut, cre, cph],
  Amogh Joshi [ctb],
  Shuyi Tan [ctb],
  Zhipeng Zhu [ctb],
  olivroy [ctb] (GitHub contributor)
Maintainer: Vincenzo Coia <vincenzo.coia@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-27 09:00:02 UTC
Built: R 4.6.0; ; 2026-04-27 09:52:18 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(mean,dst)
S3method(median,dst)
S3method(plot,dst)
S3method(print,dst)
S3method(range,dst)
export("parameters<-")
export(dgev)
export(dgp)
export(distribution)
export(dst_bern)
export(dst_beta)
export(dst_binom)
export(dst_cauchy)
export(dst_chisq)
export(dst_degenerate)
export(dst_empirical)
export(dst_exp)
export(dst_f)
export(dst_finite)
export(dst_gamma)
export(dst_geom)
export(dst_gev)
export(dst_gp)
export(dst_hyper)
export(dst_lnorm)
export(dst_lp3)
export(dst_nbinom)
export(dst_norm)
export(dst_null)
export(dst_pearson3)
export(dst_pois)
export(dst_t)
export(dst_unif)
export(dst_weibull)
export(enframe_cdf)
export(enframe_chf)
export(enframe_density)
export(enframe_hazard)
export(enframe_odds)
export(enframe_pmf)
export(enframe_quantile)
export(enframe_return)
export(enframe_survival)
export(eval_cdf)
export(eval_chf)
export(eval_density)
export(eval_hazard)
export(eval_odds)
export(eval_pmf)
export(eval_property)
export(eval_quantile)
export(eval_return)
export(eval_survival)
export(is.distribution)
export(is_distribution)
export(kurtosis)
export(kurtosis_exc)
export(parameters)
export(pgev)
export(pgp)
export(pretty_name)
export(prob_left)
export(prob_right)
export(qgev)
export(qgp)
export(realise)
export(realize)
export(skewness)
export(stdev)
export(variance)
export(vtype)
importFrom(stats,median)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# distionary <img src='man/figures/logo.png' align="right" height="139" />

<!-- badges: start -->

[![License:
MIT](https://img.shields.io/badge/license-MIT-blue.svg)](https://cran.r-project.org/web/licenses/MIT)
[![R-CMD-check](https://github.com/probaverse/distionary/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/probaverse/distionary/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/probaverse/distionary/graph/badge.svg)](https://app.codecov.io/gh/probaverse/distionary)
[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![Status at rOpenSci Software Peer
Review](https://badges.ropensci.org/688_status.svg)](https://github.com/ropensci/software-review/issues/688)
<!-- badges: end -->

With `distionary`, you can:

1.  Specify a probability distribution
    ([built-in](https://distionary.probaverse.com/articles/specify-built-in.html)
    or [your
    own](https://distionary.probaverse.com/articles/specify-user-defined.html)),
    and
2.  [Evaluate](https://distionary.probaverse.com/articles/evaluate.html)
    the probability distribution.

The main purpose of `distionary` is to implement a distribution object,
and to make distribution calculations available even if they are not
specified in the distribution’s definition. `distionary` provides the
building blocks of the wider [`probaverse`
ecosystem](https://probaverse.com) for making representative statistical
models.

The name “distionary” is a portmanteau of “distribution” and
“dictionary”. While a dictionary lists and defines words, `distionary`
defines distributions and makes a list of common distribution families
available. The built-in distributions act as building blocks for the
wider probaverse.

## Statement of Need

When building statistical models, distributions should accurately
reflect your data, but out-of-the-box options like the Normal or Poisson
distributions often fall short. Achieving realistic probability
distributions demands a versatile workbench where distributions can be
manipulated, and data can inform their features. This is the goal of the
`probaverse` ecosystem, with `distionary` providing the foundational
building blocks.

`distionary` provides the fundamental `probaverse` infrastructure for
defining probability distribution objects. It allows for the evaluation
of distribution properties, even if they aren’t explicitly specified,
offering standalone utility for users needing to define a distribution
in various forms and evaluate it comprehensively.

## Target Audience

Lots of people work with probability distributions. Lots of people
*don’t* work with probability distributions but should, because they
don’t see the value or because distributions are too clumsy to work with
under existing infrastructure. And, there are lots of people learning
about probability distributions that would have an easier time if they
get to “feel” distributions and their multifaceted nature. `distionary`
is for all of these people.

`distionary` – and the `probaverse` more widely – is designed for data
scientists, statisticians, and researchers who require the flexibility
to develop custom statistical models. It caters to those in finance,
insurance, environmental science, and engineering, where nuanced
distribution modeling is crucial. Whether building complex stochastic
models or performing detailed risk assessments, `distionary` equips
users with the tools needed to explore and manipulate probability
distributions effectively.

`distionary` makes reference to common terms regarding probability
distributions. If you’re uneasy with these terms and concepts, most
intro books in probability will be a good resource to learn from. As
`distionary` develops, more documentation will be made available so that
it’s more self-contained.

## Installation

To install `distionary`, run the following code in R:

``` r
install.packages("distionary")
```

## Example: Built-in Distributions

``` r
library(distionary)
```

**Specify** a distribution like a Poisson distribution and a Generalised
Extreme Value (GEV) distribution using the `dst_*()` family of
functions.

``` r
# Create a Poisson distribution
poisson <- dst_pois(1.5)
# Inspect
poisson
#> Poisson distribution (discrete) 
#> --Parameters--
#> lambda 
#>    1.5
```

``` r
# Create a GEV distribution
gev <- dst_gev(-1, 1, 0.2)
# Inspect
gev
#> Generalised Extreme Value distribution (continuous) 
#> --Parameters--
#> location    scale    shape 
#>     -1.0      1.0      0.2
```

Here is what the distributions look like, via their probability mass
(PMF) and density functions.

``` r
plot(poisson)
```

<img src="man/figures/README-unnamed-chunk-5-1.png" alt="" width="100%" />

``` r
plot(gev)
```

<img src="man/figures/README-unnamed-chunk-5-2.png" alt="" width="100%" />

**Evaluate** various *distributional properties* such as mean, skewness,
and range of valid values.

``` r
mean(gev)
#> [1] -0.1788514
skewness(poisson)
#> [1] 0.8164966
range(gev)
#> [1]  -6 Inf
```

Properties that completely define the distribution are called
*distributional representations*, and can be accessed by the `eval_*()`
functions. such as the PMF or quantiles. The `eval_*()` functions simply
evaluate the representation, whereas the `enframe_*()` functions place
the output alongside the input in a data frame or tibble.

``` r
eval_pmf(poisson, at = 0:4)
#> [1] 0.22313016 0.33469524 0.25102143 0.12551072 0.04706652
enframe_quantile(gev, at = c(0.2, 0.5, 0.9))
#> # A tibble: 3 × 2
#>    .arg quantile
#>   <dbl>    <dbl>
#> 1   0.2   -1.45 
#> 2   0.5   -0.620
#> 3   0.9    1.84
```

## Example: Custom Distributions

You can create a custom distribution using `distribution()`. The
innovative aspect of `distionary` is its ability to automatically
compute properties from the specified representations. By providing just
one or two representations (such as CDF and density), `distionary` can
derive other properties as needed.

``` r
# Make a distribution by specifying only density and CDF
linear <- distribution(
  density = function(x) {
    d <- 2 * (1 - x)
    d[x < 0 | x > 1] <- 0
    d
  },
  cdf = function(x) {
    p <- 2 * x * (1 - x / 2)
    p[x < 0] <- 0
    p[x > 1] <- 1
    p
  },
  .vtype = "continuous",
  .name = "My Linear"
)
# Inspect
linear
#> My Linear distribution (continuous) 
#> --Parameters--
#> NULL
```

Here is what it looks like (density function).

``` r
plot(linear)
```

<img src="man/figures/README-unnamed-chunk-9-1.png" alt="" width="100%" />

Even though only the density and CDF were specified, other properties
can be evaluated, like its mean and quantiles:

``` r
mean(linear)
#> [1] 0.3333333
enframe_quantile(linear, at = c(0.2, 0.5, 0.9))
#> # A tibble: 3 × 2
#>    .arg quantile
#>   <dbl>    <dbl>
#> 1   0.2    0.106
#> 2   0.5    0.293
#> 3   0.9    0.684
```

## `distionary` in the Context of Other Packages

The R ecosystem offers several packages for working with probability
distributions, each with unique strengths:

- **`stats` Package**: Provides fundamental functions for standard
  distributions but lacks a unified object-oriented approach for complex
  manipulations.

- **`distr` Package**: Introduces an object-oriented framework for
  distribution objects using S4, offering flexible manipulation, though
  it can be complex to extend and use.

- **`distributions3` Package**: Utilizes S3 classes for a
  straightforward interface focused on simplicity, suitable for basic
  tasks but may not meet advanced application needs.

- **`distributional` Package**: Extends `distributions3` to support
  vectorized operations, aiding statistical modelling but lacking
  extensive tools for custom distribution creation.

In this landscape, `distionary` addresses the need for a cohesive and
flexible API that can seamlessly integrate the strengths of these
packages. It provides a unified framework for defining, manipulating,
and evaluating probability distributions. Because `distionary` only
needs some distribution properties to be specified, it offers a level of
flexibility central to the `probaverse` ecosystem.

## Acknowledgements

The creation of `distionary` would not have been possible without the
support of BGC Engineering Inc., the R Consortium, the Politecnico di
Milano, the European Space Agency, The University of British Columbia,
and the Natural Science and Engineering Research Council of Canada
(NSERC). The authors would also like to thank the reviewers from
ROpenSci for their insightful feedback, which greatly contributed to
enhancing the quality of this R package.

## Citation

To cite package `distionary` in publications use:

Coia V (2025). *distionary: Create and Evaluate Probability
Distributions*. R package version 0.1.0,
<https://github.com/probaverse/distionary>,
<https://distionary.probaverse.com/>.

## Code of Conduct

Please note that the distionary project is released with a [Code of
Conduct](https://distionary.probaverse.com/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

#' Script to generate the network diagram connecting the various distribution
#' properties.
library(DiagrammeR)
library(DiagrammeRsvg)

# Define the graph using Graphviz DOT language
graph <- grViz("
digraph G {
  graph [layout = dot, rankdir = LR]

  # Define nodes
  quantile [label = 'quantile']
  cdf [label = 'cdf']
  return [label = 'return']
  range [label = 'range']
  realise [label = 'realise']
  median [label = 'median']
  survival [label = 'survival']
  chf [label = 'chf']
  hazard [label = 'hazard']
  density [label = 'density']
  moments [label = 'moments']
  stdev [label = 'stdev']
  kurtosis_exc [label = 'kurtosis_exc']
  odds [label = 'odds']
  pmf [label = 'pmf']
  junction [label = '', shape = point, width = 0, height = 0, style = invis]

  # Define edges
  cdf -> quantile
  quantile -> return
  quantile -> range
  quantile -> realise
  quantile -> median
  cdf -> survival
  survival -> chf
  survival -> junction [arrowhead = none]
  density -> junction [arrowhead = none]
  junction -> hazard
  density -> moments
  moments -> stdev
  stdev -> moments
  moments -> kurtosis_exc
  kurtosis_exc -> moments
  pmf -> odds

  # Arrange nodes
  {rank = same; cdf; density}
}
")

svg <- export_svg(graph)
writeLines(svg, here::here("vignettes/network_diagram.svg"))

# Export `graph` to png
rsvg::rsvg_png(charToRaw(svg), file = here::here("vignettes/network_diagram.png"), width = 1600)

#' Weibull Distribution
#'
#' Makes a Weibull distribution.
#'
#' @param scale Scale parameter; single positive numeric.
#' @param shape Shape parameter; single positive numeric.
#' @returns A Weibull distribution.
#' @examples
#' dst_weibull(1, 1)
#' @export
dst_weibull <- function(shape, scale) {
  checkmate::assert_numeric(shape, 0, len = 1)
  checkmate::assert_numeric(scale, 0, len = 1)
  if (is.na(shape) || is.na(scale)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(shape = shape, scale = scale),
    density = function(x) {
      stats::dweibull(x, shape = shape, scale = scale)
    },
    cdf = function(x) {
      stats::pweibull(x, shape = shape, scale = scale)
    },
    quantile = function(p) {
      stats::qweibull(p, shape = shape, scale = scale)
    },
    realise = function(n) {
      stats::rweibull(n, shape = shape, scale = scale)
    },
    survival = function(x) {
      stats::pweibull(x, shape = shape, scale = scale, lower.tail = FALSE)
    },
    mean = scale * gamma(1 + 1 / shape),
    variance = scale^2 * (gamma(1 + 2 / shape) - gamma(1 + 1 / shape)^2),
    skewness = {
      g1 <- gamma(1 + 1 / shape)
      g2 <- gamma(1 + 2 / shape)
      g3 <- gamma(1 + 3 / shape)
      (2 * g1^3 - 3 * g1 * g2 + g3) / (g2 - g1^2)^(3 / 2)
    },
    kurtosis_exc = {
      g1 <- gamma(1 + 1 / shape)
      g2 <- gamma(1 + 2 / shape)
      g3 <- gamma(1 + 3 / shape)
      g4 <- gamma(1 + 4 / shape)
      (-6 * g1^4 + 12 * g1^2 * g2 - 3 * g2^2 - 4 * g1 * g3 + g4) /
        (g2 - g1^2)^2
    },
    range = c(0, Inf),
    .name = "Weibull",
    .vtype = "continuous"
  )
}

test_that("Variable type specified correctly.", {
  expect_equal(vtype(dst_bern(0.5)), "discrete")
  expect_equal(vtype(dst_beta(1, 1)), "continuous")
  expect_equal(vtype(dst_binom(size = 1, prob = 0.4)), "discrete")
  expect_equal(vtype(dst_cauchy(1, 1)), "continuous")
  expect_equal(vtype(dst_chisq(1)), "continuous")
  expect_equal(vtype(dst_degenerate(1)), "discrete")
  expect_equal(vtype(dst_exp(1)), "continuous")
  expect_equal(vtype(dst_f(1, 1)), "continuous")
  expect_equal(vtype(dst_gamma(1, 1)), "continuous")
  expect_equal(vtype(dst_geom(0.5)), "discrete")
  expect_equal(vtype(dst_gev(1, 1, 1)), "continuous")
  expect_equal(vtype(dst_gp(1, 1)), "continuous")
  expect_equal(vtype(dst_hyper(1, 1, 1)), "discrete")
  expect_equal(vtype(dst_lnorm(1, 1)), "continuous")
  expect_equal(vtype(dst_lp3(1, 1, 1)), "continuous")
  expect_equal(vtype(dst_nbinom(size = 1, prob = 0.4)), "discrete")
  expect_equal(vtype(dst_norm(1, 1)), "continuous")
  expect_true(is.na(vtype(dst_null())))
  expect_equal(vtype(dst_pearson3(1, 1, 1)), "continuous")
  expect_equal(vtype(dst_pois(1)), "discrete")
  expect_equal(vtype(dst_t(1)), "continuous")
  expect_equal(vtype(dst_unif(0, 1)), "continuous")
  expect_equal(vtype(dst_weibull(1, 1)), "continuous")
})

test_that("Variable type matches manual distribution input", {
  expect_identical(vtype(distribution(.vtype = "foofy")), "foofy")
  expect_identical(vtype(distribution(.vtype = "FoOFy")), "foofy")
  expect_identical(vtype(distribution()), "unknown")
  expect_identical(vtype(distribution()), vtype(distribution(.vtype = NULL)))
  expect_identical(vtype(distribution(.vtype = NA_character_)), NA_character_)
})

# Make sure that defined distributions evaluate NA inputs properly.
# Make sure they evaluate edge cases properly.
test_that("`eval_()` functions handle edge cases properly", {
  for (item in test_distributions) {
    d <- rlang::exec(item$distribution, !!!item$valid[[1]])
    r <- range(d)
    eps <- .Machine$double.eps
    p <- c(0.4, NA_real_)
    ## Quantile. Also use x for downstream tests.
    x <- eval_quantile(d, at = p)
    expect_true(is.numeric(x[1]))
    expect_true(is.na(x[2]))
    expect_length(eval_quantile(d, at = numeric()), 0)
    expect_equal(eval_quantile(d, at = 0:1), r)
    ## CDF
    y <- eval_cdf(d, at = x)
    expect_true(is.numeric(y[1]))
    expect_true(is.na(y[2]))
    expect_length(eval_cdf(d, at = numeric()), 0)
    if (attr(d, "name") == "Hypergeometric") {
      # The hypergeometric distribution from the stats package
      # needs to move about 1e-7 units to the left of the minimum value
      # in order to register a drop of the CDF to 0. This will be
      # deemed acceptable since it's acceptable in the stats package,
      # although may be up for consideration in a future version of
      # distionary.
      expect_equal(
        eval_cdf(d, at = c(-Inf, r[1] - 1e-6, r[2], Inf)),
        c(0, 0, 1, 1)
      )
    } else {
      expect_equal(
        eval_cdf(d, at = c(-Inf, r[1] - eps, r[2], Inf)),
        c(0, 0, 1, 1)
      )
    }
    ## Survival
    y <- eval_survival(d, at = x)
    expect_true(is.numeric(y[1]))
    expect_true(is.na(y[2]))
    expect_length(eval_survival(d, at = numeric()), 0)
    if (attr(d, "name") == "Hypergeometric") {
      expect_equal(
        eval_survival(d, at = c(-Inf, r[1] - 1e-6, r[2], Inf)),
        c(1, 1, 0, 0)
      )
    } else {
      expect_equal(
        eval_survival(d, at = c(-Inf, r[1] - eps, r[2], Inf)),
        c(1, 1, 0, 0)
      )
    }
    ## Mass
    y <- eval_pmf(d, at = x)
    expect_true(is.numeric(y[1]))
    expect_true(is.na(y[2]))
    expect_length(eval_pmf(d, at = numeric()), 0)
    expect_equal(
      eval_pmf(d, at = c(-Inf, r[1] - 1, r[2] + 1, Inf)),
      rep(0, 4)
    )
    ## Density
    if (vtype(d) == "continuous") {
      y <- eval_density(d, at = x)
      expect_true(is.numeric(y[1]))
      expect_true(is.na(y[2]))
      expect_length(eval_density(d, at = numeric()), 0)
      dens <- eval_density(d, at = c(-Inf, Inf))
      expect_equal(dens, c(0, 0))
      # expect_true(all(is.infinite(dens) | dens == rep(0, 4)))
    }
    ## Realise
    expect_length(realise(d, n = 0), 0)
  }
})

test_that("Random number generation works", {
  # Strategy:
  # Generate more and more data until the p-value is > 0.05.
  # If the distribution is not the correct specification, the p-value
  # will get closer and closer to 0. If the distribution is correct,
  # then p ~ Unif(0,1), and if p < 0.05 (unlucky), extending the
  # sample will eliminate the luck factor.
  # If the number generator is *not* correct, then the large sample
  # size drawn (1000) is likely to trigger a small p-value; to mitigate
  # the risk that a large p-value is obtained anyway, the process is
  # repeated 4 times with different seeds.
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (vtype(d) == "continuous") {
        cdf <- representation_as_function(d, "cdf")
        for (sd in 1:4) {
          set.seed(sd)
          x <- realise(d, n = 1000)
          pval <- ks.test(x, cdf)$p.value
          maxiter <- 10L
          i <- 1
          while (pval < 0.05 && i < maxiter) {
            x <- append(x, realise(d, n = 1000))
            pval <- ks.test(x, cdf)$p.value
            i <- i + 1
          }
          expect_gt(pval, 0.05)
        }
      } else if (pretty_name(d) != "Degenerate") {
        for (sd in 1:4) {
          # For tapering distributions like the Poisson,
          # generate data only within a finite range encompassing most of the
          # possibilities, and compare to the distribution trimmed to those
          # values. This is because generating more and more data will
          # result in more and more new (large) values with low frequency,
          # throwing off the chi-squared test.
          set.seed(sd)
          x <- realise(d, n = 1000)
          tbl <- table(x)
          bnds <- range(x)
          if (pretty_name(d) == "Finite") {
            rng <- parameters(d)$outcomes
            rng <- rng[rng >= bnds[1] & rng <= bnds[2]]
          } else {
            rng <- bnds[1]:bnds[2]
          }
          freq <- rep(0, length(rng))
          names(freq) <- rng
          freq[names(tbl)] <- unname(unclass(tbl))
          p_orig <- eval_pmf(d, at = rng)
          p <- p_orig / sum(p_orig)
          pval <- suppressWarnings(chisq.test(freq, p = p)$p.value)
          maxiter <- 10L
          i <- 1
          while (pval < 0.05 && i < maxiter) {
            x <- append(x, realise(d, n = 10000))
            x <- x[x >= bnds[1] & x <= bnds[2]]
            tbl <- table(x)
            freq <- rep(0, length(rng))
            names(freq) <- rng
            freq[names(tbl)] <- unname(unclass(tbl))
            pval <- suppressWarnings(chisq.test(freq, p = p)$p.value)
            i <- i + 1
          }
          if (pval < 0.05) {
            print(pval)
            print(sd)
            print(d)
          }
          expect_gt(pval, 0.05)
        }
      } else if (pretty_name(d) == "Degenerate") {
        for (sd in 1:4) {
          set.seed(sd)
          x <- realise(d, n = 1000)
          expect_identical(unique(x), parameters(d)[[1]])
        }
      }
    }
  }
})

#' @section Overview:
#' The \pkg{distionary} package provides a comprehensive framework for working
#' with probability distributions in R. With \pkg{distionary}, you can:
#' \enumerate{
#'   \item Specify probability distributions from common families or create
#'         custom distributions.
#'   \item Evaluate distributional properties and representations.
#'   \item Access distributional calculations even when they're not directly
#'         specified.
#' }
#'
#' The main purpose of \pkg{distionary} is to implement a distribution object
#' that powers the wider \href{https://probaverse.com}{probaverse ecosystem} for
#' making probability distributions that are representative of your data.
#'
#' @section Creating Distributions:
#' Use the \code{dst_*()} family of functions to create distributions from
#' common families:
#' \itemize{
#'   \item \code{\link{dst_norm}()}, \code{\link{dst_exp}()},
#'         \code{\link{dst_unif}()}, etc. are some continuous distributions.
#'   \item \code{\link{dst_pois}()}, \code{\link{dst_binom}()},
#'         \code{\link{dst_geom}()}, etc. are some discrete distributions.
#'   \item \code{\link{dst_empirical}()} is useful for creating a non-parametric
#'         distribution from data.
#' }
#' You can also make your own distribution using the
#' \code{\link{distribution}()} function, which allows you to specify
#' any combination of distributional representations and properties. For this
#' version of \pkg{distionary}, the CDF and density/PMF are required
#' in order to access all functionality.
#'
#' @section Evaluating Distributions:
#' A distribution's _representations_ are functions that fully describe the
#' distribution. They can be accessed with the \code{eval_*()} functions.
#' For example, \code{\link{eval_cdf}()} and \code{\link{eval_quantile}()}
#' invoke the distribution's cumulative distribution function (CDF)
#' and quantile function.
#'
#' Other properties of the distribution can be calculated by functions of the
#' property's name, such as \code{\link{mean}()} and \code{\link{range}()}.
#'
#' @section Random Samples:
#' Generate random samples from a distribution using
#' \code{\link{realise}()}.
#'
#' @section Getting Started:
#' New users should start with the package vignettes:
#' \itemize{
#'   \item \code{vignette("specify", package = "distionary")} -
#'         Learn how to specify distributions.
#'   \item \code{vignette("evaluate", package = "distionary")} -
#'         Learn how to evaluate distributions.
#' }
#'
#' @examples
#' # Create a Poisson distribution.
#' poisson <- dst_pois(lambda = 1.5)
#' poisson
#'
#' # Evaluate the probability mass function.
#' eval_pmf(poisson, at = 0:4)
#' plot(poisson)
#'
#' # Get distribution properties.
#' mean(poisson)
#' variance(poisson)
#'
#' # Create a continuous distribution (Normal).
#' normal <- dst_norm(mean = 0, sd = 1)
#'
#' # Evaluate quantiles.
#' eval_quantile(normal, at = c(0.025, 0.5, 0.975))
#'
#' # Create a custom distribution.
#' my_dist <- distribution(
#'   density = function(x) ifelse(x >= 0 & x <= 1, 2 * (1 - x), 0),
#'   cdf = function(x) ifelse(x >= 0 & x <= 1, 1 - (1 - x)^2, 0),
#'   .vtype = "continuous",
#'   .name = "Linear"
#' )
#' plot(my_dist)
#' plot(my_dist, "cdf")
#'
#' # Even without specifying all properties, they can still be computed.
#' mean(my_dist)
#'
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' Parameters of a Distribution
#'
#' Get or set the parameters of a distribution, if applicable. See
#' details.
#'
#' @param distribution Distribution.
#' @param value A list of named parameter values, or `NULL`.
#' @details If a distribution is made by specifying parameter values
#' (e.g., mean and variance for a Normal distribution; shape parameters
#' for a Beta distribution), it is useful to keep track of what
#' these parameters are. This is done by adding `parameters`
#' to the list of objects defining the distribution; for instance,
#' `distribution(parameters = c(shape1 = 1.4, shape2 = 3.4))`.
#' Note that no checks are made to ensure the parameters are valid.
#' It's important to note that, in this version of distionary,
#' manually changing the parameters after the distribution has been
#' created will not change the functionality of the distribution,
#' because the parameters are never referred to when making calculations.
#' @returns A list of the distribution parameters. More specifically,
#' returns the `"parameters"` entry of the list making up the
#' probability distribution.
#' @examples
#' a <- dst_beta(1, 2)
#' parameters(a)
#'
#' b <- distribution(mean = 5)
#' parameters(b)
#' parameters(b) <- list(t = 7)
#' parameters(b)
#' @rdname parameters
#' @export
parameters <- function(distribution) {
  checkmate::assert_class(distribution, "dst")
  attr(distribution, "parameters")
}

#' @export
#' @rdname parameters
`parameters<-` <- function(distribution, value) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_list(value, names = "named", null.ok = TRUE)
  attr(distribution, "parameters") <- value
  distribution
}

# Testing that each representation satisfies its definition.
#
# Check that each representation satisfies the properties of that
# representation. For example, a CDF should be non-decreasing and
# evaluate between 0 and 1, reaching 0 and 1 at the endpoints.
#
# Systematic testing uses the `test_distributions` object.
# To modify it, see `data-raw/test_distributions.R`.
# To update it, run `Rscript data-raw/test_distributions.R`.


test_that("Each representation satisfies its definition.", {
  for (i in seq_along(test_distributions)) {
    item <- test_distributions[[i]]
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      v <- vtype(d)
      p <- 0:50 / 50
      x <- eval_quantile(d, at = p)
      r <- range(d)
      ## CDF should increase from y=0 to y=1.
      cdf_vals <- eval_cdf(d, at = x)
      if (v == "continuous") {
        expect_equal(cdf_vals[1], 0)
      } else {
        expect_gt(cdf_vals[1], 0)
      }
      expect_lte(cdf_vals[49], 1)
      expect_true(all(diff(cdf_vals) >= 0))
      ## Survival
      if (!is.null(d$survival)) {
        surv <- eval_survival(d, at = x)
        if (v == "continuous") {
          expect_lte(surv[1], 1)
        }
        expect_gte(surv[49], 0)
        expect_true(all(diff(surv) <= 0))
      }
      ## Density
      if (v == "continuous") {
        a <- eval_quantile(d, at = c(0.005, 0.995))
        dens_vals <- eval_density(d, at = x[!is.infinite(x)])
        expect_true(all(dens_vals >= 0))
        dens_fun <- representation_as_function(d, "density")
        int <- integrate(dens_fun, lower = a[1], upper = a[2])
        expect_equal(int$value, 0.99, tolerance = 1e-5)
      }
      ## Mass
      pmf_fun <- d$pmf
      if (!is.null(pmf_fun)) {
        # Most of the distributions tested will have mass here:
        pmf_vals <- eval_pmf(d, at = -40:1000)
        # Test finite distributions specially:
        if (pretty_name(d) == "Finite") {
          outcomes <- parameters(d)$outcomes
          pmf_vals <- eval_pmf(d, at = outcomes)
          expect_equal(sum(pmf_vals), 1)
        }
        expect_true(all(pmf_vals >= 0))
        expect_gt(sum(pmf_vals), 0.9)
        expect_lt(sum(pmf_vals), 1 + 2 * .Machine$double.eps)
      }
      ## Quantile
      qf <- d$quantile
      if (!is.null(qf)) {
        expect_true(all(diff(x) >= 0))
        if (v == "continuous") {
          expect_equal(eval_cdf(d, at = x), p)
        }
      }
      ## Hazard
      if (v == "continuous") {
        expect_true(all(eval_hazard(d, at = x[-c(1, 51)]) >= 0))
      }
      ## CHF
      if (v == "continuous") {
        chf <- eval_chf(d, at = x[-51])
        expect_true(all(diff(chf) >= 0))
      }
      ## PMF
      if (v == "discrete") {
        if (pretty_name(d) == "Finite") {
          xx <- parameters(d)$outcomes
          ## From CDF
          pmf_evald <- eval_pmf(d, at = xx)
          pmf_derived <- prob_left(d, of = xx, inclusive = TRUE) -
            prob_left(d, of = xx, inclusive = FALSE)
          expect_equal(pmf_derived, pmf_evald)
          ## Sum to 1
          pmf_sum <- sum(pmf_evald)
          diff_from_one <- 1 - pmf_sum
        } else if (pretty_name(d) != "Degenerate") {
          if (is.infinite(r[2])) {
            xx <- 0:99
          } else {
            xx <- r[1]:r[2]
          }
          ## From CDF
          pmf_evald <- eval_pmf(d, at = xx)
          pmf_derived <- prob_left(d, of = xx, inclusive = TRUE) -
            prob_left(d, of = xx, inclusive = FALSE)
          expect_equal(pmf_derived, pmf_evald)
          ## Sum to 1
          pmf_sum <- sum(pmf_evald)
          diff_from_one <- 1 - pmf_sum
          while (diff_from_one > 1e-9) {
            xx <- 100 + xx
            pmf_evald <- eval_pmf(d, at = xx)
            pmf_sum <- pmf_sum + sum(pmf_evald)
            diff_from_one <- 1 - pmf_sum
          }
          expect_lt(diff_from_one, 1e-9)
        }
      }
    }
  }
})

test_that("Parameters are specified and retrieved correctly w/ parameters()", {
  param_retrieved <- list()
  # Remove the finite and empirical distributions, which are special.
  test_suite <- test_distributions
  test_suite[["empirical"]] <- NULL
  test_suite[["finite"]] <- NULL
  for (i in seq_along(test_suite)) {
    item <- test_suite[[i]]
    paramset <- item$valid[[1]]
    d <- rlang::exec(item$distribution, !!!paramset)
    param_retrieved[[i]] <- parameters(d)
    expect_identical(param_retrieved[[!!i]], !!paramset)
    parameters(d)[[1]] <- "hello"
    expect_identical(parameters(d)[[1]], "hello")
    parameters(d)[[1]] <- NaN
    expect_identical(parameters(d)[[1]], NaN)
    parameters(d)[[1]] <- data.frame(x = 1)
    expect_identical(parameters(d)[[1]], data.frame(x = 1))
  }
})

test_that("Empirical distribution parameters() are correct.", {
  item <- test_distributions[["empirical"]]
  dst_finite_paramnames <- names(parameters(dst_finite(1:2, c(0.5, 0.5))))
  for (paramset in item$valid) {
    d <- rlang::exec(item$distribution, !!!paramset)
    param_retrieved <- parameters(d)
    if (length(paramset) == 1) {
      paramset[[2]] <- rep(1, length(paramset[[1]]))
    }
    names(paramset) <- dst_finite_paramnames
    i_mask <- paramset[[2]] != 0
    paramset[[1]] <- paramset[[1]][i_mask]
    paramset[[2]] <- paramset[[2]][i_mask]
    paramset[[2]] <- paramset[[2]][order(paramset[[1]])]
    paramset[[1]] <- sort(paramset[[1]])
    paramset[[2]] <- paramset[[2]] / sum(paramset[[2]])
    expect_identical(param_retrieved, paramset)
  }
})

test_that("Finite distribution parameters() are correct.", {
  item <- test_distributions[["finite"]]
  for (paramset in item$valid) {
    d <- rlang::exec(item$distribution, !!!paramset)
    if (length(paramset[[1]]) == 1) {
      expect_identical(d, dst_degenerate(paramset[[1]]))
      next
    }
    param_retrieved <- parameters(d)
    i_mask <- paramset[[2]] != 0
    paramset[[1]] <- paramset[[1]][i_mask]
    paramset[[2]] <- paramset[[2]][i_mask]
    paramset[[2]] <- paramset[[2]][order(paramset[[1]])]
    paramset[[1]] <- sort(paramset[[1]])
    expect_identical(param_retrieved, paramset)
  }
})

test_that("Mean calculated thru network matches known vals", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      # print(d)
      supposed_mean <- mean(d)
      if (is.infinite(supposed_mean)) {
        supposed_mean <- NaN  # For direct comparison to integral output.
      }
      if (is_intrinsic(d, "mean")) {
        if (vtype(d) == "continuous") {
          if (
            pretty_name(d) == "Generalised Extreme Value" &&
            parameters(d)$shape == 0
          ) {
            # The density becomes NaN if x is too small. Manually check.
            integrand <- function(x) {
              x * eval_density(d, at = x)
            }
            expect_equal(
              stats::integrate(integrand, -1000, 0)$value +
                stats::integrate(integrand, 0, Inf)$value,
              supposed_mean
            )
          } else {
            expect_equal(
              suppressMessages(eval_mean_from_network(d)),
              supposed_mean
            )
          }
        } else if (
          pretty_name(d) %in%
          c("Hypergeometric", "Bernoulli", "Binomial", "Finite")
        ) {
          # Finite support.
          expect_error(eval_mean_from_network(d))
          r <- range(d)
          x <- seq(r[1], r[2], by = 1L)
          if (pretty_name(d) == "Finite") {
            # Finite distribution can have non-integer support.
            x <- parameters(d)$outcomes
          }
          p <- eval_pmf(d, at = x)
          expect_equal(sum(p * x), supposed_mean)
        } else if (
          pretty_name(d) %in%
          c("Negative Binomial", "Poisson", "Geometric")
        ) {
          # Infinite support.
          expect_error(eval_mean_from_network(d))
          to_add <- Inf
          i <- 0
          mean <- 0
          while (to_add > 1e-9) {
            x <- 1:100 + 100 * i
            to_add <- sum(eval_pmf(d, x) * x)
            mean <- mean + to_add
            i <- i + 1
          }
          expect_equal(mean, supposed_mean)
        } else if (pretty_name(d) == "Degenerate") {
          # Do nothing; checked in a separate script.
        } else {
          # Shouldn't be any distributions left.
          stop("At least one distribution family has not been accounted for.")
        }
      }
    }
  }
})

test_that("Network is invoked in priority: variance", {
  # First look for `stdev`, then invoke algorithm if not found.
  d <- distribution(
    density = function(x) {
      stats::dnorm(x, sd = 3)
    },
    range = c(-Inf, Inf),
    stdev = 10, # deliberately incorrect
    .vtype = "continuous"
  )
  expect_equal(variance(d), 100)
  expect_equal(eval_variance_from_network(d), 100)
  d$stdev <- NULL
  expect_equal(variance(d), 9)
  expect_equal(eval_variance_from_network(d), 9)
})

test_that("Variance algorithm matches known vals", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (is_intrinsic(d, "variance") || is_intrinsic(d, "stdev")) {
        supposed_var <- variance(d)
        if (is.infinite(supposed_var)) {
          supposed_var <- NaN  # To align with numerical integration output.
        }
        if (vtype(d) == "continuous") {
          if (
            pretty_name(d) == "Generalised Extreme Value" &&
            parameters(d)$shape == 0
          ) {
            # GEV with shape = 0 has NaN density for very negative values.
            mu <- mean(d)
            integrand <- function(x) {
              (x - mu)^2 * eval_density(d, at = x)
            }
            expect_equal(
              stats::integrate(integrand, -1000, 0)$value +
                stats::integrate(integrand, 0, Inf)$value,
              supposed_var
            )
          } else {
            expect_equal(suppressMessages(algorithm_variance(d)), supposed_var)
          }
        } else if (
          pretty_name(d) %in%
          c("Hypergeometric", "Bernoulli", "Binomial", "Finite")
        ) {
          # Finite support.
          expect_error(algorithm_variance(d))
          r <- range(d)
          x <- seq(r[1], r[2], by = 1L)
          if (pretty_name(d) == "Finite") {
            # Finite distribution can have non-integer support.
            x <- parameters(d)$outcomes
          }
          mu <- mean(d)
          x2 <- (x - mu)^2
          p <- eval_pmf(d, at = x)
          expect_equal(sum(p * x2), supposed_var)
        } else if (
          pretty_name(d) %in%
          c("Negative Binomial", "Poisson", "Geometric")
        ) {
          # Infinite support.
          expect_error(algorithm_variance(d))
          mu <- mean(d)
          to_add <- Inf
          i <- 0
          v <- 0
          while (to_add > 1e-9) {
            x <- 0:99 + 100 * i
            to_add <- sum(eval_pmf(d, x) * (x - mu)^2)
            v <- v + to_add
            i <- i + 1
          }
          expect_equal(v, supposed_var)
        } else if (pretty_name(d) == "Degenerate") {
          # Do nothing; checked in a separate script.
        } else {
          # Shouldn't be any distributions left.
          stop("At least one distribution family has not been accounted for.")
        }
      }
    }
  }
})

test_that("Network is invoked in priority: kurtosis", {
  # First look for `kurtosis_exc`, then invoke algorithm if not found.
  d <- distribution(
    density = stats::dnorm,
    range = c(-Inf, Inf),
    kurtosis_exc = 100 - 3, # deliberately incorrect
    .vtype = "continuous"
  )
  expect_equal(kurtosis(d), 100)
  expect_equal(eval_kurtosis_from_network(d), 100)
  d$kurtosis_exc <- NULL
  expect_equal(kurtosis(d), 3)
  expect_equal(eval_kurtosis_from_network(d), 3)
})

test_that("Original kurtosis is never specified in our distributions", {
  # ...so don't need to verify their entries.
  for (item in test_distributions) {
    d <- rlang::exec(item$distribution, !!!item$valid[[1]])
    expect_false(is_intrinsic(d, "kurtosis"))
  }
})

test_that("Kurtosis algorithm matches known vals", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (is_intrinsic(d, "kurtosis_exc") || is_intrinsic(d, "kurtosis")) {
        supposed_kurt <- kurtosis(d)
        if (is.infinite(supposed_kurt)) {
          supposed_kurt <- NaN  # To align with numerical integration output.
        }
        if (vtype(d) == "continuous") {
          expect_equal(suppressMessages(algorithm_kurtosis(d)), supposed_kurt)
        } else if (
          pretty_name(d) %in%
          c("Hypergeometric", "Bernoulli", "Binomial", "Finite")
        ) {
          # Finite support.
          expect_error(algorithm_kurtosis(d))
          r <- range(d)
          x <- seq(r[1], r[2], by = 1L)
          if (pretty_name(d) == "Finite") {
            # Finite distribution can have non-integer support.
            x <- parameters(d)$outcomes
          }
          mu <- mean(d)
          sigma <- stdev(d)
          x4 <- ((x - mu) / sigma)^4
          p <- eval_pmf(d, at = x)
          expect_equal(sum(p * x4), supposed_kurt)
        } else if (
          pretty_name(d) %in%
          c("Negative Binomial", "Poisson", "Geometric")
        ) {
          # Infinite support.
          expect_error(algorithm_kurtosis(d))
          mu <- mean(d)
          sigma <- stdev(d)
          to_add <- Inf
          i <- 0
          kur <- 0
          while (to_add > 1e-9) {
            x <- 0:99 + 100 * i
            to_add <- sum(eval_pmf(d, x) * ((x - mu) / sigma)^4)
            kur <- kur + to_add
            i <- i + 1
          }
          expect_equal(kur, supposed_kurt)
        } else if (pretty_name(d) == "Degenerate") {
          # Do nothing; checked in a separate script.
        } else {
          # Shouldn't be any distributions left.
          stop("At least one distribution family has not been accounted for.")
        }
      }
    }
  }
})

test_that("Quantile function calculated thru network matches known vals", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      p <- 1:99 / 100
      if (is_intrinsic(d, "quantile")) {
        if (vtype(d) == "continuous") {
          expect_equal(
            eval_quantile_from_network(d, at = p),
            eval_quantile(d, at = p)
          )
        } else if (pretty_name(d) == "Finite") {
          expect_error(eval_quantile_from_network(d, at = 0.4))
          support <- parameters(d)$outcomes
          x <- support[1] - 1
          q_derived <- numeric(0L)
          j <- 0
          for (i in seq_along(p)) {
            p_ <- p[i]
            below <- eval_cdf(d, at = x) < p_
            while (below) {
              j <- j + 1
              x <- support[j]
              below <- eval_cdf(d, at = x) < p_
            }
            q_derived[i] <- x
          }
          expect_equal(
            q_derived,
            eval_quantile(d, at = p)
          )
        } else if (pretty_name(d) != "Degenerate") {
          expect_error(eval_quantile_from_network(d, at = 0.4))
          # Distributions in distionary start at 0 at least.
          x <- -1
          q_derived <- numeric(0L)
          for (i in seq_along(p)) {
            p_ <- p[i]
            below <- eval_cdf(d, at = x) < p_
            while (below) {
              x <- x + 1
              below <- eval_cdf(d, at = x) < p_
            }
            q_derived[i] <- x
          }
          expect_equal(
            q_derived,
            eval_quantile(d, at = p)
          )
        }
      }
    }
  }
})

#' Build a Distribution Object
#'
#' Make a distribution object by specifying properties
#' (e.g., cdf, density, mean, etc.).
#' Some properties, if not included, will be calculated based
#' on other properties that are included (e.g., quantile from cdf;
#' variance from standard deviation).
#' A list of these representations can be found in the details.
#'
#' @param ... Name-value pairs for defining the distribution.
#' @param .vtype The variable type, typically "discrete" or "continuous".
#' Can be any character vector of length 1, but is converted to
#' lowercase with `tolower()` for compliance with known types.
#' @param .name A name to give to the distribution.
#' Can be any character vector of length 1.
#' @param .parameters A named list with one entry per distribution parameter,
#' each of which can be any data type. In this version of distionary,
#' these parameters are only stored for the benefit of the user to know
#' what distribution they are working with; the code never looks at these
#' parameters to inform its calculations. This is anticipated to change in
#' a future version of distionary.
#' @return A distribution object.
#' @details
#' Currently, the CDF (`cdf`) is required to be specified, along with the PMF
#' (`pmf`) for discrete distributions and density (`density`) for continuous
#' distributions. Otherwise, the full extent of distribution properties will
#' not be accessible.
#'
#' A distributional representation is a function that fully describes the
#' distribution. Besides `cdf`, `density`, and `pmf`, other options
#' understood by `distionary` include:
#'
#' - `survival`: the survival function, or one minus the cdf.
#' - `hazard`: the hazard function, for continuous variables only.
#' - `chf`: the cumulative hazard function, for continuous variables only.
#' - `quantile`: the quantile function, or left-inverse of the cdf.
#' - `realise` or `realize`: a function that takes an integer and generates
#'   a vector of that many random draws from the distribution.
#' - `odds`: for discrete variables, the probability odds function
#'   (pmf / (1 - pmf))
#' - `return`: the quantiles associated with the provided return periods,
#'   where events are exceedances.
#'
#' All functions should be vectorized.
#'
#' Other properties that are understood by `distionary` include:
#'
#' - `mean`, `stdev`, `variance`, `skewness`, `median` are self-explanatory.
#' - `kurtosis_exc` and `kurtosis` are the distribution's excess
#'   kurtosis and regular kurtosis.
#' - `range`: A vector of the minimum and maximum value of a distribution's
#'   support.
#' @examples
#' linear <- distribution(
#'   density = function(x) {
#'     d <- 2 * (1 - x)
#'     d[x < 0 | x > 1] <- 0
#'     d
#'   },
#'   cdf = function(x) {
#'     p <- 2 * x * (1 - x / 2)
#'     p[x < 0] <- 0
#'     p[x > 1] <- 1
#'     p
#'   },
#'   .vtype = "continuous",
#'   .name = "My Linear",
#'   .parameters = list(could = "include", anything = data.frame(x = 1:10))
#' )
#'
#' # Inspect
#' linear
#'
#' # Plot
#' plot(linear)
#' @family Distribution Construction
#' @export
distribution <- function(...,
                         .vtype = NULL,
                         .name = NULL,
                         .parameters = list()) {
  if (!is.null(.vtype)) {
    .vtype <- as.character(.vtype)
  } else {
    .vtype <- "unknown"
  }
  if (!is.null(.name)) {
    .name <- as.character(.name)
  } else {
    .name <- "Unnamed"
  }
  checkmate::assert_character(.vtype, len = 1, null.ok = TRUE)
  checkmate::assert_character(.name, len = 1, null.ok = TRUE)
  checkmate::assert_list(.parameters, names = "named", null.ok = TRUE)
  .vtype <- tolower(.vtype)
  dots <- rlang::enquos(...)
  checkmate::assert_list(dots, names = "named", null.ok = TRUE)
  representations <- lapply(dots, rlang::eval_tidy)
  # Check for required properties.
  reps_missing <- is.null(representations$cdf) ||
    (is.null(representations$density) && is.null(representations$pmf))
  if (reps_missing) {
    warning(
      "Full suite of distribution properties may not be accessible ",
      "without specifying 'cdf', and either 'density' or 'pmf'."
    )
  }
  # Typo detection for variable type.
  vtypes <- c("discrete", "continuous", "ordinal", "categorical", "mixed")
  vtype_match <- agrep(.vtype, vtypes, max.distance = 0.1, value = TRUE)
  if (length(vtype_match) > 0 && !(.vtype %in% vtype_match)) {
    warning(paste0(
      "The .vtype '", .vtype, "' looks similar to ",
      paste(vtype_match, collapse = ", "), "."
    ))
  }
  new_distribution(
    representations,
    vtype = .vtype, name = .name, parameters = .parameters
  )
}

test_that("Object is a distribution.", {
  expect_true(is_distribution(distribution()))
})

test_that("distribution() edge cases satisfied.", {
  expect_true(is_distribution(distribution()))
  expect_error(distribution(1:10))
  expect_error(distribution(.parameters = "foofy"))
  expect_error(distribution(.parameters = list("foofy")))
  expect_error(distribution(.parameters = c(alpha = 4)))
  expect_error(distribution(.name = c("my", "name", "is")))
  expect_error(distribution(.name = character(0)))
  expect_error(distribution(.vtype = c("my", "name", "is")))
  expect_error(distribution(.vtype = character(0)))
})

test_that("Typo warning works: vtype", {
  expect_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "discreet")
  )
  expect_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "continus")
  )
  expect_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "mixedd")
  )
  expect_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "categorcal")
  )
  expect_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "ordinale")
  )
  expect_no_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "discrete")
  )
  expect_no_warning(
    distribution(cdf = pnorm, density = dnorm, .vtype = "my_type")
  )
})


test_that("Warning when cdf and pmf/density missing.", {
  expect_warning(
    distribution()
  )
  expect_warning(
    distribution(pmf = dpois, quantile = qpois)
  )
  expect_warning(
    distribution(density = dnorm, quantile = qnorm)
  )
  expect_no_warning(
    distribution(cdf = pnorm, density = dnorm)
  )
  expect_no_warning(
    distribution(cdf = pnorm, pmf = dpois)
  )
})

test_that("Network is invoked in priority: stdev", {
  # First look for `variance`, then invoke algorithm if not found.
  d <- distribution(
    density = function(x) {
      stats::dnorm(x, sd = 3)
    },
    range = c(-Inf, Inf),
    variance = 100, # deliberately incorrect
    .vtype = "continuous"
  )
  expect_equal(stdev(d), 10)
  expect_equal(eval_stdev_from_network(d), 10)
  d$variance <- NULL
  expect_equal(stdev(d), 3)
  expect_equal(eval_stdev_from_network(d), 3)
})

test_that("stdev matches variance.", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (is_intrinsic(d, "variance") || is_intrinsic(d, "stdev")) {
        expect_equal(sqrt(variance(d)), stdev(d))
      }
    }
  }
})

#' Null Distribution
#'
#' Sometimes it's convenient to work with a distribution object that is
#' akin to a missing value. This is especially true when programmatically
#' outputting distributions, such as when a distribution fails to fit to
#' data. This function makes such a distribution object. It always evaluates
#' to `NA`.
#'
#' @returns A Null distribution.
#' @examples
#' x <- dst_null()
#' mean(x)
#' eval_pmf(x, at = 1:10)
#' @export
dst_null <- function() {
  distribution(
    .parameters = NULL,
    cdf = function(x) {
      rep(NA_real_, length(x))
    },
    density = function(x) {
      rep(NA_real_, length(x))
    },
    hazard = function(x) {
      rep(NA_real_, length(x))
    },
    chf = function(x) {
      rep(NA_real_, length(x))
    },
    pmf = function(x) {
      rep(NA_real_, length(x))
    },
    odds = function(x) {
      rep(NA_real_, length(x))
    },
    return = function(x) {
      rep(NA_real_, length(x))
    },
    quantile = function(x) {
      rep(NA_real_, length(x))
    },
    mean = NA_real_,
    variance = NA_real_,
    skewness = NA_real_,
    kurtosis_exc = NA_real_,
    .name = "Null",
    .vtype = NA_character_
  )
}

test_that("Pretty name is pretty.", {
  for (item in test_distributions) {
    d <- rlang::exec(item$distribution, !!!item$valid[[1]])
    prettynm <- pretty_name(d)
    expect_length(prettynm, 1)
    expect_true(!is.na(prettynm))
    prettynm <- pretty_name(d, 3)
    expect_length(prettynm, 1)
    expect_true(!is.na(prettynm))
  }
  expect_equal(pretty_name(dst_null()), "Null")
  expect_equal(pretty_name(dst_null(), 3), "Null")
})

test_that("Built-in distributions resolve to Null dist with NA param.", {
  for (item in test_distributions) {
    paramset <- item$valid[[1]]
    for (i in seq_along(paramset)) {
      this_paramset <- paramset
      this_paramset[[i]] <- NA
      expect_equal(
        rlang::exec(!!item$distribution, !!!this_paramset),
        dst_null()
      )
    }
  }
})

test_that("Built-in distributions error with non-numeric input.", {
  for (item in test_distributions) {
    paramset <- item$valid[[1]]
    for (i in seq_along(paramset)) {
      this_paramset <- paramset
      this_paramset[[i]] <- "peekaboo"
      expect_error(rlang::exec(!!item$distribution, !!!this_paramset))
      this_paramset[[i]] <- TRUE
      expect_error(rlang::exec(!!item$distribution, !!!this_paramset))
      this_paramset[[i]] <- list(5)
      expect_error(rlang::exec(!!item$distribution, !!!this_paramset))
      this_paramset[[i]] <- factor("peekaboo")
      expect_error(rlang::exec(!!item$distribution, !!!this_paramset))
    }
  }
})

test_that("Built-in distributions require parameters of length 1.", {
  for (item in test_distributions) {
    paramset <- item$valid[[1]]
    for (i in seq_along(paramset)) {
      this_paramset <- paramset
      this_paramset[[i]] <- rep(paramset[[i]], 2)
      expect_error(rlang::exec(!!item$distribution, !!!this_paramset))
      this_paramset[[i]] <- numeric(0L)
      expect_error(rlang::exec(!!item$distribution, !!!this_paramset))
    }
  }
})

test_that("Some built-in distributions have invalid parameter ranges.", {
  for (item in test_distributions) {
    for (paramset in item$invalid) {
      expect_error(rlang::exec(item$distribution, !!!paramset))
    }
  }
})

#' Aggregate discrete values
#'
#' Aggregates discrete values together with their weights
#' into a data frame or tibble.
#'
#' @param y Vector of outcomes.
#' @param weights Vector of weights, one for each of `y`.
#' These need not sum to one, but must not be negative and non-NA.
#' @param sum_to_one Logical; should the weights be normalized
#' to sum to 1? Default is FALSE.
#' @return Data frame with the following columns:
#' - `y`: Increasing vector of unique values of `y` that have positive weight.
#' - `weight`: Weights corresponding to each outcome.
#' @details
#' For a vector of outcomes `y` with a matching vector of `weights`,
#' `aggregate_weights()` provides a single non-zero, non-NA
#' weight per unique value of `y`.
#' @keywords internal
aggregate_weights <- function(y,
                              weights,
                              sum_to_one = FALSE) {
  checkmate::assert_numeric(y)
  checkmate::assert_numeric(weights, lower = 0, any.missing = FALSE)
  checkmate::assert_logical(sum_to_one, len = 1L, any.missing = FALSE)
  if (length(y) != length(weights)) {
    stop("`y` and `weights` must have the same length.")
  }
  zero_w <- weights == 0
  y <- y[!zero_w]
  weights <- weights[!zero_w]
  if (length(y) == 0L) {
    return(convert_dataframe_to_tibble(
      data.frame(location = numeric(), size = numeric())
    ))
  }
  if (sum_to_one) {
    weights <- weights / sum(weights)
  }
  grps <- match(y, y)
  agg_w <- tapply(weights, grps, FUN = sum, simplify = TRUE)
  agg_y <- tapply(y, grps, FUN = unique, simplify = TRUE)
  df <- data.frame(y = agg_y, weight = agg_w)
  convert_dataframe_to_tibble(df)
}

#' Find the probability left or right of a number
#'
#' Probability to the left or right of a number, inclusive or not.
#' `prob_left()` is a more general cdf defined using either `<` or `<=`, and
#' `prob_right()` is a more general survival function defined using either
#' `>` or `>=`.
#'
#' @param distribution Distribution to find probabilities of.
#' @param of Find the probability to the left or right *of* this number.
#' Could be a vector.
#' @param inclusive Should `of` be included in the probability calculation?
#' Logical.
#' @returns A vector of probabilities.
#' @rdname flexible_cdf
#' @examples
#' d <- dst_pois(5)
#' prob_left(d, of = 3, inclusive = TRUE)
#' prob_left(d, of = 3, inclusive = FALSE)
#' prob_right(d, of = 0:3, inclusive = TRUE)
#' @export
prob_left <- function(distribution, of, inclusive) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(of)
  checkmate::assert_logical(inclusive, len = 1)
  p_left <- eval_cdf(distribution, at = of)
  if (!inclusive) {
    p_break <- eval_pmf(distribution, at = of)
    p_left <- p_left - p_break
  }
  p_left
}

#' @rdname flexible_cdf
#' @export
prob_right <- function(distribution, of, inclusive) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(of)
  checkmate::assert_logical(inclusive, len = 1)
  p_right <- eval_survival(distribution, at = of)
  if (inclusive) {
    p_break <- eval_pmf(distribution, at = of)
    p_right <- p_right + p_break
  }
  p_right
}

#' Generate a Sample from a Distribution
#'
#' Draw `n` independent observations from a distribution.
#'
#' @param distribution Distribution object.
#' @param n Number of observations to generate.
#' @returns Vector of independent values drawn from the inputted distribution.
#' @note `realise()` and `realize()` are aliases and do the same thing.
#' @rdname realise
#' @examples
#' d <- dst_pois(5)
#' set.seed(2)
#' realise(d, n = 10)
#' @export
realise <- function(distribution, n = 1) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_integerish(n, lower = 0, len = 1)
  eval_property(distribution, "realise", n)
}

#' @rdname realise
#' @export
realize <- function(distribution, n = 1) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_integerish(n, lower = 0, len = 1)
  eval_property(distribution, "realize", n)
}

#' Student t Distribution
#'
#' Makes a Student t distribution.
#'
#' @param df Degrees of freedom; single positive numeric.
#' @returns A Student t distribution.
#' @examples
#' dst_t(3)
#' @export
dst_t <- function(df) {
  checkmate::assert_numeric(df, 0, len = 1)
  if (is.na(df)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(df = df),
    density = function(x) {
      stats::dt(x, df = df)
    },
    cdf = function(x) {
      stats::pt(x, df = df)
    },
    quantile = function(p) {
      stats::qt(p, df = df)
    },
    realise = function(n) {
      stats::rt(n, df = df)
    },
    survival = function(x) {
      stats::pt(x, df = df, lower.tail = FALSE)
    },
    mean = ifelse(df > 1, 0, NaN),
    median = 0,
    variance = {
      if (df > 2) {
        df / (df - 2)
      } else if (df > 1) {
        Inf
      } else {
        NaN
      }
    },
    skewness = ifelse(df > 3, 0, NaN),
    kurtosis_exc = {
      if (df > 4) {
        6 / (df - 4)
      } else if (df > 2) {
        Inf
      } else {
        NaN
      }
    },
    range = c(-Inf, Inf),
    .name = "Student t",
    .vtype = "continuous",
  )
}

#' Pearson Type III distribution
#'
#' Makes a Pearson Type III distribution, which is a Gamma distribution,
#' but shifted.
#'
#' @param location Location parameter, specifying how to shift the
#' Gamma distribution; single numeric.
#' @param scale Scale parameter of the Gamma distribution;
#' single positive numeric.
#' @param shape Shape parameter of the Gamma distribution;
#' single positive numeric.
#' @returns A Pearson Type III distribution.
#' @examples
#' dst_pearson3(1, 1, 1)
#' @export
dst_pearson3 <- function(location, scale, shape) {
  checkmate::assert_numeric(location, len = 1)
  checkmate::assert_numeric(scale, 0, len = 1)
  checkmate::assert_numeric(shape, 0, len = 1)
  if (is.na(location) || is.na(scale) || is.na(shape)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(
      location = location, scale = scale, shape = shape
    ),
    cdf = function(x) {
      stats::pgamma(x - location, scale = scale, shape = shape)
    },
    survival = function(x) {
      stats::pgamma(
        x - location, scale = scale, shape = shape, lower.tail = FALSE
      )
    },
    density = function(x) {
      stats::dgamma(x - location, scale = scale, shape = shape)
    },
    quantile = function(p) {
      location + stats::qgamma(p, shape = shape, scale = scale)
    },
    realise = function(n) {
      location + stats::rgamma(n, shape = shape, scale = scale)
    },
    mean = location + scale * shape,
    variance = shape * scale^2,
    skewness = 2 / sqrt(shape),
    kurtosis_exc = 6 / shape,
    .name = "Pearson Type III",
    .vtype = "continuous"
  )
}

#' Cumulative Hazard Function
#'
#' Access a distribution's cumulative hazard function (chf).
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_unif(0, 4)
#' eval_chf(d, at = 0:4)
#' enframe_chf(d, at = 0:4)
#' @family distributional representations
#' @rdname chf
#' @export
eval_chf <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "chf", at)
}

#' @noRd
eval_chf_from_network <- function(distribution, at) {
  if (vtype(distribution) == "continuous") {
    sf <- eval_survival(distribution, at = at)
    -log(sf)
  } else {
    stop("Not valid for non-continuous distributions.")
  }
}

#' @rdname chf
#' @export
enframe_chf <- function(..., at, arg_name = ".arg", fn_prefix = "chf",
                        sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_chf
  )
}

#' CHF is never intrinsic, which means there are no known values to compare
#' against. Compare built-in algorithm with another algorithm for deriving
#' CHF: integrating the hazard.
#' This also checks that the hazard function is correct.
test_that("CHF algorithm validated by comparing to a different algorithm.", {
  # - `eval_chf()` goes through the survival function only.
  # - integrating the hazard goes through both the density and survival.
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (vtype(d) == "continuous") {
        p <- 1:9 / 10
        x <- eval_quantile(d, at = p)
        haz_fun <- representation_as_function(d, "hazard")
        increments <- vapply(
          seq_along(x[-1L]),
          function(t) {
            stats::integrate(haz_fun, x[1], x[t + 1])$value
          },
          FUN.VALUE = numeric(1L)
        )
        chf_derived <- eval_chf(d, at = x[1]) + append(0, (increments))
        expect_equal(chf_derived, eval_chf(d, at = x))
      }
    }
  }
})

test_that("enframe matches eval", {
  d <- dst_norm(0, 1)
  p <- 1:9 / 10
  x <- -3:3
  expect_equal(enframe_cdf(d, at = x)$cdf, eval_cdf(d, at = x))
  expect_equal(enframe_density(d, at = x)$density, eval_density(d, at = x))
  expect_equal(
    enframe_pmf(d, at = x)$pmf,
    eval_pmf(d, at = x)
  )
  expect_equal(
    enframe_odds(d, at = x)$odds,
    eval_odds(d, at = x)
  )
  expect_equal(enframe_survival(d, at = x)$survival, eval_survival(d, at = x))
  expect_equal(enframe_hazard(d, at = x)$hazard, eval_hazard(d, at = x))
  expect_equal(enframe_chf(d, at = x)$chf, eval_chf(d, at = x))
  expect_equal(
    enframe_quantile(d, at = p)$quantile,
    eval_quantile(d, at = p)
  )
})

test_that("enframe throws error if an ellipsis entry is not a distribution", {
  d <- dst_norm(0, 1)
  expect_error(enframe_cdf(5, at = 1:10))
  expect_error(enframe_cdf(d, 5, at = 1:10))
  expect_error(enframe_cdf(at = 1:10))
})

test_that("column names match the function, by default.", {
  d <- dst_norm(0, 1)
  expect_equal(names(enframe_cdf(d, at = 0))[2L], "cdf")
  expect_equal(names(enframe_pmf(d, at = 0))[2L], "pmf")
  expect_equal(names(enframe_odds(d, at = 0))[2L], "odds")
  expect_equal(names(enframe_density(d, at = 0))[2L], "density")
  expect_equal(names(enframe_quantile(d, at = 0.1))[2L], "quantile")
  expect_equal(names(enframe_chf(d, at = 0))[2L], "chf")
  expect_equal(names(enframe_hazard(d, at = 0))[2L], "hazard")
  expect_equal(names(enframe_survival(d, at = 0))[2L], "survival")
})

test_that("Manual column names work.", {
  d <- dst_norm(0, 1)
  x <- 1:3
  df <- enframe_cdf(d, at = x, arg_name = "one", fn_prefix = "two")
  expect_equal(sort(c("one", "two")), sort(names(df)))
  df <- enframe_cdf(
    mod1 = d, mod2 = d, at = x, arg_name = "one", fn_prefix = "two", sep = "."
  )
  expect_equal(sort(c("one", "two.mod1", "two.mod2")), sort(names(df)))
  df <- enframe_cdf(
    mod1 = d, d, at = x, arg_name = "one", fn_prefix = "two", sep = "."
  )
  expect_equal(sort(c("one", "two.mod1", "two.d")), sort(names(df)))
})

test_that("Dimensions and type are as expected.", {
  d <- dst_norm(0, 1)
  x <- -3:3
  df <- enframe_cdf(d, at = x, arg_name = "one", fn_prefix = "two")
  expect_true(is.data.frame(df))
  expect_equal(dim(df), c(7, 2))
  df <- enframe_cdf(mod1 = d, mod2 = d, at = x)
  expect_equal(dim(df), c(7, 3))
  df <- enframe_cdf(d, at = numeric())
  expect_equal(nrow(df), 0)
  expect_length(eval_cdf(d, at = numeric()), 0)
})

#' Exponential Distribution
#'
#' Makes an Exponential distribution.
#'
#' @param rate Rate parameter; single positive numeric.
#' @returns An Exponential distribution.
#' @examples
#' dst_exp(1)
#' @export
dst_exp <- function(rate) {
  checkmate::assert_numeric(rate, 0, len = 1)
  if (is.na(rate)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(rate = rate),
    density = function(x) {
      stats::dexp(x, rate = rate)
    },
    cdf = function(x) {
      stats::pexp(x, rate = rate)
    },
    quantile = function(p) {
      stats::qexp(p, rate = rate)
    },
    realise = function(n) {
      stats::rexp(n, rate = rate)
    },
    survival = function(x) {
      stats::pexp(x, rate = rate, lower.tail = FALSE)
    },
    mean = 1 / rate,
    median = log(2) / rate,
    variance = 1 / rate^2,
    skewness = 2,
    kurtosis_exc = 6,
    range = c(0, Inf),
    .name = "Exponential",
    .vtype = "continuous"
  )
}

#' @export
print.dst <- function(x, ...) {
  nm <- attributes(x)[["name"]]
  vtype <- attributes(x)[["vtype"]]
  param <- parameters(x)
  # Family name
  if (is.null(nm)) {
    cat("Unnamed distribution")
  } else {
    cat(nm, "distribution")
  }
  # Variable type
  if (!is.null(vtype)) {
    cat(paste0(" (", vtype, ")", collapse = ""), "\n")
  }
  # Parameters
  if (!is.null(param) && all(!is.na(param))) {
    all_numeric <- all(vapply(
      param,
      function(x) is.numeric(x) && length(x) == 1,
      FUN.VALUE = logical(1)
    ))
    if (all_numeric) {
      param <- unlist(param)
    }
    if (pretty_name(x) == "Finite") {
      param <- as.data.frame(param)
      param <- convert_dataframe_to_tibble(param)
    }
    cat("--Parameters--\n")
    print(param, ...)
  }
  invisible(x)
}

#' Median of a Distribution
#'
#' Finds the median of a distribution.
#'
#' @importFrom stats median
#'
#' @param x Distribution to calculate median from.
#' @param ... Not used.
#' @returns Median of a distribution; single numeric.
#' @details
#' Median is calculated as the 0.5-quantile when not found in the
#' distribution. So, when the median is
#' non-unique, the smallest of the possibilities is taken.
#' @examples
#' d <- dst_gamma(3, 3)
#' median(d)
#' @export
median.dst <- function(x, ...) {
  rlang::check_dots_empty()
  eval_property(x, "median")
}

#' Probability Density Function
#'
#' Access a distribution's probability density function (pdf).
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_unif(0, 4)
#' eval_density(d, at = 0:4)
#' enframe_density(d, at = 0:4)
#' @family distributional representations
#' @rdname density
#' @export
eval_density <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "density", at)
}

#' @noRd
eval_density_from_network <- function(distribution, at) {
  stop(
    "Cannot find density function. Density must be specified ",
    "in the distribution."
  )
}

#' @rdname density
#' @export
enframe_density <- function(..., at, arg_name = ".arg", fn_prefix = "density",
                            sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_density
  )
}

#' Test that the hazard function is correct by comparing it to the limit of
#' the probability of an event happening immediately, given that it hasn't
#' happened already, using the density function.
test_that("Hazard function is validated using the density function.", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (vtype(d) == "continuous") {
        p <- 1:9 / 10
        x <- eval_quantile(d, at = p)
        r <- range(d)
        supposed_hazard <- eval_hazard(d, at = x)
        expect_equal(
          eval_density(d, at = x) / eval_survival(d, at = x),
          supposed_hazard
        )
      }
    }
  }
})

#' Representations of the Generalized Pareto Distribution
#'
#' @param x,q Vector of quantiles.
#' @param p Vector of probabilities.
#' @param scale Vector of scale parameters; positive numeric.
#' @param shape Vector of shape parameters; positive numeric.
#' @param lower.tail Single logical. If `TRUE`, cdf (default);
#' if `FALSE`, survival function.
#' @returns Vector of evaluated GP distribution, with length
#' equal to the recycled lengths of `q`/`x`/`p`, `scale`, and `shape`.
#' @examples
#' pgp(1:10, 1, 1)
#' dgp(1:10, 2, 0)
#' qgp(1:9 / 10, 10, -2)
#' @rdname gp_raw
#' @export
pgp <- function(q, scale, shape, lower.tail = TRUE) {
  checkmate::assert_numeric(q)
  checkmate::assert_numeric(scale, 0)
  checkmate::assert_numeric(shape)
  checkmate::assert_logical(lower.tail, len = 1, any.missing = FALSE)
  l <- vctrs::vec_recycle_common(q, scale, shape)
  q <- l[[1]]
  scale <- l[[2]]
  shape <- l[[3]]
  hi <- gp_upper(scale = scale, shape = shape)
  left <- q < 0
  right <- q > hi
  t <- gev_t_function(q, location = 0, scale = scale, shape = shape)
  if (lower.tail) {
    res <- 1 - t
    res[left] <- 0
    res[right] <- 1
  } else {
    res <- t
    res[left] <- 1
    res[right] <- 0
  }
  res
}


#' @rdname gp_raw
#' @export
qgp <- function(p, scale, shape) {
  # z = (RP^xi - 1) / xi -> log(RP)
  checkmate::assert_numeric(p)
  checkmate::assert_numeric(scale, 0)
  checkmate::assert_numeric(shape)
  l <- vctrs::vec_recycle_common(p, scale, shape)
  p <- l[[1]]
  scale <- l[[2]]
  shape <- l[[3]]
  rp <- 1 / (1 - p)
  res <- ifelse(
    p >= 0 & p <= 1,
    ifelse(shape == 0, log(rp), (rp^shape - 1) / shape),
    NaN
  )
  scale * res
}

#' @rdname gp_raw
#' @export
dgp <- function(x, scale, shape) {
  checkmate::assert_numeric(x)
  checkmate::assert_numeric(scale, 0)
  checkmate::assert_numeric(shape)
  l <- vctrs::vec_recycle_common(x, scale, shape)
  x <- l[[1]]
  scale <- l[[2]]
  shape <- l[[3]]
  z <- x / scale
  hi <- gp_upper(scale = scale, shape = shape)
  t <- gev_t_function(x, location = 0, scale = scale, shape = shape)
  res <- ifelse(shape == 0, t, (1 + shape * z)^(-1 / shape - 1))
  # i_zero <- shape == 0
  # i_non0 <- shape != 0
  # res <- numeric(length(x))
  # res[i_zero] <- t[i_zero]
  # res[i_non0] <- (1 + shape[i_non0] * z[i_non0])^(-1 / shape[i_non0] - 1)
  res[x < 0 | x > hi] <- 0
  res / scale
}


#' Upper endpoint of GP
#'
#' @inheritParams dst_gp
#' @returns A vector of maxima corresponding to the input parameters.
#' @noRd
gp_upper <- function(scale, shape) {
  l <- vctrs::vec_recycle_common(scale, shape)
  scale <- l[[1]]
  shape <- l[[2]]
  res <- shape # For NA propagation
  ifelse(shape < 0, -scale / shape, Inf)
  # i_calc <- shape < 0
  # res[i_calc] <- -scale[i_calc] / shape[i_calc]
  # res[shape >= 0] <- Inf
  # res
}

test_that("GP quantities match known cases, shape > 0", {
  scale <- 1
  shape <- 1
  med <- scale * (2^shape - 1) / shape
  expect_equal(
    pgp(c(-1, 0, med), scale = scale, shape = shape), c(0, 0, 0.5)
  )
  expect_equal(
    qgp(c(-1, 0, 0.5, 1, 2), scale = scale, shape = shape),
    c(NaN, 0, med, Inf, NaN)
  )
  expect_equal(dgp(c(-1, Inf), scale = scale, shape = shape), c(0, 0))
  expect_true(all(diff(dgp(1:10, scale = scale, shape = shape)) < 0))
  pdf <- function(x) {
    dgp(x, scale = scale, shape = shape)
  }
  expect_equal(stats::integrate(pdf, 0, Inf)$value, 1, tolerance = 0.0001)
})


test_that("GP quantities match known cases, shape = 0", {
  scale <- 1
  shape <- 0
  med <- log(2)
  expect_equal(
    pgp(c(-1, 0, med, Inf), scale = scale, shape = shape), c(0, 0, 0.5, 1)
  )
  expect_equal(
    suppressWarnings(qgp(c(-1, 0, 0.5, 1, 2), scale = scale, shape = shape)),
    c(NaN, 0, med, Inf, NaN)
  )
  expect_equal(dgp(c(-1, Inf), scale = scale, shape = shape), c(0, 0))
  pdf <- function(x) {
    dgp(x, scale = scale, shape = shape)
  }
  expect_true(all(diff(pdf(1:10)) < 0))
  expect_equal(stats::integrate(pdf, 0, Inf)$value, 1, tolerance = 0.0001)
})

test_that("GP quantities match known cases, shape < 0", {
  scale <- 2
  shape <- -1
  rightend <- -scale / shape
  med <- scale * (2^shape - 1) / shape
  expect_equal(
    pgp(
      c(-1, 0, med, rightend, rightend + 1, Inf),
      scale = scale, shape = shape
    ),
    c(0, 0, 0.5, 1, 1, 1)
  )
  expect_equal(
    qgp(c(-1, 0, 0.5, 1, 2), scale = scale, shape = shape),
    c(NaN, 0, med, rightend, NaN)
  )
  expect_equal(
    dgp(c(-1, rightend + 1, Inf), scale = scale, shape = shape),
    c(0, 0, 0)
  )
  pdf <- function(x) {
    dgp(x, scale = scale, shape = shape)
  }
  expect_equal(stats::integrate(pdf, 0, rightend)$value, 1, tolerance = 0.0001)
})

test_that("GP negative scale not allowed.", {
  expect_error(dgev(1:10, -4, 1))
  expect_error(pgev(1:10, -4, 1))
  expect_error(qgev(1:10 / 10, -4, 1))
})

test_that("cdf and pdf of GP align via numerical derivative.", {
  d <- list(
    dst_gp(1, 1),
    dst_gp(1, 0),
    dst_gp(10, -1)
  )
  x <- 1:11
  eps <- 1e-6
  for (i in seq_along(d)) {
    pdf <- eval_density(d[[i]], at = x)
    cdf1 <- eval_cdf(d[[i]], at = x - eps)
    cdf2 <- eval_cdf(d[[i]], at = x)
    pdf_num <- (cdf2 - cdf1) / eps
    expect_equal(pdf, pdf_num, tolerance = 1e-6)
  }
})

test_that("cdf and qf of GP align.", {
  d <- list(
    dst_gp(1, 1),
    dst_gp(1, 0),
    dst_gp(10, -1)
  )
  p <- 1:9 / 10
  for (i in seq_along(d)) {
    qf <- eval_quantile(d[[i]], at = p)
    cdf <- eval_cdf(d[[i]], at = qf)
    expect_equal(cdf, p)
  }
})

test_that("quantile function of GP is valid, validating the distribution.", {
  d <- list(
    list(scale = 1.2, shape = 1),
    list(scale = 1.2, shape = 0),
    list(scale = 10, shape = -1)
  )
  p <- 0:100 / 100
  for (i in seq_along(d)) {
    hi <- rlang::exec(distionary:::gp_upper, !!!d[[i]])
    qf <- rlang::exec(qgp, p, !!!d[[i]])
    expect_true(all(diff(qf) > 0))
    expect_equal(qf[c(1L, 101L)], c(0, hi))
  }
})

test_that("vectorisation of p/d/q GP functions works.", {
  y <- 1:10
  x <- 1:10
  x[4] <- NA_real_
  x[2] <- NaN
  p <- x / 11
  v <- y / 11
  # Mismatched lengths
  expect_error(pgp(x, 1:2, 1))
  expect_error(pgp(x, 1, 0:1))
  expect_error(dgp(x, 1:2, 1))
  expect_error(dgp(x, 1, 0:1))
  expect_error(qgp(p, 1:2, 1))
  expect_error(qgp(p, 1, 0:1))
  expect_error(distionary:::gp_lower(1:2, 1:4))
  expect_error(distionary:::gp_upper(1:2, 1:4))
  # Lengths input correctly; should be length 10.
  expect_length(pgp(x, 1, 1), 10)
  expect_length(dgp(x, 1, 1), 10)
  expect_length(qgp(p, 1, 1), 10)
  expect_length(pgp(x, 1:10, 1), 10)
  expect_length(dgp(x, 1:10, 1), 10)
  expect_length(qgp(p, 1:10, 1), 10)
  # NA and NaN gets projected forward. NaN may convert to NA.
  # --> main argument
  expect_true(is.na(pgp(x, 1, 1)[4]))
  expect_true(is.na(dgp(x, 1, 1)[4]))
  expect_true(is.na(qgp(p, 1, 1)[4]))
  expect_true(is.na(pgp(x, 1, 1)[2]))
  expect_true(is.na(dgp(x, 1, 1)[2]))
  expect_true(is.na(qgp(p, 1, 1)[2]))
  # --> scale
  expect_true(is.na(pgp(y, x, 1)[4]))
  expect_true(is.na(dgp(y, x, 1)[4]))
  expect_true(is.na(qgp(v, x, 1)[4]))
  expect_true(is.na(pgp(y, x, 1)[2]))
  expect_true(is.na(dgp(y, x, 1)[2]))
  expect_true(is.na(qgp(v, x, 1)[2]))
  # --> shape
  expect_true(is.na(pgp(y, 1, x)[4]))
  expect_true(is.na(dgp(y, 1, x)[4]))
  expect_true(is.na(qgp(v, 1, x)[4]))
  expect_true(is.na(pgp(y, 1, x)[2]))
  expect_true(is.na(dgp(y, 1, x)[2]))
  expect_true(is.na(qgp(v, 1, x)[2]))
  # Max
  expect_length(distionary:::gp_upper(x, 1), 10)
  expect_length(distionary:::gp_upper(1, x), 10)
})

#' Finite Distribution
#'
#' Makes a finite distribution, which is a distribution with a finite number
#' of possible outcomes.
#'
#' @param outcomes Numeric vector representing the potential
#' outcomes of the distribution.
#' @param probs Numeric vector of probabilities corresponding to the outcomes
#' in `outcomes`. Must not be negative and must sum to 1.
#' @returns A distribution with finite outcomes.
#' @seealso [dst_empirical()]
#' @examples
#' dst_finite(2:5, probs = 1:4 / 10)
#' @export
dst_finite <- function(outcomes, probs) {
  if (any(is.na(outcomes)) || any(is.na(probs))) {
    return(dst_null())
  }
  if (length(outcomes) != length(probs)) {
    stop("Inputs `outcomes` and `probs` should have the same length.")
  }
  if (length(outcomes) == 0) {
    warning(
      "Can't make a finite distribution from empty data. ",
      "Returning a Null distribution."
    )
    return(dst_null())
  }
  mask <- probs != 0
  outcomes <- outcomes[mask]
  probs <- probs[mask]
  if (any(probs < 0)) {
    stop("Probabilities must be between 0 and 1.")
  }
  if (abs(sum(probs) - 1) > .Machine$double.eps * length(probs)) {
    stop("Probabilities must add up to 1.")
  }
  if (length(outcomes) == 1L) {
    return(dst_degenerate(outcomes))
  }
  ## Helpful Quantities
  mu <- sum(probs * outcomes)
  ss <- sum(probs * (outcomes - mu)^2)
  sigma <- sqrt(ss)
  ## Sort outcomes and probabilities for the stepfuns
  order <- order(outcomes)
  outcomes <- outcomes[order]
  probs <- probs[order]
  cumsum_p <- cumsum(probs)
  heights <- c(0, cumsum_p)
  taus <- cumsum_p[-length(cumsum_p)]
  distribution(
    .parameters = list(outcomes = outcomes, probs = probs),
    pmf = function(q) {
      na_mask <- is.na(q)
      matched <- match(q, outcomes)
      res <- probs[matched]
      res[is.na(matched)] <- 0
      res[na_mask] <- NA_real_
      res
    },
    cdf = stats::stepfun(outcomes, heights, right = FALSE),
    quantile = stats::stepfun(taus, outcomes, right = TRUE),
    realise = function(n) {
      sample(outcomes, size = n, replace = TRUE, prob = probs)
    },
    survival = stats::stepfun(outcomes, 1 - heights, right = FALSE),
    mean = mu,
    variance = ss,
    skewness = sum(probs * ((outcomes - mu) / sigma)^3),
    kurtosis_exc = sum(probs * ((outcomes - mu) / sigma)^4) - 3,
    range = range(outcomes),
    .vtype = "discrete",
    .name = "Finite"
  )
}

#' Integration regime used in distionary
#'
#' Integrating distribution-related functions can be finicky. This function
#' is designed to get around the shortcomings of built-in integration
#' regimes like stats::integrate and cubature::hcubature.
#'
#' @param fun Function (integrand) to be integrated.
#' @param lower,upper Limits of integration. Can be infinite.
#' @param tol Tolerance; small positive number.
#' @param ... Other arguments to pass to `cubature::hcubature()`.
#' @returns A single numeric, the integral of the function.
#' @details Here are the integration shortfalls encountered with distributions:
#'
#' - The mean of a standard Cauchy distribution integrates to 0, when it
#'   should not exist. This happens because the function is symmetric and
#'   the negative part cancels out the positive part numerically, but
#'   both sides of the integral are not well-defined.
#'
#' The solution involves a simple `if` statement to break the integral in two.
#'
#' NOTE: if `cubature::hcubature()` is used, there have been other challenges,
#' and so it's best to use `stats::integrate()`.
#'
#' - Sometimes, if either end of the integral is finite but evaluates to
#'   infinity (like the integrand for calculating the mean of the
#'   Beta(0.5, 0.5) distribution through its density), the integral evaluates
#'   to infinity.
#' - The integrand for the Cauchy(0, 1) mean via the density function should
#'   be infinite on either side of x=0, but comes back as finite.
#'
#' **Change of variables math if having trouble integrating to a finite
#'   endpoint that evaluates to infinity**
#' Integral is I = int f(x) dx, x = x0..x1
#'
#' _Upper Integral_
#' Let t = 1 / (x1 - x)
#' ==> x = x1 - 1/t
#' t goes from 1 / (x1 - x0) .. Inf
#' dx = 1/t^2 dt
#' Integral becomes:
#' I = int f(x1 - 1/t) / t^2 dt
#'
#' _Lower Integral_
#' Let t = 1 / (x - x0)
#' ==> x = x0 + 1/t
#' dx = -1/t^2 dt
#' t goes from Inf .. 1 / (x1 - x0)
#' Integral becomes
#' I = int -f(x0 + 1/t) / t^2 dt
#' Or, switch the order of integration, and get
#' I = int  f(x0 + 1/t) / t^2 dt
#'
#' _Both_
#' When both finite endpoints have non-finite evaluation, split the
#' integral in half using the midpoint.
#'
#' @examples
#' # Beta(0.5, 0.5) should have mean of 0.5.
#' integrand <- function(x) {
#'   x * stats::dbeta(x, 0.5, 0.5)
#' }
#' plot(integrand)
#' distionary_integrate(integrand, 0, 1)
#' stats::integrate(integrand, 0, 1)  # Good
#' cubature::hcubature(integrand, 0, 1, tol = 1e-7) # Bad
#' cubature::hcubature(integrand, 0, 1)  # Works but not as accurate
#'
#' # Cauchy(0, 1)
#' integrand <- function(x) {
#'   x * stats::dcauchy(x)
#' }
#' plot(integrand, -10, 10)
#' distionary_integrate(integrand, 0, Inf) # NaN. Ideally would be Inf.
#' distionary_integrate(integrand, -Inf, Inf) # NaN, good.
#' stats::integrate(integrand, -Inf, Inf)  # 0, bad.
#' cubature::hcubature(integrand, -Inf, Inf) # Sometimes hangs, or NaN.
#' @noRd
distionary_integrate <- function(fun, lower, upper, tol = 1e-9, ...) {
  checkmate::assert_function(fun)
  checkmate::assert_numeric(lower, len = 1)
  checkmate::assert_numeric(upper, lower = lower, len = 1)
  checkmate::assert_numeric(tol, lower = 0, len = 1)
  if (lower == upper) return(0)
  if (is.infinite(lower) && is.infinite(upper)) {
    ## Solves the problem of the standard Cauchy mean evaluating to 0,
    ## and the t(3) distribution evaluating to finite skewness.
    i_lower <- distionary_integrate(
      fun,
      lower = lower, upper = 0,
      tol = tol / 2,
      ...
    )
    i_upper <- distionary_integrate(
      fun,
      lower = 0, upper = upper,
      tol = tol / 2,
      ...
    )
    return(i_lower + i_upper)
  }
  int <- try(
    stats::integrate(fun, lower, upper, rel.tol = tol, ...)$value,
    silent = TRUE
  )
  if (inherits(int, "try-error")) {
    message(
      "Integration routine for numerical computation failed. ",
      "This could be because the property is not well-defined. Returning NaN."
    )
    return(NaN)
  }
  int
}

#' Chi-Squared Distribution
#'
#' Makes a Chi-Squared distribution.
#'
#' @param df degrees of freedom parameter; single positive numeric.
#' @returns A Chi-Squared distribution
#' @examples
#' dst_chisq(3)
#' @export
dst_chisq <- function(df) {
  checkmate::assert_numeric(df, 0, len = 1)
  if (is.na(df)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(df = df),
    density = function(x) {
      stats::dchisq(x, df = df)
    },
    cdf = function(x) {
      stats::pchisq(x, df = df)
    },
    quantile = function(p) {
      stats::qchisq(p, df = df)
    },
    realise = function(n) {
      stats::rchisq(n, df = df)
    },
    survival = function(x) {
      stats::pchisq(x, df = df, lower.tail = FALSE)
    },
    mean = df,
    median = df * (1 - 2 / 9 * df)^3,
    variance = 2 * df,
    skewness = sqrt(8 / df),
    kurtosis_exc = 12 / df,
    range = c(0, Inf),
    .name = "Chi-Squared",
    .vtype = "continuous"
  )
}

#' Odds Function
#'
#' Access a distribution's odds function. The odds of an event having
#' probability `p` is `p / (1 - p)`.
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_pois(1)
#' eval_pmf(d, at = c(1, 2, 2.5))
#' eval_odds(d, at = c(1, 2, 2.5))
#' enframe_odds(d, at = 0:4)
#' @family distributional representations
#' @rdname odds
#' @export
eval_odds <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "odds", at)
}

#' @noRd
eval_odds_from_network <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  p <- eval_pmf(distribution, at = at)
  p / (1 - p)
}

#' @rdname odds
#' @export
enframe_odds <- function(..., at, arg_name = ".arg", fn_prefix = "odds",
                         sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_odds
  )
}

test_that("Odds matches some known values", {
  expect_equal(eval_odds(dst_bern(0.5), at = 1), 1)
  expect_equal(eval_odds(dst_bern(0.2), at = 1), 0.25)
  expect_equal(eval_odds(dst_bern(0.8), at = 1), 4)
  d <- dst_pois(4)
  expect_equal(
    eval_odds(d, at = 0:10),
    eval_pmf(d, at = 0:10) / (1 - eval_pmf(d, at = 0:10))
  )
})

#' Log Normal Distribution
#'
#' Makes a Log Normal distribution, which is the distribution of
#' the exponential of a Normally distributed random variable.
#' @param meanlog Mean of the log of the random variable;
#' single numeric.
#' @param sdlog Standard deviation of the log of the random variable;
#' single positive numeric.
#' @returns A Log Normal distribution.
#' @examples
#' dst_lnorm(0, 1)
#' @export
dst_lnorm <- function(meanlog, sdlog) {
  checkmate::assert_numeric(meanlog, len = 1)
  checkmate::assert_numeric(sdlog, 0, len = 1)
  if (is.na(meanlog) || is.na(sdlog)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(meanlog = meanlog, sdlog = sdlog),
    density = function(x) {
      stats::dlnorm(x, meanlog = meanlog, sdlog = sdlog)
    },
    cdf = function(x) {
      stats::plnorm(x, meanlog = meanlog, sdlog = sdlog)
    },
    quantile = function(p) {
      stats::qlnorm(p, meanlog = meanlog, sdlog = sdlog)
    },
    realise = function(n) {
      stats::rlnorm(n, meanlog = meanlog, sdlog = sdlog)
    },
    survival = function(x) {
      stats::plnorm(x, meanlog = meanlog, sdlog = sdlog, lower.tail = FALSE)
    },
    mean = exp(meanlog + sdlog^2 / 2),
    median = exp(meanlog),
    variance = {
      ev <- exp(sdlog^2)
      (ev - 1) * ev * exp(2 * meanlog)
    },
    skewness = {
      ev <- exp(sdlog^2)
      (ev + 2) * sqrt(ev - 1)
    },
    kurtosis_exc = {
      e4 <- exp(4 * sdlog^2)
      e3 <- exp(3 * sdlog^2)
      e2 <- exp(2 * sdlog^2)
      e4 + 2 * e3 + 3 * e2 - 6
    },
    range = c(0, Inf),
    .name = "Log Normal",
    .vtype = "continuous"
  )
}

#' Probability Mass Function
#'
#' Access a distribution's probability mass function (pmf).
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_pois(5)
#' eval_pmf(d, at = c(1, 2, 2.5))
#' enframe_pmf(d, at = 0:4)
#' eval_pmf(dst_norm(0, 1), at = -3:3)
#' @family distributional representations
#' @rdname pmf
#' @export
eval_pmf <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "pmf", at)
}

#' @noRd
eval_pmf_from_network <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  if (vtype(distribution) == "continuous") {
    at[!is.na(at)] <- 0
    return(at)
  }
  stop("Cannot find pmf, which must be specified in the distribution.")
}

#' @rdname pmf
#' @export
enframe_pmf <- function(..., at, arg_name = ".arg", fn_prefix = "pmf",
                        sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_pmf
  )
}

# Validate PMF can be recovered from CDF
#
# Systematic testing uses the `test_distributions` object.
# To modify it, see `data-raw/test_distributions.R`.
# To update it, run `Rscript data-raw/test_distributions.R`.


test_that("PMF calculated thru CDF matches known vals.", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (is_intrinsic(d, "pmf")) {
        expect_false(is_intrinsic(d, "density"))
        p <- 1:99 / 100
        x <- unique(eval_quantile(d, at = p))
        pmf_evald <- eval_cdf(d, at = x) - eval_cdf(d, at = x - 1e-6)
        expect_equal(pmf_evald, eval_pmf(d, at = x))
      }
    }
  }
})

#' Convert a data frame to a tibble
#'
#' Converts a data frame to a tibble, if the user has the `tibble`
#' package installed.
#' @param res Data frame.
#' @returns A tibble, if the user has the `tibble` package installed.
#' @noRd
convert_dataframe_to_tibble <- function(res) {
  checkmate::assert_data_frame(res)
  if (requireNamespace("tibble", quietly = TRUE)) {
    res <- tibble::as_tibble(res)
  }
  res
}

#' Cauchy Distribution
#'
#' Makes a Cauchy distribution.
#'
#' @param location Location parameter; single numeric.
#' @param scale Scale parameter; single positive numeric.
#' @returns A Cauchy distribution.
#' @examples
#' d <- dst_cauchy(0, 1)
#'
#' # Moments do not exist for the Cauchy distribution.
#' mean(d)
#' variance(d)
#' @export
dst_cauchy <- function(location, scale) {
  checkmate::assert_numeric(location, len = 1)
  checkmate::assert_numeric(scale, 0, len = 1)
  if (is.na(location) || is.na(scale)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(location = location, scale = scale),
    density = function(x) {
      stats::dcauchy(x, location = location, scale = scale)
    },
    cdf = function(x) {
      stats::pcauchy(x, location = location, scale = scale)
    },
    quantile = function(p) {
      stats::qcauchy(p, location = location, scale = scale)
    },
    realise = function(n) {
      stats::rcauchy(n, location = location, scale = scale)
    },
    survival = function(x) {
      stats::pcauchy(x, location = location, scale = scale, lower.tail = FALSE)
    },
    mean = NaN,
    median = location,
    variance = NaN,
    skewness = NaN,
    kurtosis_exc = NaN,
    range = c(-Inf, Inf),
    .vtype = "continuous",
    .name = "Cauchy"
  )
}

#' Binomial Distribution
#'
#' Makes a Binomial distribution, representing the number of successes
#' in a fixed number of independent trials.
#'
#' @param size Number of trials; single positive integer.
#' @param prob Success probability of each trial; single numeric
#' between 0 and 1.
#' @returns A binomial distribution.
#' @examples
#' dst_binom(10, 0.6)
#' @export
dst_binom <- function(size, prob) {
  checkmate::assert_integerish(size, lower = 0, len = 1)
  checkmate::assert_numeric(prob, 0, 1, len = 1)
  if (is.na(size) || is.na(prob)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(size = size, prob = prob),
    pmf = function(x) {
      stats::dbinom(x, size = size, prob = prob)
    },
    cdf = function(x) {
      stats::pbinom(x, size = size, prob = prob)
    },
    quantile = function(p) {
      stats::qbinom(p, size = size, prob = prob)
    },
    realise = function(n) {
      stats::rbinom(n, size = size, prob = prob)
    },
    survival = function(x) {
      stats::pbinom(x, size = size, prob = prob, lower.tail = FALSE)
    },
    mean = size * prob,
    variance = size * prob * (1 - prob),
    skewness = (1 - 2 * prob) / sqrt(size * prob * (1 - prob)),
    kurtosis_exc = (1 - 6 * prob * (1 - prob)) / (size * prob * (1 - prob)),
    range = c(0, size),
    .vtype = "discrete",
    .name = "Binomial"
  )
}

#' Distribution Quantiles
#'
#' Access a distribution's quantiles.
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_unif(0, 4)
#' eval_quantile(d, at = 1:9 / 10)
#' enframe_quantile(d, at = 1:9 / 10)
#' @family distributional representations
#' @details When a quantile function does not exist, an algorithm is
#' deployed that calculates the left inverse of the CDF. This algorithm
#' works by progressively cutting the specified range in half, moving
#' into the left or right half depending on where the solution is.
#' The algorithm is not currently fast and is subject to improvement,
#' and is a simple idea that has been passed around on the internet
#' here and there. Tolerance is less than 1e-9, unless the maximum
#' number of iterations (200) is reached.
#'
#' The algorithm is not new, and is rather simple.
#' The algorithm works by progressively
#' cutting an initially wide range in half, moving into the left or right
#' half depending on where the solution is. I found the idea on Stack
#' Overflow somewhere, but unfortunately cannot find the location anymore.
#' @rdname quantile
#' @export
eval_quantile <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "quantile", at)
}

#' @rdname quantile
#' @export
enframe_quantile <- function(
    ..., at, arg_name = ".arg", fn_prefix = "quantile",
    sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_quantile
  )
}

# Validate density can be recovered from CDF
#
# Systematic testing uses the `test_distributions` object.
# To modify it, see `data-raw/test_distributions.R`.
# To update it, run `Rscript data-raw/test_distributions.R`.



test_that("Built-in density functions match cdf", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (is_intrinsic(d, "density")) {
        expect_false(is_intrinsic(d, "pmf"))
        rng <- range(d)
        p <- 1:99 / 100
        x <- eval_quantile(d, at = p)
        dens_fun <- representation_as_function(d, "density")
        increments <- vapply(
          seq_along(x[-1L]),
          function(t) {
            stats::integrate(dens_fun, lower = x[t], upper = x[t + 1])$value
          },
          FUN.VALUE = numeric(1L)
        )
        cdf_derived <- p[1] + append(0, cumsum(increments))
        expect_equal(cdf_derived, eval_cdf(d, at = x))
      }
    }
  }
})

#' Enframe a distributional representation
#'
#' This is the workhorse for the `enframe_` family of functions.
#' `enframe_general()` evaluates a specified distributional representation
#' for multiple distributions, and places the results in a data frame
#' or tibble.
#'
#' @inheritParams eval_cdf
#' @param ... Distributions to enframe; optionally, named.
#' @param eval_fn The `eval_` function for the desired distributional
#' representation, such as `eval_cdf` and `eval_density`.
#' @param fn_args A named list of arguments to pass to the `eval_fn` function,
#' besides the distribution and `at` argument.
#' @returns A data frame or tibble of the input argument (`at`), with the
#' evaluated distributional representation for each distribution in
#' `...` in its own column.
#' @details If only one distribution is specified in `...`, then the evaluation
#' column will be named that of `fn_prefix`.
#'
#' If more than one distribution
#' is specified in `...`, the evaluation columns will be named by the
#' prefix `fn_prefix` followed by the distribution names, with `sep` in between.
#'
#' Distributions are named first by their argument names, if given, or if not,
#' the input text. Names are then made unique using `vctrs::vec_as_names()`
#' with the "unique" names repair. "Unique" is chosen instead of "universal"
#' because names are anticipated to be syntactic with the `eval_fn` prefix;
#' "minimal" is not sufficient because it may result in columns having the
#' same names.
#' @noRd
enframe_general <- function(...,
                            at,
                            arg_name,
                            fn_prefix,
                            sep,
                            eval_fn,
                            fn_args = list()) {
  arg_name <- as.character(arg_name)
  fn_prefix <- as.character(fn_prefix)
  sep <- as.character(sep)
  checkmate::assert_numeric(at)
  checkmate::assert_character(arg_name, len = 1)
  checkmate::assert_character(fn_prefix, len = 1)
  checkmate::assert_character(sep, len = 1)
  checkmate::assert_function(eval_fn)
  checkmate::assert_list(fn_args)
  ellipsis <- rlang::quos(...)
  ellipsis <- rlang::quos_auto_name(ellipsis)
  distributions <- lapply(ellipsis, rlang::eval_tidy)
  checkmate::assert_list(distributions, types = "dst")
  is_distributions <- vapply(
    distributions, is_distribution,
    FUN.VALUE = logical(1L)
  )
  if (!all(is_distributions)) {
    stop(
      "`enframe_*()` functions only accept distributions. ",
      "Entries that are not distributions: ",
      paste(which(!is_distributions), collapse = ", ")
    )
  }

  n <- length(distributions)
  if (n == 0L) {
    stop("Need at least one distribution in the `enframe_*()` function.")
  }
  f <- list()
  for (i in seq_len(n)) {
    f[[i]] <- rlang::exec(
      eval_fn,
      distribution = distributions[[i]], at = at, !!!fn_args
    )
  }
  if (n == 1L) {
    eval_col_names <- fn_prefix
  } else {
    ellipsis_names <- rlang::names2(ellipsis)
    dist_names <- vctrs::vec_as_names(ellipsis_names, repair = "unique")
    eval_col_names <- paste0(fn_prefix, sep, dist_names)
  }
  names(f) <- eval_col_names
  arg <- list(at)
  names(arg) <- arg_name
  res <- as.data.frame(c(arg, f))
  convert_dataframe_to_tibble(res)
}

#' Gamma Distribution
#'
#' Makes a Gamma distribution.
#'
#' @param shape Shape parameter; single positive numeric.
#' @param rate Rate parameter; single positive numeric.
#' @returns A Gamma distribution.
#' @examples
#' dst_gamma(2, 1)
#' @export
dst_gamma <- function(shape, rate) {
  checkmate::assert_numeric(shape, 0, len = 1)
  checkmate::assert_numeric(rate, 0, len = 1)
  if (is.na(shape) || is.na(rate)) {
    return(dst_null())
  }
  d <- dst_pearson3(location = 0, scale = 1 / rate, shape = shape)
  attributes(d)[["name"]] <- "Gamma"
  parameters(d) <- list(shape = shape, rate = rate)
  d
}

#' Constructor Function for Distribution Objects
#'
#' Low level constructor function for creating distribution objects.
#' Consider the user-facing constructor function, `distribution()`.
#'
#' @param l List containing the components of a distribution object.
#' @param ... Attributes to add to the list.
#' @param class If making a subclass, specify its name here.
#' @returns The original list `l` with the specified attributes.
#' @noRd
new_distribution <- function(l, ..., class = character()) {
  checkmate::assert_list(l)
  checkmate::assert_character(class)
  structure(l, ..., class = append(class, "dst"))
}

#' Distribution Objects
#' @param object Object to be tested
#' @rdname distribution
#' @export
is_distribution <- function(object) inherits(object, "dst")

#' @rdname distribution
#' @export
is.distribution <- function(object) inherits(object, "dst")

#' Variable Type of a Distribution
#'
#' Retrieve the variable type of a distribution, such as
#' "continuous" or "discrete".
#'
#' @param distribution Distribution object.
#' @returns Single character with the variable type.
#' @examples
#' vtype(dst_beta(1, 2))
#' vtype(dst_bern(0.4))
#' vtype(distribution())
#' @export
vtype <- function(distribution) {
  checkmate::assert_class(distribution, "dst")
  attributes(distribution)[["vtype"]]
}

#' Degenerate Distribution
#'
#' A degenerate distribution assigns a 100% probability to one outcome.
#' @param location Outcome of the distribution; single positive numeric.
#' @returns A degenerate distribution
#' @examples
#' d <- dst_degenerate(5)
#' realise(d)
#' variance(d)
#' @export
dst_degenerate <- function(location) {
  checkmate::assert_numeric(location, len = 1)
  if (is.na(location)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(location = location),
    cdf = function(x) {
      as.numeric(x >= location)
    },
    quantile = function(p) {
      res <- rep(location, length(p))
      res[p < 0 | p > 1] <- NaN
      res[is.na(p)] <- NA_real_
      res
    },
    pmf = function(x) {
      as.numeric(x == location)
    },
    realise = function(n) {
      rep(location, n)
    },
    survival = function(x) {
      as.numeric(x < location)
    },
    mean = location,
    stdev = 0,
    variance = 0,
    skewness = NaN,
    kurtosis_exc = NaN,
    .name = "Degenerate",
    .vtype = "discrete"
  )
}

test_that("Degenerate Distribution quantities are appropriate", {
  d <- dst_degenerate(0.5)
  expect_equal(eval_pmf(d, -10:10), rep(0, 21))
  expect_equal(eval_pmf(d, 0.5), 1)
  expect_equal(eval_cdf(d, -10:10), c(rep(0, 11), rep(1, 10)))
  expect_equal(eval_cdf(d, 0.5), 1)
  expect_equal(eval_survival(d, c(0, 0.5, 1)), c(1, 0, 0))
  expect_equal(realise(d, 10), rep(0.5, 10))
  expect_equal(eval_quantile(d, -1:11 / 10), c(NaN, rep(0.5, 11), NaN))
  expect_equal(mean(d), 0.5)
  expect_equal(variance(d), 0)
  expect_equal(skewness(d), NaN)
  expect_equal(kurtosis_exc(d), NaN)
  expect_equal(range(d), c(0.5, 0.5))
})

#' Generalised Extreme Value Distribution
#'
#' Makes a Generalised Extreme Value (GEV) distribution, which is the
#' limiting distribution of the maximum.
#'
#' @param location Location parameter; single numeric.
#' @param scale Scale parameter; single positive numeric.
#' @param shape Shape parameter; single numeric.
#' This is also the extreme value index,
#' so that `shape > 0` is heavy tailed, and `shape < 0` is short-tailed.
#' @returns A GEV distribution.
#' @examples
#' # Short-tailed example
#' short <- dst_gev(0, 1, -1)
#' range(short)
#' mean(short)
#'
#' # Heavy-tailed example
#' heavy <- dst_gev(0, 1, 1)
#' range(heavy)
#' mean(heavy)
#'
#' # Light-tailed example (a Gumbel distribution)
#' light <- dst_gev(0, 1, 0)
#' range(light)
#' mean(light)
#' @export
dst_gev <- function(location, scale, shape) {
  checkmate::assert_numeric(location, len = 1)
  checkmate::assert_numeric(scale, 0, len = 1)
  checkmate::assert_numeric(shape, len = 1)
  if (is.na(location) || is.na(scale) || is.na(shape)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(location = location, scale = scale, shape = shape),
    cdf = function(x) {
      pgev(x, location = location, scale = scale, shape = shape)
    },
    quantile = function(p) {
      qgev(p, location = location, scale = scale, shape = shape)
    },
    density = function(x) {
      dgev(x, location = location, scale = scale, shape = shape)
    },
    mean = {
      if (shape >= 1) {
        Inf
      } else if (shape == 0) {
        location - scale * digamma(1)
      } else {
        location + (scale * (gamma(1 - shape) - 1)) / shape
      }
    },
    variance = {
      if (shape >= 0.5) {
        Inf
      } else if (shape == 0) {
        scale^2 * pi^2 / 6
      } else {
        scale^2 * (gamma(1 - 2 * shape) - gamma(1 - shape)^2) / shape^2
      }
    },
    range = c(
      gev_lower(location, scale, shape),
      gev_upper(location, scale, shape)
    ),
    .name = "Generalised Extreme Value",
    .vtype = "continuous"
  )
}

#' Empirical Distribution
#'
#' An empirical distribution is a non-parametric way to
#' estimate a distribution using data. By default,
#' it assigns equal probability to all observations
#' (this can be overridden with the `weights` argument).
#' Identical to [dst_finite()] with NA handling and with weights not needing
#' to add to 1.
#'
#' @param y <`data-masking`> Numeric vector representing the potential
#' outcomes of the distribution.
#' @param weights <`data-masking`> Numeric vector of weights corresponding to
#' to the outcomes `y`. These will be scaled so that they add up to 1.
#' @param data Optionally, a data frame to compute `y` and `weights` from.
#' `NULL` if data are not coming from a data frame (the default).
#' @param na_action_y,na_action_w What should be done with `NA` entries in
#' `y` and `w`eights? Character vector of length 1: one of `"fail"`,
#' `"null"` (default), or `"drop"`. See details.
#' @returns A finite distribution. If only one outcome, returns a degenerate
#' distribution. Returns a Null distribution if `NA` values are present
#' and `"null"` is specified as an NA action.
#' @details
#' `y` and `weights` are recycled to have the same length, but only
#' if one of them has length 1 (via `vctrs::vec_recycle_common()`).
#'
#' `na_action_y` and `na_action_w` specify the NA action for `y` and `weights`.
#' Options are, in order of precedence:
#'
#' - `"fail"`: Throw an error in the presence of `NA`.
#' - `"null"`: Return a Null distribution (`dst_null()`) in the presence
#'   of `NA`.
#' - `"drop"` (the default for `na_action_w`): Remove outcome-weight pairs
#'   having an `NA` value in the specified vector.
#'
#' @seealso [dst_finite()]
#' @examples
#' t <- -2:7
#' dst_empirical(t)
#'
#' # Using a data frame
#' df <- data.frame(time = c(NA, NA, t))
#' dst_empirical(time * 60, data = df)  # Null, since `NA` in `time`.
#'
#' # Drop NA `time` values.
#' dst_empirical(time * 60, data = df, na_action_y = "drop")
#'
#' # Weights explicit. Zero-weight outcomes ("-120") are gone.
#' df$w <- c(1, 1, 0:9)
#' dst_empirical(time * 60, w, data = df, na_action_y = "drop")
#'
#' # "Null" takes precedence over "drop".
#' df$w <- c(NA, NA, 0:9)
#' df$time[1] <- -3
#' df$time[12] <- NA
#' dst_empirical(time, w, data = df, na_action_w = "null", na_action_y = "drop")
#' dst_empirical(time, w, data = df, na_action_w = "drop", na_action_y = "null")
#' dst_empirical(time, w, data = df, na_action_w = "drop", na_action_y = "drop")
#' @export
dst_empirical <- function(y,
                          weights = 1,
                          data = NULL,
                          na_action_y = c("null", "drop", "fail"),
                          na_action_w = c("null", "drop", "fail")) {
  enquo_y <- rlang::enquo(y)
  enquo_w <- rlang::enquo(weights)
  y <- rlang::eval_tidy(enquo_y, data = data)
  w <- rlang::eval_tidy(enquo_w, data = data)
  yw <- vctrs::vec_recycle_common(y, w)
  y <- yw[[1]]
  w <- yw[[2]]
  na_action_y <- rlang::arg_match(na_action_y)
  na_action_w <- rlang::arg_match(na_action_w)
  na_w <- is.na(w)
  na_y <- is.na(y)
  has_na_w <- any(na_w)
  has_na_y <- any(na_y)
  # Fail first
  if (has_na_w && na_action_w == "fail") {
    stop(
      "Weights have NA values. You can either deal with these, or choose ",
      "an alternate option for `na_action_w`."
    )
  }
  if (has_na_y && na_action_y == "fail") {
    stop(
      "Outcomes have NA values. You can either deal with these, or choose ",
      "an alternate option for `na_action_y`."
    )
  }
  # Null next
  if (has_na_w && na_action_w == "null") {
    return(dst_null())
  }
  if (has_na_y && na_action_y == "null") {
    return(dst_null())
  }
  # Drop last
  if (na_action_w == "drop") {
    y <- y[!na_w]
    w <- w[!na_w]
  }
  if (na_action_y == "drop") {
    na_y <- is.na(y) # Because y has a new length.
    y <- y[!na_y]
    w <- w[!na_y]
  }
  steps <- aggregate_weights(y, w, sum_to_one = TRUE)
  dst_finite(outcomes = steps$y, probs = steps$weight)
}

#' F Distribution
#'
#' Makes an F distribution.
#'
#' @param df1,df2 Degrees of freedom of the numerator and denominator,
#' both single positive numerics.
#' @returns An F distribution.
#' @examples
#' dst_f(2, 3)
#' @export
dst_f <- function(df1, df2) {
  checkmate::assert_numeric(df1, 0, len = 1)
  checkmate::assert_numeric(df2, 0, len = 1)
  if (is.na(df1) || is.na(df2)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(df1 = df1, df2 = df2),
    density = function(x) {
      stats::df(x, df1 = df1, df2 = df2)
    },
    cdf = function(x) {
      stats::pf(x, df1 = df1, df2 = df2)
    },
    quantile = function(p) {
      stats::qf(p, df1 = df1, df2 = df2)
    },
    realise = function(n) {
      stats::rf(n, df1 = df1, df2 = df2)
    },
    survival = function(x) {
      stats::pf(x, df1 = df1, df2 = df2, lower.tail = FALSE)
    },
    mean = ifelse(df2 > 2, df2 / (df2 - 2), NaN),
    variance = ifelse(
      df2 > 4,
      2 * df2^2 * (df1 + df2 - 2) / (df1 * (df2 - 2)^2 * (df2 - 4)),
      NaN
    ),
    skewness = ifelse(
      df2 > 6,
      (2 * df1 + df2 - 2) * sqrt(8 * (df2 - 4)) /
        ((df2 - 6) * sqrt(df1 * (df1 + df2 - 2))),
      NaN
    ),
    kurtosis_exc = ifelse(
      df2 > 8,
      12 * (df1 * (5 * df2 - 22) * (df1 + df2 - 2) + (df2 - 4) * (df2 - 2)^2) /
        (df1 * (df2 - 6) * (df2 - 8) * (df1 + df2 - 2)),
      NaN
    ),
    range = c(0, Inf),
    .name = "F",
    .vtype = "continuous"
  )
}

#' Cumulative Distribution Function
#'
#' Access a distribution's cumulative distribution function (cdf).
#'
#' @param distribution,... A distribution, or possibly multiple
#' distributions in the case of `...`.
#' @param at Vector of values to evaluate the representation at.
#' @param arg_name For `enframe_`, name of the column containing
#' the function arguments. Length 1 character vector.
#' @param fn_prefix For `enframe_`, name of the function to
#' appear in the column(s). Length 1 character vector.
#' @param sep When `enframe`'ing more than one distribution, the
#' character that will be separating the `fn_name` and the distribution name.
#' Length 1 character vector.
#' @returns The evaluated representation in vector form (for `eval_`)
#' with length matching the length of `at`, and data frame
#' or tibble form (for `enframe_`) with number of rows matching the
#' length of `at`. The `at` input occupies the first column,
#' named `.arg` by default, or the specification in `arg_name`;
#' the evaluated representations for each distribution in `...`
#' go in the subsequent columns (one column per distribution). For a
#' single distribution, this column is named according to the
#' representation by default (cdf, survival, quantile, etc.),
#' or the value in `fn_prefix`. For multiple distributions, unnamed
#' distributions are auto-named, and columns are named
#' `<fn_prefix><sep><distribution_name>` (e.g., `cdf_distribution1`).
#' @family distributional representations
#' @examples
#' d1 <- dst_unif(0, 4)
#' d2 <- dst_pois(1.1)
#' eval_cdf(d1, at = 0:4)
#' enframe_cdf(d1, at = 0:4)
#' enframe_cdf(d1, d2, at = 0:4)
#' enframe_cdf(model1 = d1, model2 = d2, at = 0:4)
#' enframe_cdf(
#'   model1 = d1, model2 = d2, at = 0:4, arg_name = "value"
#' )
#' @rdname cdf
#' @export
eval_cdf <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "cdf", at)
}

#' @rdname cdf
#' @export
enframe_cdf <- function(..., at, arg_name = ".arg", fn_prefix = "cdf",
                        sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_cdf
  )
}

#' Survival Function
#'
#' Access a distribution's survival function.
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_unif(0, 4)
#' eval_survival(d, at = 0:4)
#' enframe_survival(d, at = 0:4)
#' @family distributional representations
#' @rdname survival
#' @export
eval_survival <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "survival", at)
}

#' @noRd
eval_survival_from_network <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  1 - eval_cdf(distribution, at = at)
}

#' @rdname survival
#' @export
enframe_survival <- function(..., at, arg_name = ".arg", fn_prefix = "survival",
                             sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_survival
  )
}

# Validates survival can be recovered from network
#
# Systematic testing uses the `test_distributions` object.
# To modify it, see `data-raw/test_distributions.R`.
# To update it, run `Rscript data-raw/test_distributions.R`.


test_that("Survival function calculated thru network matches known vals", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      p <- 0:100 / 100
      x <- eval_quantile(d, at = p)
      if (is_intrinsic(d, "survival")) {
        expect_equal(
          eval_survival_from_network(d, at = x),
          eval_survival(d, at = x)
        )
      }
    }
  }
})

#' Hazard Function
#'
#' Access a distribution's hazard function.
#'
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @examples
#' d <- dst_unif(0, 4)
#' eval_hazard(d, at = 0:4)
#' enframe_hazard(d, at = 0:4)
#' @family distributional representations
#' @rdname hazard
#' @export
eval_hazard <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "hazard", at)
}

#' @noRd
eval_hazard_from_network <- function(distribution, at) {
  if (vtype(distribution) != "continuous") {
    stop("Hazard function requires a continuous distribution.")
  }
  sf <- eval_survival(distribution, at)
  pdf <- eval_density(distribution, at)
  pdf / sf
}

#' @rdname hazard
#' @export
enframe_hazard <- function(..., at, arg_name = ".arg", fn_prefix = "hazard",
                           sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_hazard
  )
}

#' Bernoulli Distribution
#'
#' Makes a Bernoulli distribution, representing the outcome of a
#' single trial with a given success probability.
#'
#' @param prob Probability of success; single numeric between 0 and 1.
#' @returns A Bernoulli distribution.
#' @examples
#' dst_bern(0.3)
#' @export
dst_bern <- function(prob) {
  checkmate::assert_numeric(prob, 0, 1, len = 1)
  if (is.na(prob)) {
    return(dst_null())
  }
  d <- dst_binom(size = 1, prob = prob)
  parameters(d)$size <- NULL
  attr(d, "name") <- "Bernoulli"
  d
}

#' Range of Distribution
#'
#' Range returns a vector of length two, with the minimum and maximum
#' values of the (support of the) distribution.
#'
#' @param distribution Distribution to compute range from.
#' @param ... Not used; vestige of the `base::range()` S3 generic.
#' @details If there are no methods for the distribution's class,
#' the range is calculated
#' using `eval_quantile()` at 0 and at 1.
#' @returns Vector of length two, containing the minimum and maximum
#' values of a distribution.
#' @examples
#' a <- dst_gp(1, 0.5)
#' b <- dst_unif(0, 1)
#' c <- dst_norm(3, 4)
#' range(a)
#' range(b)
#' range(c)
#' @rdname range
#' @export
range.dst <- function(distribution, ...) {
  checkmate::assert_class(distribution, "dst")
  dots <- rlang::enexprs(...)
  dots[["na.rm"]] <- NULL
  if (length(dots) > 0) {
    stop(
      "`range()` is expecting no arguments in `...`. ",
      "Did you accidentally misspell 'distribution'?"
    )
  }
  eval_property(distribution, "range")
}

#' Moments of a Distribution
#'
#' Get common moment-related quantities of a
#' distribution: `mean`, `variance`, standard deviation (`stdev`),
#' `skewness`, and `kurtosis` or excess kurtosis (`kurtosis_exc`).
#' If these quantities are not supplied in the
#' distribution's definition, a numerical algorithm may be used.
#'
#' @param x,distribution Distribution to evaluate.
#' @param ... When calculating the mean via integration of the quantile
#' function, arguments passed to `stats::integrate()`.
#' @note Beware that if a quantity is being calculated numerically
#' for a non-continuous (e.g., discrete) distribution, the calculation
#' could be highly approximate. An upcoming version of distionary will
#' resolve this issue.
#' @details If there is no method associated with a subclass of
#' \code{x}, then moments are calculated using
#' \code{stats::integrate()} from the density function.
#'
#' @returns A single numeric.
#' @examples
#' a <- dst_gp(1, 0.5)
#' b <- dst_unif(0, 1)
#' c <- dst_norm(3, 4)
#' mean(a)
#' variance(b)
#' kurtosis(c)
#' kurtosis_exc(c)
#' @rdname moments
#' @export
mean.dst <- function(x, ...) {
  rlang::check_dots_empty()
  eval_property(x, "mean")
}

#' Uniform Distribution
#'
#' Makes a Uniform distribution.
#'
#' @param min,max Minimum and maximum of the distribution.
#' Single numerics.
#' @returns A Uniform distribution.
#' @examples
#' dst_unif(0, 1)
#' @export
dst_unif <- function(min, max) {
  checkmate::assert_numeric(min, len = 1)
  if (is.na(min)) {
    return(dst_null())
  }
  checkmate::assert_numeric(max, min, len = 1)
  if (is.na(max)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(min = min, max = max),
    density = function(x) {
      stats::dunif(x, min = min, max = max)
    },
    cdf = function(x) {
      stats::punif(x, min = min, max = max)
    },
    quantile = function(p) {
      stats::qunif(p, min = min, max = max)
    },
    realise = function(n) {
      stats::runif(n, min = min, max = max)
    },
    survival = function(x) {
      stats::punif(x, min = min, max = max, lower.tail = FALSE)
    },
    mean = (min + max) / 2,
    median = (min + max) / 2,
    variance = (min - max)^2 / 12,
    skewness = 0,
    kurtosis_exc = -6 / 5,
    range = c(min, max),
    .name = "Uniform",
    .vtype = "continuous"
  )
}

test_that("Return levels match some known values", {
  expect_equal(eval_return(dst_unif(0, 1), at = 1), 0)
  expect_equal(eval_return(dst_unif(0, 1), at = 2), 0.5)
  expect_equal(eval_return(dst_unif(0, 1), at = 10), 0.9)
  expect_equal(eval_return(dst_unif(0, 1), at = Inf), 1)
  expect_equal(
    suppressWarnings(eval_return(dst_unif(0, 1), at = c(-1, 0.5))),
    c(NaN, NaN)
  )
})

#' Evaluate Quantiles from a CDF
#'
#' Pulls together all the pieces needed to calculate the left inverse
#' of the CDF. Intended for internal use only. Calls the
#' `encapsulate_p()` function to figure out where to start looking for
#' solutions, then `directional_inverse()` to run the algorithm.
#'
#' @param distribution A distribution having access to a cdf.
#' @param at A vector of values for which to evaluate the quantile function.
#' @param tol,maxiter Tolerance (a small positive number) and maximum number
#' of iterations (at least 1); length 1 vectors.
#' @returns The `at`-quantiles of the distribution. Numeric vector
#' of values between 0 and 1.
#' @noRd
eval_quantile_from_network <- function(distribution,
                                       at,
                                       tol = 1e-9,
                                       maxiter = 200) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at, 0, 1)
  checkmate::assert_numeric(tol, 0, len = 1)
  checkmate::assert_integerish(maxiter, lower = 1, len = 1)
  n <- length(at)
  if (n == 0) {
    return(numeric(0L))
  }
  if (vtype(distribution) != "continuous") {
    stop(
      "The current quantile algorithm can suffer from low precision with ",
      "non-continuous distributions, so this functionality is disabled ",
      "for now."
    )
  }
  ord <- order(at)
  at <- at[ord]
  x <- at
  i_na <- which(is.na(at))
  i_zero <- which(at == 0)
  n_zero <- length(i_zero)
  i_positive <- which(at > 0 & at <= 1)
  n_positive <- length(i_positive)
  i_other <- setdiff(seq_len(n), c(i_na, i_zero, i_positive))
  x[i_other] <- NaN
  if (n_zero > 0) {
    r <- encapsulate_p(distribution, p = 0, direction = "right")
    if (is.infinite(r[1L])) {
      x[i_zero] <- r[1L]
    } else {
      x[i_zero] <- directional_inverse(
        distribution,
        p = 0, low = r[1L], high = r[2L], tol = tol,
        maxiter = maxiter, direction = "right"
      )
    }
  }
  r <- encapsulate_p(distribution, p = at[i_positive], direction = "left")
  low <- rep(r[1L], n)
  high <- r[2L]
  for (i in i_positive) {
    p <- at[i]
    if (isTRUE(p == at[i - 1L])) {
      x[i] <- low[i + 1L] <- x[i - 1L]
    } else {
      x[i] <- low[i + 1L] <- directional_inverse(
        distribution,
        p = p, low = low[i], high = r[2L], tol = tol,
        maxiter = maxiter, direction = "left"
      )
    }
  }
  x[ord] <- x
  x
}

#' Find a range of possible outcomes
#'
#' In order to run the directional inverse algorithm, we need to know
#' approximately where the solution lies. This function
#' finds a range of possible outcomes where the cdf evaluates to values
#' (probabilities) contain the vector `p`, and therefore should come
#' before the inversion algorithm begins.
#'
#' @param p Vector of values between 0 and 1 (inclusive).
#' @param direction One of `"left"` for calculating left-inverse, or
#' `"right"` for calculating right-inverse.
#' @note If 0 or 1 are included in the vector `p`, one of the endpoints might
#' be infinite.
#' @returns A range of values containing the solutions to the left
#' inverse of the CDF at `p`.
#' @noRd
#' @inheritParams eval_quantile_from_network
encapsulate_p <- function(distribution, p, direction) {
  if (length(p) == 0) {
    return(c(NA, NA))
  }
  p_min <- min(p)
  p_max <- max(p)
  if (direction == "left") {
    cdf_gt <- `>=`
    cdf_lt <- `<`
    survival_gt <- `>`
  } else if (direction == "right") {
    cdf_gt <- `>`
    cdf_lt <- `<=`
    survival_gt <- `>=`
  } else {
    stop(
      "`direction` must be one of 'left' or 'right'. Received '",
      direction, "'."
    )
  }
  left <- -1
  right <- 1
  cdf_p <- eval_cdf(distribution, at = p)
  cdf_left <- eval_cdf(distribution, at = left)
  while (cdf_gt(cdf_left, p_min)) {
    left <- 2 * left
    cdf_left <- eval_cdf(distribution, at = left)
  }
  if (p_max >= 0.9 && !is.null(distribution$survival)) {
    survival_right <- eval_survival(distribution, at = right)
    while (survival_gt(survival_right, 1 - p_max)) {
      right <- 2 * right
      survival_right <- eval_survival(distribution, at = right)
    }
  } else {
    cdf_right <- eval_cdf(distribution, at = right)
    while (cdf_lt(cdf_right, p_max)) {
      right <- 2 * right
      cdf_right <- eval_cdf(distribution, at = right)
    }
  }
  if (p_min > 0 && is.infinite(left)) {
    left <- -.Machine$double.xmax
  }
  if (p_max < 1 && is.infinite(right)) {
    right <- .Machine$double.xmax
  }
  c(left, right)
}


#' Algorithm to Compute a Directional Inverse
#'
#' Calculates the smallest value for which a function `f`
#' evaluates to be greater than or equal to `y` -- that is,
#' the left inverse of `f` at `y`.
#' @param p Single value for which to calculate the left inverse.
#' @param low,high Single numeric values forming a range
#' within which to search for the solution.
#' @param tol,maxiter Tolerance (a small positive number) and maximum number
#' of iterations
#' @details This algorithm works by progressively
#' cutting the specified range in half, moving into the left or right
#' half depending on where the solution is.
#' @returns The left inverse of the CDF evaluated at `p`.
#' @noRd
#' @inheritParams encapsulate_p
directional_inverse <- function(distribution, p, low, high, tol, maxiter,
                                direction) {
  stopifnot(low <= high)
  if (is.na(p)) {
    return(p)
  }
  if (direction == "left") {
    ineq <- `<=`
  } else if (direction == "right") {
    ineq <- `<`
  } else {
    stop(
      "`direction` must be one of 'left' or 'right'. Received '",
      direction, "'."
    )
  }
  max_tol <- tol
  w <- .Machine$double.xmax
  i <- 0L
  slope <- 1
  mid <- (high + low) / 2
  while (w > tol && i <= maxiter) {
    i <- i + 1L
    cdf_low <- eval_cdf(distribution, at = low)
    cdf_mid <- eval_cdf(distribution, at = mid)
    cdf_high <- eval_cdf(distribution, at = high)
    slope_left <- (cdf_mid - cdf_low) / w * 2
    slope_right <- (cdf_high - cdf_mid) / w * 2
    slope <- max(slope, min(slope_left, slope_right, na.rm = TRUE),
      na.rm = TRUE
    )
    tol <- min(max_tol / slope, tol, na.rm = TRUE)
    if (ineq(p, cdf_mid)) {
      high <- mid
    } else {
      low <- mid
    }
    if (low == high) {
      return(low)
    }
    w <- high - low
    mid <- (high + low) / 2
  }
  # cat("w:", w, "tol:", tol, "i:", i, "slope:", slope, "\n")
  if (i == maxiter && w > tol) {
    warning(
      "Maximum number of iterations reached before ",
      "tolerance was achieved."
    )
  }
  mid
}

#' Return Level Function
#'
#' Compute return levels (quantiles) from a distribution by inputting
#' return periods. The return periods correspond to events that are
#' _exceedances_ of a quantile, not non-exceedances.
#'
#' @param at Vector of return periods >=1.
#' @inheritParams eval_cdf
#' @inherit eval_cdf return
#' @details This function is simply the quantile
#' function evaluated at `1 - 1 / at`.
#' @examples
#' d <- dst_gp(24, 0.3)
#' eval_return(d, at = c(2, 25, 100, 200))
#' @family distributional representations
#' @rdname return
#' @export
eval_return <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_property(distribution, "return", at)
}

#' @noRd
eval_return_from_network <- function(distribution, at) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_numeric(at)
  eval_quantile(distribution, at = 1 - 1 / at)
}

#' @rdname return
#' @export
enframe_return <- function(..., at, arg_name = ".arg", fn_prefix = "return",
                           sep = "_") {
  enframe_general(
    ...,
    at = at, arg_name = arg_name, fn_prefix = fn_prefix,
    sep = sep, eval_fn = eval_return
  )
}

#' Geometric Distribution
#'
#' Makes a Geometric distribution, corresponding to the number of failures
#' in a sequence of independent trials before observing a success.
#'
#' @param prob Probability of success in each trial; single numeric
#' between 0 and 1.
#' @returns A Geometric distribution.
#' @examples
#' d <- dst_geom(0.4)
#'
#' # This version of the Geometric distribution does not count the success.
#' range(d)
#' @export
dst_geom <- function(prob) {
  checkmate::assert_numeric(prob, 0, 1, len = 1)
  if (is.na(prob)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(prob = prob),
    pmf = function(x) {
      stats::dgeom(x, prob = prob)
    },
    cdf = function(x) {
      stats::pgeom(x, prob = prob)
    },
    quantile = function(p) {
      stats::qgeom(p, prob = prob)
    },
    realise = function(n) {
      stats::rgeom(n, prob = prob)
    },
    survival = function(x) {
      stats::pgeom(x, prob = prob, lower.tail = FALSE)
    },
    mean = (1 - prob) / prob,
    variance = (1 - prob) / prob^2,
    skewness = ifelse(prob < 1, (2 - prob) / sqrt(1 - prob), NaN),
    kurtosis_exc = ifelse(prob < 1, 6 + prob^2 / (1 - prob), NaN),
    range = c(0, Inf),
    .name = "Geometric",
    .vtype = "discrete"
  )
}

#' Evaluate a distribution
#'
#' Evaluate a distribution property. The distribution itself
#' is first searched for the property, and if it can't
#' be found, will attempt to calculate the property
#' from other entries.
#'
#' @param distribution Distribution object.
#' @param entry Name of the property, such as "cdf" or
#' "mean". Length 1 character vector.
#' @param ... If the property is a function, arguments to the
#' function go here. Need not be named; inserted in the order they
#' appear.
#' @returns The distribution's property, evaluated. If cannot be
#' evaluated, returns `NULL`.
#' @examples
#' d <- distribution(
#'   cdf = function(x) {
#'     (x > 0) * pmin(x^2, 1)
#'   },
#'   g = 9.81,
#'   .vtype = "continuous"
#' )
#' eval_property(d, "g")
#' eval_property(d, "quantile", 1:9 / 10)
#' eval_property(d, "mean")
#' eval_property(d, "realise", 10)
#' eval_property(d, "foofy")
#' eval_property(d, "foofy", 1:10)
#' @export
eval_property <- function(distribution, entry, ...) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_character(entry, len = 1)
  repres <- distribution[[entry]]
  if (is.null(repres)) {
    if (entry == "realize") {
      return(eval_property(distribution, "realise", ...))
    }
    eval_from_network <- paste0("eval_", entry, "_from_network")
    available <- exists(eval_from_network)
    if (!available) {
      return(NULL)
    } else {
      return(rlang::exec(eval_from_network, distribution, ...))
    }
  } else {
    if (is.function(repres)) {
      return(repres(...))
    } else {
      rlang::check_dots_empty()
      return(repres)
    }
  }
}

#' Generalised Pareto Distribution
#'
#' Makes a Generalized Pareto (GP) distribution, corresponding to the
#' limiting distribution of excesses over a threshold.
#' @param scale Scale parameter; single positive numeric.
#' @param shape Shape parameter; single positive numeric.
#' This is also the extreme value index, so that `shape > 0` is heavy
#' tailed, and `shape < 0` is short-tailed.
#' @return A Generalised Pareto Distribution.
#' @examples
#' # Short-tailed example
#' short <- dst_gp(1, -1)
#' range(short)
#' mean(short)
#'
#' # Heavy-tailed example
#' heavy <- dst_gp(1, 1)
#' range(heavy)
#' mean(heavy)
#'
#' # Light-tailed example (a Gumbel distribution)
#' light <- dst_gp(1, 0)
#' range(light)
#' mean(light)
#' @export
dst_gp <- function(scale, shape) {
  checkmate::assert_numeric(scale, 0, len = 1)
  checkmate::assert_numeric(shape, len = 1)
  if (is.na(scale) || is.na(shape)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(scale = scale, shape = shape),
    cdf = function(x) {
      pgp(x, scale = scale, shape = shape)
    },
    survival = function(x) {
      pgp(x, scale = scale, shape = shape, lower.tail = FALSE)
    },
    quantile = function(p) {
      qgp(p, scale = scale, shape = shape)
    },
    density = function(x) {
      dgp(x, scale = scale, shape = shape)
    },
    mean = ifelse(shape < 1, scale / (1 - shape), Inf),
    variance = ifelse(
      shape < 1 / 2,
      scale^2 / (1 - shape)^2 / (1 - 2 * shape),
      Inf
    ),
    skewness = ifelse(
      shape < 1 / 3,
      2 * (1 + shape) * sqrt(1 - 2 * shape) /
        (1 - 3 * shape),
      Inf
    ),
    kurtosis_exc = ifelse(
      shape < 1 / 4,
      3 * (1 - 2 * shape) * (2 * shape^2 + shape + 3) /
        ((1 - 3 * shape) * (1 - 4 * shape)) - 3,
      Inf
    ),
    range = c(0, gp_upper(scale = scale, shape = shape)),
    .name = "Generalised Pareto",
    .vtype = "continuous"
  )
}

#' Beta Distribution
#'
#' Makes a Beta distribution.
#' @param shape1,shape2 Shape parameters of the distribution;
#' single positive numerics.
#' @returns A Beta distribution.
#' @examples
#' dst_beta(2, 3)
#' @export
dst_beta <- function(shape1, shape2) {
  checkmate::assert_numeric(shape1, 0, len = 1)
  checkmate::assert_numeric(shape2, 0, len = 1)
  if (is.na(shape1) || is.na(shape2)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(shape1 = shape1, shape2 = shape2),
    density = function(x) {
      stats::dbeta(x, shape1, shape2)
    },
    cdf = function(x) {
      stats::pbeta(x, shape1, shape2)
    },
    quantile = function(p) {
      stats::qbeta(p, shape1, shape2)
    },
    realise = function(n) {
      stats::rbeta(n, shape1, shape2)
    },
    survival = function(x) {
      stats::pbeta(x, shape1, shape2, lower.tail = FALSE)
    },
    mean = shape1 / (shape1 + shape2),
    variance = shape1 * shape2 / (shape1 + shape2)^2 /
      (shape1 + shape2 + 1),
    skewness = 2 * (shape2 - shape1) * sqrt(shape1 + shape2 + 1) /
      (shape1 + shape2 + 2) / sqrt(shape1 * shape2),
    range = c(0, 1),
    .vtype = "continuous",
    .name = "Beta"
  )
}

#' Poisson Distribution
#'
#' Makes a Poisson distribution.
#' @param lambda Mean of the Poisson distribution; single positive numeric.
#' @returns A Poisson distribution.
#' @examples
#' dst_pois(1)
#' @export
dst_pois <- function(lambda) {
  checkmate::assert_numeric(lambda, 0, len = 1)
  if (is.na(lambda)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(lambda = lambda),
    pmf = function(x) {
      suppressWarnings(stats::dpois(x, lambda = lambda))
    },
    cdf = function(x) {
      stats::ppois(x, lambda = lambda)
    },
    quantile = function(p) {
      stats::qpois(p, lambda = lambda)
    },
    realise = function(n) {
      stats::rpois(n, lambda = lambda)
    },
    survival = function(x) {
      stats::ppois(x, lambda = lambda, lower.tail = FALSE)
    },
    mean = lambda,
    variance = lambda,
    skewness = lambda^(-0.5),
    kurtosis_exc = 1 / lambda,
    range = c(0, Inf),
    .name = "Poisson",
    .vtype = "discrete"
  )
}

#' Log Pearson Type III distribution
#'
#' Makes a Log Pearson Type III distribution, which is the
#' distribution of the exponential of a random variable following
#' a Pearson Type III distribution.
#'
#' @param meanlog Mean of the log of the random variable; single numeric.
#' @param sdlog Standard deviation of the log of the random variable;
#' single positive numeric.
#' @param skew Skewness of the log of the random variable;
#' single numeric.
#' @returns A Log Pearson Type III distribution.
#' @examples
#' dst_lp3(0, 1, 1)
#' @export
dst_lp3 <- function(meanlog, sdlog, skew) {
  checkmate::assert_numeric(meanlog, len = 1)
  checkmate::assert_numeric(sdlog, 0, len = 1)
  checkmate::assert_numeric(skew, len = 1)
  if (is.na(meanlog) || is.na(sdlog) || is.na(skew)) {
    return(dst_null())
  }
  # skewness = 2 / sqrt(shape), shape = 4 / skewness^2
  shape <- 4 / skew^2
  # sd = sqrt(shape) * scale
  scale <- sdlog / sqrt(shape)
  # mean = scale * shape
  shift <- meanlog - scale * shape
  distribution(
    .parameters = list(meanlog = meanlog, sdlog = sdlog, skew = skew),
    cdf = function(x) stats::pgamma(
      log(pmax(0, x)) - shift,
      shape = shape, scale = scale
    ),
    survival = function(x) stats::pgamma(
      log(pmax(0, x)) - shift,
      shape = shape, scale = scale, lower.tail = FALSE
    ),
    density = function(x) {
      res <- stats::dgamma(
        log(pmax(0, x)) - shift,
        shape = shape, scale = scale
      ) / x
      res[x == 0] <- 0
      res
    },
    quantile = function(p) exp(
      stats::qgamma(p, shape = shape, scale = scale) + shift
    ),
    realise = function(n) exp(
      stats::rgamma(n, shape = shape, scale = scale) + shift
    ),
    .name = "Log Pearson Type III",
    .vtype = "continuous"
  )
}

#' Negative binomial Distribution
#'
#' Makes a Negative Binomial distribution, corresponding to the number
#' of failures in a sequence of independent trials until a given number
#' of successes are observed.
#'
#' @param prob Probability of a successful trial; single numeric
#' between 0 and 1.
#' @param size Number of successful trials; single positive numeric.
#' @returns A Negative Binomial distribution.
#' @examples
#' d <- dst_nbinom(10, 0.5)
#'
#' # This version of the Negative Binomial distribution does not count
#' # the successes.
#' range(d)
#' @export
dst_nbinom <- function(size, prob) {
  checkmate::assert_numeric(size, 0, len = 1)
  checkmate::assert_numeric(prob, 0, 1, len = 1)
  if (is.na(size) || is.na(prob)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(size = size, prob = prob),
    pmf = function(x) {
      stats::dnbinom(x, size = size, prob = prob)
    },
    cdf = function(x) {
      stats::pnbinom(x, size = size, prob = prob)
    },
    quantile = function(p) {
      stats::qnbinom(p, size = size, prob = prob)
    },
    realise = function(n) {
      stats::rnbinom(n, size = size, prob = prob)
    },
    survival = function(x) {
      stats::pnbinom(x, size = size, prob = prob, lower.tail = FALSE)
    },
    mean = (1 - prob) * size / prob,
    variance = (1 - prob) * size / prob^2,
    skewness = (2 - prob) / sqrt((1 - prob) * size),
    kurtosis_exc = 6 / size + prob^2 / ((1 - prob) * size),
    range = c(0, Inf),
    .name = "Negative Binomial",
    .vtype = "discrete"
  )
}

#' Intrinsically defined distribution properties
#'
#' Get the names of intrinsically defined distribution properties
#' (`intrinsics()`), or check if a property is intrinsically defined
#' (`is_intrinsic()`).
#' @param distribution The distribution to check.
#' @param property Name of the property to check; character vector.
#' @returns For `intrinsics()`, a character vector of property names
#' that have been intrinsically defined in the distribution.
#'
#' For `is_intrinsic()`, a logical vector of length equal to the length
#' of `property` indicating whether each property in `property`
#' is intrinsically defined in the distribution
#' (`TRUE`) or not (`FALSE`). If `property` is NA, the output is `NA`.
#' @noRd
intrinsics <- function(distribution) {
  checkmate::assert_class(distribution, "dst")
  names(distribution)
}

#' @noRd
is_intrinsic <- function(distribution, property) {
  checkmate::assert_class(distribution, "dst")
  property <- as.character(property)
  checkmate::assert_character(property)
  res <- property %in% intrinsics(distribution)
  res[is.na(property)] <- NA
  res
}

test_that("Checking for intrinsically defined properties works", {
  d <- dst_norm(0, 1)
  int <- intrinsics(d)
  # Works as intended.
  expect_true(is.character(intrinsics(d)))
  expect_true(length(intrinsics(d)) > 1)
  expect_false(any(is.na(intrinsics(d))))
  expect_equal(is_intrinsic(d, int), rep(TRUE, length(int)))
  expect_equal(is_intrinsic(d, letters), rep(FALSE, length(letters)))
  expect_true(is_intrinsic(d, "cdf"))
  expect_false(is_intrinsic(d, "hazard"))
  # NA and 0-length inputs; non-distributions
  expect_error(intrinsics(unclass(d)))
  expect_error(is_intrinsic(unclass(d), "cdf"))
  expect_equal(
    is_intrinsic(d, append(int, NA)),
    append(rep(TRUE, length(int)), NA)
  )
  expect_equal(is_intrinsic(d, character()), logical())
})

#' Get a distribution's representation as a function
#'
#' @param distribution Distribution to extract a representation from.
#' @param representation Character, such as `"cdf"`. In general, a suffix
#' to an `eval_` function. Vector of length 1.
#' @returns A function of the representation.
#' @noRd
representation_as_function <- function(distribution, representation) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_character(representation, len = 1)
  f <- distribution[[representation]]
  if (is.null(f)) {
    f <- function(x) {
      eval_property(distribution, representation, x)
    }
  }
  f
}

#' Plot a Distribution
#'
#' Plot a distribution's representation.
#'
#' @param x Distribution object
#' @param what Name of the representation to plot.
#' @param ... Other arguments to pass to the `graphics::curve` function,
#' or `graphics::plot` in the case of the PMF.
#' @returns This function is run for its graphics byproduct, and therefore
#' returns the original distribution, invisibly.
#' @examples
#' d <- dst_norm(0, 1)
#' plot(d, from = -4, to = 4)
#' plot(d, "cdf", n = 1000)
#' plot(d, "survival")
#' plot(d, "quantile")
#' plot(d, "hazard")
#' plot(d, "chf")
#'
#' p <- dst_pois(4)
#' plot(p)
#' @export
plot.dst <- function(
  x,
  what = c("density", "pmf", "cdf", "survival", "quantile", "hazard", "chf"),
  ...
) {
  ellipsis <- rlang::list2(...)
  specified <- length(what) == 1
  fname <- match.arg(what)
  if (specified) {
    if (fname == "density" && vtype(x) != "continuous") {
      stop("Density function does not exist.")
    }
    if (fname == "pmf" && vtype(x) != "discrete") {
      warning("Probability mass function does not exist.")
    }
  } else {
    if (fname == "density" && vtype(x) == "discrete") {
      if (pretty_name(x) == "Finite") {
        fname <- "cdf"
      } else {
        fname <- "pmf"
      }
    }
  }
  if (is.null(ellipsis[["ylab"]])) {
    ellipsis[["ylab"]] <- fname
  }
  if (fname == "quantile") {
    if (is.null(ellipsis[["from"]])) {
      ellipsis[["from"]] <- 0
    }
    if (is.null(ellipsis[["to"]])) {
      ellipsis[["to"]] <- 1
    }
    if (is.null(ellipsis[["xlab"]])) {
      ellipsis[["xlab"]] <- "Probability"
    }
    f <- representation_as_function(x, "quantile")
    ellipsis[["expr"]] <- as.name("f")
    do.call(graphics::curve, args = ellipsis)
    return(invisible(x))
  }
  if (is.null(ellipsis[["from"]])) {
    q0 <- range(x)[1L]
    if (q0 == -Inf) {
      ellipsis[["from"]] <- eval_quantile(x, at = 0.001)
    } else {
      ellipsis[["from"]] <- q0
    }
  }
  if (is.null(ellipsis[["to"]])) {
    q1 <- range(x)[2L]
    if (q1 == Inf) {
      ellipsis[["to"]] <- eval_quantile(x, at = 0.999)
    } else {
      ellipsis[["to"]] <- q1
    }
  }
  if (is.null(ellipsis[["xlab"]])) {
    ellipsis[["xlab"]] <- "y"
  }
  if (fname == "pmf") {
    if (!pretty_name(x) %in% c(
      "Bernoulli", "Binomial", "Poisson", "Geometric", "Negative Binomial",
      "Hypergeometric", "Degenerate", "Finite"
    )) {
      warning(
        "This version of distionary assumes the distribution takes on ",
        "integer values. If this is not the case, then the displayed pmf ",
        "may not be representative."
      )
    }
    xvals <- ellipsis[["from"]]:ellipsis[["to"]]
    if (pretty_name(x) == "Finite") {
      outcomes <- parameters(x)[["outcomes"]]
      xvals <- outcomes[
        outcomes >= ellipsis[["from"]] &
          outcomes <= ellipsis[["to"]]
      ]
    }
    ellipsis[["from"]] <- NULL
    ellipsis[["to"]] <- NULL
    ellipsis[["x"]] <- xvals
    ellipsis[["y"]] <- eval_pmf(x, at = xvals)
    do.call("plot", args = ellipsis)
    return(invisible(x))
  }
  f <- representation_as_function(x, fname)
  ellipsis[["expr"]] <- as.name("f")
  do.call("curve", args = ellipsis)
  invisible(x)
}

#' Representations of the Generalized Extreme Value Distribution
#'
#' @param x,q Vector of quantiles.
#' @param p Vector of probabilities.
#' @param location Location parameter; numeric vector.
#' @param scale Scale parameter; positive numeric vector.
#' @param shape Shape parameter; numeric vector.
#' This is also the extreme value index,
#' so that `shape > 0` is heavy tailed, and `shape < 0` is short-tailed.
#' @returns Vector of evaluated GEV distribution, with length
#' equal to the recycled lengths of `q`/`x`/`p`, `location`, `scale`, and
#' `shape`.
#' @examples
#' pgev(1:10, 0, 1, 1)
#' dgev(1:10, 1:10, 2, 0)
#' qgev(1:9 / 10, 2, 10, -2)
#' @rdname gev_raw
#' @export
pgev <- function(q, location, scale, shape) {
  checkmate::assert_numeric(q)
  checkmate::assert_numeric(location)
  checkmate::assert_numeric(scale, 0)
  checkmate::assert_numeric(shape)
  l <- vctrs::vec_recycle_common(q, location, scale, shape)
  q <- l[[1]]
  location <- l[[2]]
  scale <- l[[3]]
  shape <- l[[4]]
  lo <- gev_lower(location, scale, shape)
  hi <- gev_upper(location, scale, shape)
  t <- gev_t_function(q, location = location, scale = scale, shape = shape)
  res <- exp(-t)
  res[q <= lo] <- 0
  res[q >= hi] <- 1
  res
}

#' @rdname gev_raw
#' @export
qgev <- function(p, location, scale, shape) {
  checkmate::assert_numeric(p)
  checkmate::assert_numeric(location)
  checkmate::assert_numeric(scale, 0)
  checkmate::assert_numeric(shape)
  l <- vctrs::vec_recycle_common(p, location, scale, shape)
  p <- l[[1]]
  location <- l[[2]]
  scale <- l[[3]]
  shape <- l[[4]]
  neglogp <- -log(p)
  res <- ifelse(
    p >= 0 & p <= 1,
    ifelse(shape == 0, -log(neglogp), (neglogp^(-shape) - 1) / shape),
    NaN
  )
  location + scale * res
}

#' @rdname gev_raw
#' @export
dgev <- function(x, location, scale, shape) {
  checkmate::assert_numeric(x)
  checkmate::assert_numeric(location)
  checkmate::assert_numeric(scale, 0)
  checkmate::assert_numeric(shape)
  l <- vctrs::vec_recycle_common(x, location, scale, shape)
  x <- l[[1]]
  location <- l[[2]]
  scale <- l[[3]]
  shape <- l[[4]]
  lo <- gev_lower(location, scale, shape)
  hi <- gev_upper(location, scale, shape)
  t <- gev_t_function(x, location = location, scale = scale, shape = shape)
  res <- t^(shape + 1) / scale * exp(-t)
  ifelse(x <= lo | x > hi, 0, res)
  # res[x <= lo] <- 0
  # res[x > hi] <- 0
  # res
}

#' 't()' function for calculating GEV quantities
#'
#' Decides whether to use `exp()` when the shape parameter is 0,
#' or the non-limiting form otherwise. Also useful for the GP.
#'
#' @param x Argument of the function; vectorized.
#' @returns A vector of the t function evaluated.
#' @inheritParams pgev
#' @seealso See the Wikipedia entry for the GEV distribution,
#' https://en.wikipedia.org/wiki/Generalized_extreme_value_distribution
#' @note The shape parameter is not vectorized. This function is only
#' intended to be used when location, scale, and shape are scalars.
#' @noRd
gev_t_function <- function(x, location, scale, shape) {
  l <- vctrs::vec_recycle_common(x, location, scale, shape)
  x <- l[[1]]
  location <- l[[2]]
  scale <- l[[3]]
  shape <- l[[4]]
  z <- (x - location) / scale
  # res <- numeric(length(x))
  # i_zero <- shape == 0
  # i_non0 <- shape != 0
  ifelse(shape == 0, exp(-z), (1 + shape * z)^(-1 / shape))
}

#' Range of a GEV distribution
#'
#' @inheritParams pgev
#' @returns A vector of mimima for `gev_lower()`, or maxima
#' for `gev_upper()`, corresponding to the input parameters.
#' @noRd
gev_lower <- function(location, scale, shape) {
  l <- vctrs::vec_recycle_common(location, scale, shape)
  location <- l[[1]]
  scale <- l[[2]]
  shape <- l[[3]]
  ifelse(shape > 0, location - scale / shape, -Inf)
  # res <- shape # For NA propagation
  # i_calc <- shape > 0
  # res[i_calc] <- location[i_calc] - scale[i_calc] / shape[i_calc]
  # res[shape <= 0] <- -Inf
  # res
}

#' @inheritParams gev_lower
#' @noRd
gev_upper <- function(location, scale, shape) {
  l <- vctrs::vec_recycle_common(location, scale, shape)
  location <- l[[1]]
  scale <- l[[2]]
  shape <- l[[3]]
  ifelse(shape < 0, location - scale / shape, Inf)
  # res <- shape # For NA propagation
  # i_calc <- shape < 0
  # res[i_calc] <- location[i_calc] - scale[i_calc] / shape[i_calc]
  # res[shape >= 0] <- Inf
  # res
}

test_that("GEV negative scale not allowed.", {
  expect_error(dgev(1:10, 3, -4, 1))
  expect_error(pgev(1:10, 3, -4, 1))
  expect_error(qgev(1:10 / 10, 3, -4, 1))
})

test_that("cdf and pdf of GEV align via numerical derivative.", {
  d <- list(
    list(location = 0.2, scale = 1.2, shape = 0.8),
    list(location = 0.2, scale = 1.2, shape = 0),
    list(location = 0.2, scale = 1.2, shape = -0.8)
  )
  x <- -5:5
  eps <- 1e-6
  for (i in seq_along(d)) {
    pdf <- rlang::exec(dgev, x, !!!d[[i]])
    cdf1 <- rlang::exec(pgev, x - eps, !!!d[[i]])
    cdf2 <- rlang::exec(pgev, x, !!!d[[i]])
    pdf_num <- (cdf2 - cdf1) / eps
    expect_equal(pdf, pdf_num, tolerance = 1e-6)
  }
})

test_that("cdf and qf of GEV align.", {
  d <- list(
    list(location = 0.2, scale = 1.2, shape = 0.8),
    list(location = 0.2, scale = 1.2, shape = 0),
    list(location = 0.2, scale = 1.2, shape = -0.8)
  )
  p <- 1:9 / 10
  for (i in seq_along(d)) {
    qf <- rlang::exec(qgev, p, !!!d[[i]])
    cdf <- rlang::exec(pgev, qf, !!!d[[i]])
    expect_equal(cdf, p)
  }
})

test_that("quantile function of GEV is valid, validating the distribution.", {
  d <- list(
    list(location = 0.2, scale = 1.2, shape = 0.8),
    list(location = 0.2, scale = 1.2, shape = 0),
    list(location = 0.2, scale = 1.2, shape = -0.8)
  )
  p <- 0:100 / 100
  for (i in seq_along(d)) {
    lo <- rlang::exec(distionary:::gev_lower, !!!d[[i]])
    hi <- rlang::exec(distionary:::gev_upper, !!!d[[i]])
    r <- c(lo, hi)
    qf <- rlang::exec(qgev, p, !!!d[[i]])
    expect_true(all(diff(qf) > 0))
    expect_equal(min(qf), lo)
    expect_equal(max(qf), hi)
  }
})

test_that("vectorisation of p/d/q GEV functions works.", {
  y <- 1:10
  x <- 1:10
  x[4] <- NA_real_
  x[2] <- NaN
  p <- x / 11
  v <- y / 11
  # Mismatched lengths
  expect_error(pgev(x, 0:1, 1, 1))
  expect_error(pgev(x, 0, 1:2, 1))
  expect_error(pgev(x, 0, 1, 0:1))
  expect_error(dgev(x, 0:1, 1, 1))
  expect_error(dgev(x, 0, 1:2, 1))
  expect_error(dgev(x, 0, 1, 0:1))
  expect_error(qgev(p, 0:1, 1, 1))
  expect_error(qgev(p, 0, 1:2, 1))
  expect_error(qgev(p, 0, 1, 0:1))
  expect_error(distionary:::gev_lower(0:10, 1:2, 1))
  expect_error(distionary:::gev_lower(1, 1:2, 1:4))
  expect_error(distionary:::gev_lower(0:10, 2, 1:3))
  expect_error(distionary:::gev_upper(0:10, 1:2, 1))
  expect_error(distionary:::gev_upper(1, 1:2, 1:4))
  expect_error(distionary:::gev_upper(0:10, 2, 1:3))
  # Lengths input correctly; should be length 10.
  expect_length(pgev(x, 0, 1, 1), 10)
  expect_length(dgev(x, 0, 1, 1), 10)
  expect_length(qgev(p, 0, 1, 1), 10)
  expect_length(pgev(x, 0, 1:10, 1), 10)
  expect_length(dgev(x, 0, 1:10, 1), 10)
  expect_length(qgev(p, 0, 1:10, 1), 10)
  # NA and NaN gets projected forward. NaN may convert to NA.
  # --> main argument
  expect_true(is.na(pgev(x, 0, 1, 1)[4]))
  expect_true(is.na(dgev(x, 0, 1, 1)[4]))
  expect_true(is.na(qgev(p, 0, 1, 1)[4]))
  expect_true(is.na(pgev(x, 0, 1, 1)[2]))
  expect_true(is.na(dgev(x, 0, 1, 1)[2]))
  expect_true(is.na(qgev(p, 0, 1, 1)[2]))
  # --> location
  expect_true(is.na(pgev(y, x, 1, 0)[4]))
  expect_true(is.na(dgev(y, x, 1, 0)[4]))
  expect_true(is.na(qgev(v, x, 1, 0)[4]))
  expect_true(is.na(pgev(y, x, 1, 0)[2]))
  expect_true(is.na(dgev(y, x, 1, 0)[2]))
  expect_true(is.na(qgev(v, x, 1, 0)[2]))
  # --> scale
  expect_true(is.na(pgev(y, 0, x, 1)[4]))
  expect_true(is.na(dgev(y, 0, x, 1)[4]))
  expect_true(is.na(qgev(v, 0, x, 1)[4]))
  expect_true(is.na(pgev(y, 0, x, 1)[2]))
  expect_true(is.na(dgev(y, 0, x, 1)[2]))
  expect_true(is.na(qgev(v, 0, x, 1)[2]))
  # --> shape
  expect_true(is.na(pgev(y, 0, 1, x)[4]))
  expect_true(is.na(dgev(y, 0, 1, x)[4]))
  expect_true(is.na(qgev(v, 0, 1, x)[4]))
  expect_true(is.na(pgev(y, 0, 1, x)[2]))
  expect_true(is.na(dgev(y, 0, 1, x)[2]))
  expect_true(is.na(qgev(v, 0, 1, x)[2]))
  # Min and Max
  expect_length(distionary:::gev_lower(x, 1, 1), 10)
  expect_length(distionary:::gev_lower(1, x, 1), 10)
  expect_length(distionary:::gev_lower(1, 1, x), 10)
  expect_length(distionary:::gev_upper(x, 1, 1), 10)
  expect_length(distionary:::gev_upper(1, x, 1), 10)
  expect_length(distionary:::gev_upper(1, 1, x), 10)
})

#' Hypergeometric Distribution
#'
#' Creates a Hypergeometric distribution. The parameterization used here
#' is the same as for `stats::phyper()`, where the outcome can be thought
#' of as the number of red balls drawn from an urn of coloured balls,
#' using a scoop that holds a fixed number of balls.
#'
#' @param m The number of red balls in the urn; single positive integer.
#' @param n The number of non-red balls in the urn; single positive integer.
#' @param k the number of balls drawn from the urn (between 0 and `m + n`);
#' single positive integer.
#' @returns A Hypergeometric distribution.
#' @examples
#' dst_hyper(15, 50, 10)
#' @export
dst_hyper <- function(m, n, k) {
  checkmate::assert_numeric(m, 0, len = 1)
  checkmate::assert_numeric(n, 0, len = 1)
  if (is.na(m) || is.na(n)) {
    return(dst_null())
  }
  N <- m + n
  checkmate::assert_numeric(k, 0, N, len = 1)
  if (is.na(k)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(m = m, n = n, k = k),
    pmf = function(x) {
      stats::dhyper(x, m = m, n = n, k = k)
    },
    cdf = function(x) {
      stats::phyper(x, m = m, n = n, k = k)
    },
    quantile = function(p) {
      stats::qhyper(p, m = m, n = n, k = k)
    },
    realise = function(nn) {
      stats::rhyper(nn, m = m, n = n, k = k)
    },
    survival = function(x) {
      stats::phyper(x, m = m, n = n, k = k, lower.tail = FALSE)
    },
    mean = k * m / N,
    variance = k * (m / N) * (n / N) * ((N - k) / (N - 1)),
    skewness = (N - 2 * m) * sqrt(N - 1) * (N - 2 * k) /
      (sqrt(k * m * n * (N - k)) * (N - 2)),
    kurtosis_exc = (
      (N - 1) * N^2 *
        (N * (N + 1) - 6 * m * n - 6 * k * (N - k)) +
        6 * k * m * n * (N - k) * (5 * N - 6)
    ) / (
      k * m * n * (N - k) * (N - 2) * (N - 3)
    ),
    range = c(max(0, k - n), min(m, k)),
    .name = "Hypergeometric",
    .vtype = "discrete"
  )
}

#' Distribution name
#'
#' Print the name of a distribution, possibly with parameters.
#'
#' @param distribution Distribution object.
#' @param param_digits How many significant digits to include when displaying
#' the parameters? `0` if you don't want to display parameters. Length 1
#' vector.
#' @returns A character containing the distribution's name, possibly
#' followed by parameters in brackets.
#' @examples
#' d <- dst_norm(0.3552, 1.1453)
#' pretty_name(d)
#' pretty_name(d, 2)
#' @export
pretty_name <- function(distribution, param_digits = 0) {
  checkmate::check_class(distribution, "dst")
  checkmate::check_integerish(param_digits, 0, len = 1)
  name <- attributes(distribution)$name
  if (is.null(name)) {
    return("Unnamed distribution")
  }
  low_name <- tolower(name)
  if (low_name == "frechet") {
    name <- "Fr\xE9chet"
    Encoding(name) <- "latin1"
    name <- iconv(
      name,
      "latin1",
      "UTF-8"
    )
  }
  if (low_name == "gev") name <- "GEV"
  if (low_name == "gp") name <- "GP"
  if (param_digits > 0 && !is.null(parameters(distribution))) {
    brackets <- bracket_parameters(distribution, param_digits = param_digits)
    name <- paste0(name, brackets)
  }
  name
}

#' Present parameters using bracket notation
#'
#' For distributions that have parameters, this function places those
#' parameters in bracket notation, such as (0, 1) for a Normal distribution
#' with a mean of 0 and variance of 1.
#'
#' @param distribution A single distplyr distribution for which the
#' `parameters()` function can be applied.
#' @param param_digits Number of significant digits to include in the
#' output.
#' @returns a single character with the parameters separated by commas and
#' sandwiched by brackets.
#' @noRd
bracket_parameters <- function(distribution, param_digits = 2) {
  p <- signif(unlist(parameters(distribution)), digits = param_digits)
  paste0("(", paste0(p, collapse = ", "), ")")
}

#' Normal (Gaussian) Distribution
#'
#' Makes a Normal (Gaussian) distribution.
#'
#' @param mean Mean of the distribution. Single numeric.
#' @param sd Standard deviation of the distribution.
#' Single positive numeric.
#' @returns A Normal distribution.
#' @examples
#' dst_norm(0, 1)
#' @export
dst_norm <- function(mean, sd) {
  checkmate::assert_numeric(mean, len = 1)
  checkmate::assert_numeric(sd, 0, len = 1)
  if (is.na(mean) || is.na(sd)) {
    return(dst_null())
  }
  distribution(
    .parameters = list(mean = mean, sd = sd),
    density = function(x) {
      stats::dnorm(x, mean = mean, sd = sd)
    },
    cdf = function(x) {
      stats::pnorm(x, mean = mean, sd = sd)
    },
    quantile = function(p) {
      stats::qnorm(p, mean = mean, sd = sd)
    },
    realise = function(n) {
      stats::rnorm(n, mean = mean, sd = sd)
    },
    survival = function(x) {
      stats::pnorm(x, mean = mean, sd = sd, lower.tail = FALSE)
    },
    mean = mean,
    median = mean,
    variance = sd^2,
    stdev = sd,
    skewness = 0,
    kurtosis_exc = 0,
    range = c(-Inf, Inf),
    .name = "Normal",
    .vtype = "continuous"
  )
}

#' @noRd
eval_mean_from_network <- function(distribution, tol = 1e-9, ...) {
  checkmate::assert_class(distribution, "dst")
  if (vtype(distribution) != "continuous") {
    stop(
      "Numerical computation for non-continuous distributions is ",
      "not yet supported in this version of distionary."
    )
  }
  dens <- representation_as_function(distribution, representation = "density")
  integrand <- function(x) x * dens(x)
  r <- range(distribution)
  distionary_integrate(integrand, lower = r[1], upper = r[2], tol = tol, ...)
}

#' @rdname moments
#' @export
kurtosis <- function(distribution) {
  checkmate::assert_class(distribution, "dst")
  eval_property(distribution, "kurtosis")
}

#' @rdname moments
#' @export
kurtosis_exc <- function(distribution) {
  checkmate::assert_class(distribution, "dst")
  eval_property(distribution, "kurtosis_exc")
}

test_that("Network is invoked in priority: kurtosis_exc", {
  # First look for `kurtosis`, then invoke algorithm if not found.
  d <- distribution(
    density = stats::dnorm,
    range = c(-Inf, Inf),
    kurtosis = 100, # deliberately incorrect
    .vtype = "continuous"
  )
  expect_equal(kurtosis_exc(d), 100 - 3)
  expect_equal(eval_kurtosis_exc_from_network(d), 100 - 3)
  d$kurtosis <- NULL
  expect_equal(kurtosis_exc(d), 0)
  expect_equal(eval_kurtosis_exc_from_network(d), 0)
})

test_that("Excess kurtosis matches kurtosis.", {
  for (item in test_distributions) {
    for (paramset in item$valid) {
      d <- rlang::exec(item$distribution, !!!paramset)
      if (is_intrinsic(d, "kurtosis_exc") || is_intrinsic(d, "kurtosis")) {
        expect_equal(kurtosis(d) - 3, kurtosis_exc(d))
      }
    }
  }
})

test_that("Null distribution works", {
  d <- dst_null()
  expect_equal(mean(d), NA_real_)
  expect_equal(variance(d), NA_real_)
  expect_equal(stdev(d), NA_real_)
  expect_equal(skewness(d), NA_real_)
  expect_equal(kurtosis(d), NA_real_)
  expect_equal(kurtosis_exc(d), NA_real_)
  expect_equal(eval_cdf(d, at = 0), NA_real_)
  expect_equal(eval_density(d, at = 0), NA_real_)
  expect_equal(eval_pmf(d, at = 0), NA_real_)
  expect_equal(eval_quantile(d, at = 0), NA_real_)
  expect_equal(realise(d), NA_real_)
  expect_equal(vtype(d), NA_character_)
})

test_that("Distributions are not sensitive to machine tolerance.", {
  for (item in test_distributions) {
    paramset_orig <- item$valid[[1]]
    eps <- .Machine$double.eps
    paramset_eps <- lapply(paramset_orig, \(x) x - eps)
    d_orig <- rlang::exec(item$distribution, !!!paramset_orig)
    d_eps <- rlang::exec(item$distribution, !!!paramset_eps)
    p <- 0.4
    ## Quantile. Also use x for downstream tests.
    x <- eval_quantile(d_orig, at = p)
    expect_equal(x, eval_quantile(d_eps, p + eps))
    ## CDF
    expect_equal(
      eval_cdf(d_orig, x),
      eval_cdf(d_eps, x + eps)
    )
    ## Survival
    expect_equal(
      eval_survival(d_orig, x),
      eval_survival(d_eps, x + eps)
    )
    ## Mass
    if (!attr(d_orig, "name") %in% c("Degenerate", "Finite")) {
      # Degenerate and Finite distribution's x would also need to increase
      # by eps, which is not meaningful to test.
      expect_equal(
        eval_pmf(d_orig, x),
        eval_pmf(d_eps, x)
      )
    }
    ## Density
    if (vtype(d_orig) == "continuous") {
      expect_equal(
        eval_density(d_orig, x),
        eval_density(d_eps, x + eps)
      )
    }
    ## Mean
    expect_equal(
      mean(d_orig),
      mean(d_eps)
    )
    ## Variance
    expect_equal(
      variance(d_orig),
      variance(d_eps)
    )
    ## Standard Deviation
    expect_equal(
      stdev(d_orig),
      stdev(d_eps)
    )
    ## Skewness
    if (pretty_name(d_orig) == "Log Pearson Type III") {
      expect_equal(
        suppressMessages(skewness(d_orig)),
        suppressMessages(skewness(d_eps))
      )
    } else {
      expect_equal(
        skewness(d_orig),
        skewness(d_eps)
      )
    }
    ## Kurtosis
    if (pretty_name(d_orig) == "Log Pearson Type III") {
      expect_equal(
        suppressMessages(kurtosis(d_orig)),
        suppressMessages(kurtosis(d_eps))
      )
    } else {
      expect_equal(
        kurtosis(d_orig),
        kurtosis(d_eps)
      )
    }
    ## Excess Kurtosis
    if (pretty_name(d_orig) == "Log Pearson Type III") {
      expect_equal(
        suppressMessages(kurtosis_exc(d_orig)),
        suppressMessages(kurtosis_exc(d_eps))
      )
    } else {
      expect_equal(
        kurtosis_exc(d_orig),
        kurtosis_exc(d_eps)
      )
    }
  }
})

#' @noRd
eval_realise_from_network <- function(distribution, n) {
  checkmate::assert_class(distribution, "dst")
  checkmate::assert_integerish(n, lower = 0, len = 1)
  u <- stats::runif(n)
  eval_quantile(distribution, at = u)
}

test_that("All dists are being tested thru test_distributions object", {
  all_namespace <- ls(getNamespace("distionary"))
  dst_namespace <- all_namespace[grep("^dst_", all_namespace)]
  dst_namespace <- dst_namespace[dst_namespace != "dst_null"]
  dst_tested <- vapply(
    test_distributions, function(x) x[[1]], FUN.VALUE = character(1L)
  )
  dst_tested <- unname(dst_tested)
  expect_equal(sort(dst_tested), sort(dst_namespace))
})

---
title: "Pseudocode"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{pseudocode}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


This document outlines the core functionalities of `distionary` for
other programming languages. Developers can add extra checks as
needed.

## Distribution Objects

Distribution objects store user-defined properties and distribution
metadata. The key metadata field is `vtype` (e.g., continuous,
discrete, mixed). Properties and metadata are bundled in a
dictionary-like structure (`{}`).

```
distribution = function(properties, vtype = Null):
  return {properties = properties, vtype = vtype}
```

Currently, `distionary` supports only numeric, univariate 
distributions (i.e., single variable). Future versions
will support multivariate distributions.

## Common Distribution Families

Wrappers allow access to common distributions like Normal and Poisson,
prefixed with `dst_`. For example:

```
dst_norm = function(mean, sd):
  return distribution(
    properties = [
      density = function(x) dnorm(x, mean = mean, sd = sd),
      realise = function(n) rnorm(n, mean = mean, sd = sd),
      ...,
      range = [-Inf, Inf],
      mean = mean,
      stdev = sd,
      skewness = 0,
      ...
    ],
    vtype = "continuous"
  )
```

Here, `[]` is a list-like structure that can be used to pass arguments
to the function. `dnorm` and `rnorm` are assumed to be available or
should be implemented manually.

In addition to common families, a Null distribution that always
evaluates to NA should be created. This object is useful
when external algorithms fail, where a Null distribution can be output
instead of an error.

## Evaluating

Start with thin wrappers for evaluating specific properties,
which invoke the general function `eval_property`.

Properties that are not functions can be accessed by
functions of the same name:

```
mean = function(distribution):
  eval_property(distribution, name = "mean")

range = function(distribution):
  eval_property(distribution, name = "range")
```

Properties that are functions are prefixed with `eval_` to indicate they
require an argument, passed with `at`. All such properties currently
in distionary only have one argument, but in principle could accept
others; these all go in `[]`.

```
eval_density = function(distribution, at):
  eval_property(distribution, name = "density", args = [at])

eval_survival = function(distribution, at):
  eval_property(distribution, name = "cdf", args = [at])
```

The `at` parameter can accept a vector of values. The current
version supports a single distribution input, but future 
versions might allow recycling both arguments.

The main function `eval_property` evaluates properties if
present or delegates network evaluation.
If the property is a function, it will evaluate it
using the provided arguments in `args`. By default, `args` 
is `Null`, meaning arguments are not relevant (therefore,
this argument can be ignored for numeric properties).

```
eval_property = function(distribution, property_name, args = Null):
  properties = subset(distribution, "properties")
  property = subset(properties, property_name)
  if is_null(property):
    return eval_from_network(distribution, name = property_name, args = args)
  else if is_function(property):
    return property(args)
  else:
    return property
```

Use `subset` to retrieve entries from dictionary-like
structures. `property(args)` means to call the function
property with arguments.

The absence of a property is evaluated using a
[network of properties](https://distionary.probaverse.com/articles/specify-user-defined#network-of-properties),
specifying how one property can be retrieved from others.
For example, the cumulative hazard function (CHF)
can be calculated using the survival function:

```
eval_from_network = function(distribution, name, args):
  if name == "chf":
    return -log(eval_survival(distribution, args))
  if name == "survival":
    return 1 - eval_cdf(distribution, args)
  if name == "mean":
  ...
```

Network evaluation is still robust against whether the survival function
has been specified for the distribution due to the use of 
`eval_survival`: if the survival function can't be retrieved
from the distribution, it will evaluate it from the network,
this time from the cumulative distribution function (CDF).
The network has been set up to avoid getting caught in a loop,
so that the survival function doesn't look back to the CHF
for evaluation -- it's (mostly) a directed acyclic graph.

Network computation failures throw errors, though future 
versions might allow user-specified error-handling strategies.

The defined network is not unique. For example, the CHF could
be retrieved from the hazard function rather than the survival
function. The network was minimized for numerical efficiency.


Currently, the CDF and density/PMF anchor the property network,
meaning that they are assumed available. Future
improvements may implement networks as their own objects,
customized to each distribution to be seeded
from the properties specified in each distribution.

## Enframing

Instead of `eval`uating properties that are functions, `enframe` these to place
inputs and outputs in tabular form. This is useful for table-centric
data analyses. Start with simple wrappers:

```
enframe_density = function(distributions, at):
  enframe_general(distributions, at = at, eval_fn = eval_density)

enframe_survival = function(distributions, at):
  enframe_general(distributions, at = at, eval_fn = eval_survival)
```

`distributions` might contain multiple distributions. The
main function loops over each distribution, adding results 
to new columns:

```
enframe_general = function(distributions, at, eval_fn):
  tbl = [x = at]
  for d in distributions:
    y = eval_fn(d, at)
    tbl = append(tbl, [y])
  return tbl
```

The developer may wish to consider features related to column naming.

---
title: "Evaluate a Distribution"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Evaluate a Distribution}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(distionary)
```

This vignette covers the second goal of `distionary`: to evaluate probability distributions, even when that property is not specified in the distribution's definition.

## Distributional Representations

A _distributional representation_ is a mathematical function that completely defines a probability distribution. Unlike a simple property (such as the mean or variance), a representation contains enough information that any other property or representation can be calculated from it.

The key innovation in `distionary` is that these representations are interconnected through a network of relationships, allowing you to specify a distribution using any available representation and automatically derive others as needed. For example, if you specify only a CDF, `distionary` can compute the quantile function, mean, variance, and other properties.

Here is a list of representations recognised by `distionary`, and the functions for accessing them. 

| Representation                   | `distionary` Functions                  |
|----------------------------------|-----------------------------------------|
| Cumulative Distribution Function | `eval_cdf()`,      `enframe_cdf()`      |
| Survival Function                | `eval_survival()`, `enframe_survival()` |
| Quantile Function                | `eval_quantile()`, `enframe_quantile()` |
| Hazard Function                  | `eval_hazard()`,   `enframe_hazard()`   |
| Cumulative Hazard Function       | `eval_chf()`,      `enframe_chf()`      |
| Probability density Function     | `eval_density()`,  `enframe_density()`  |
| Probability mass Function (PMF)  | `eval_pmf()`,      `enframe_pmf()`      |
| Odds Function                    | `eval_odds()`,     `enframe_odds()`     |
| Return Level Function            | `eval_return()`,   `enframe_return()`   |

All representations can either be accessed by the `eval_*()` family of functions, providing a vector of the evaluated representation.

```{r}
d1 <- dst_geom(0.6)
eval_pmf(d1, at = 0:5)
```

Alternatively, the `enframe_*()` family of functions provides the results in a tibble or data frame paired with the inputs, useful in a data wrangling workflow.

```{r}
enframe_pmf(d1, at = 0:5)
```

The `enframe_*()` functions allow for insertion of multiple distributions, placing a column for each distribution. The column names can be changed in three ways:

1. The input column `.arg` can be renamed with the `arg_name` argument.
2. The `pmf` prefix on the evaluation columns can be changed with the `fn_prefix` argument.
3. The distribution names can be changed by assigning name-value pairs for the input distributions.

Let's practice this with the addition of a second distribution.

```{r}
d2 <- dst_geom(0.4)
enframe_pmf(
  model1 = d1, model2 = d2, at = 0:5,
  arg_name = "num_failures", fn_prefix = "probability"
)
```

## Drawing a random sample

To draw a random sample from a distribution, use the `realise()` or `realize()` function:

```{r}
set.seed(42)
realise(d1, n = 5)
```

You can read this call as "realise distribution `d` five times". By default, `n` is set to 1, so that realising converts a distribution to a numeric draw:

```{r}
realise(d1)
```

While random sampling falls into the same family as the `p*/d*/q*/r*` functions from the `stats` package (e.g., `rnorm()`), this function is not a distributional representation, hence does not have a `eval_*()` or `enframe_*()` counterpart. This is because it's impossible to perfectly describe a distribution based on a sample. 

## Properties of Distributions

`distionary` distinguishes between _distributional representations_ (which fully define a distribution) and _distributional properties_ (which are characteristics that can be computed from representations). 

A distribution _property_ is any measurable characteristic that can be calculated from a distribution's representation. Unlike representations, properties do not contain enough information to fully reconstruct the distribution. For example, knowing the mean and variance of a distribution doesn't tell you whether it's a Normal, Gamma, or some other distribution family. Properties include statistical moments and other summary measures.

Below is a table of the properties incorporated in `distionary`, and the corresponding functions for accessing them.

| Property | `distionary` Function |
|----------|---------------------|
| Mean                       | `mean()` |
| Median                     | `median()` |
| Variance                   | `variance()` |
| Standard Deviation         | `sd()` |
| Skewness                   | `skewness()` |
| Excess Kurtosis            | `kurtosis_exc()` |
| Kurtosis                   | `kurtosis()` |
| Range                      | `range()` |

Here's the mean and variance of our original distribution.

```{r}
mean(d1)
variance(d1)
```

---
title: "Built-In Distributions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Built-In Distributions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(distionary)
```

This vignette covers built-in distribution families available in `distionary`. It provides insight into each family and its counterparts in the `stats` package.

## Built-In Distribution Families

All distribution families found in the `stats` package are integrated into `distionary`, along with a few others. They are shown in the table below with notes on whether they have counterparts in the `stats` package.

| Distribution                     | `distionary` Function  |  Has counterpart in `stats` |
|----------------------------------|------------------------|-----------------------------|
| Bernoulli | `dst_bern()` | Yes |
| Beta | `dst_beta()` | Yes |
| Binomial | `dst_binom()` | Yes |
| Cauchy | `dst_cauchy()` | Yes |
| Chi Squared | `dst_chisq()` | Yes |
| Degenerate | `dst_degenerate()` | No |
| Exponential | `dst_exp()` | Yes |
| Empirical | `dst_empirical()` | No |
| F | `dst_f()` | Yes |
| Finite | `dst_finite()` | No |
| Gamma | `dst_gamma()` | Yes |
| Geometric | `dst_geom()` | Yes |
| Generalised Extreme Value (GEV) | `dst_gev()` | No |
| Generalised Pareto (GP) | `dst_gp()` | No |
| Hypergeometric | `dst_hyper()` | Yes |
| Log Normal | `dst_lnorm()` | Yes |
| Log Pearson Type III | `dst_lp3()` | No |
| Negative Binomial | `dst_nbinom()` | Yes |
| Normal | `dst_norm()` | Yes |
| Pearson Type III | `dst_pearson3()` | No |
| Poisson | `dst_pois()` | Yes |
| Student _t_ | `dst_t()` | Yes |
| Uniform | `dst_unif()` | Yes |
| Weibull | `dst_weibull()` | Yes |

In addition, there is a special "Null" distribution object akin to a missing or unknown distribution. This is useful, for example, if an algorithm fails to return a distribution: instead of throwing an error, a Null distribution can be returned.

```{r}
# Make a Null distribution.
null <- dst_null()
# Inspect
null
```

This is the behaviour when specifying `NA` as a distribution parameter:

```{r}
dst_norm(mean = 0, sd = NA)
```

The Null distribution always evaluates to `NA`.

```{r}
mean(null)
eval_pmf(null, at = 1:10)
range(null)
```

## The Empirical Distribution

The Empirical distribution is different from the others in terms of its utility because it makes _your data_ the distribution, and is therefore far more flexible than any of the other distributions.

In statistical terminology, this is because the other built-in distributions are _parametric_, defined by a fixed amount of numeric parameters. For example, exactly two parameters make up the family of Normal distributions. The empirical distribution, on the other hand, is a _non-parametric_ model, because there isn't a fixed number of numeric parameters that can describe it.

The Empirical distribution is particularly advantageous when a distribution's standard form is unknown or infeasible to model, or when comparing a custom model against data is needed. 

Here is an example of forming an empirical distribution from a dataset generated from a Normal distribution. First, define the normal distribution and generate a sample; the first few observations are shown below.

```{r}
set.seed(42)
normal <- dst_norm(0, 1)
x <- realise(normal, n = 100)
# Inspect the first few observations
head(x)
```

An Empirical distribution can be made from these values.

```{r}
empirical <- dst_empirical(x)
# Inspect
empirical
```

Comparing its CDF to the normal distribution, one can see the two are similar:

```{r}
plot(empirical, "cdf", from = -4, to = 4, n = 1000)
plot(normal, "cdf", add = TRUE, col = "red")
legend(
  "topleft", 
  legend = c("Empirical", "Normal"), 
  col = c("black", "red"), 
  lty = 1
)
```

Although this distribution is non-parametric, the `parameters()` function is still applicable because it's not tied to the statistical definition of "parametric", which is concerned with parameter dimension. In `distionary`, the dataset itself and their probabilities comprise the distribution parameters (using `str()` here for better printing):

```{r}
str(parameters(empirical))
```

One should also note that Empirical distributions are discrete, with outcomes defined by the observed values:

```{r}
vtype(empirical)
```

---
title: "Specifying Your Own Distribution"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Specifying Your Own Distribution}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(distionary)
```

This vignette explains how to create user-defined distributions using `distionary`.

## User-Defined Distributions

You can make your own distribution using the `distribution()` function. Specify distribution properties as name-value pairs, where the name is the distribution property. Distributional _representations_ like the cumulative distribution function (CDF) can be specified as functions.

```{r}
my_normal <- distribution(
  density = stats::dnorm,
  cdf = stats::pnorm,
  range = c(-Inf, Inf),
  g = 9.81,
  another_representation = function(x) x^2,
  .vtype = "continuous"
)
# Inspect
my_normal
```

The usual evaluation framework can now be accessed, such as evaluating the CDF.

```{r}
eval_cdf(my_normal, at = c(0.2, 0.5, 0.7))
```

Even though the mean has not been specified, it can still be evaluated.

```{r}
mean(my_normal)
```

Some of the specified properties are special because they correspond to a property that's known to distionary -- in this example, `density`, `cdf`, and `range`. In general, the special names can be identified as follows:

- The suffix following `eval_` for any distributional representation (e.g., `quantile` for `eval_quantile()`). These should be specified as functions.
- `realise` or `realize`, which is used to generate random samples from the distribution. This should be a function of the number of observations to draw.
- The same name as the `distionary` functions used to extract any other property (e.g., `mean` for `mean()`).

See the [Evaluate a Distribution](./evaluate.html) vignette for more details on these evaluation functions and the network of relationships between the understood properties. Note that, in the current version, specifying a `cdf` along with a `density` or `pmf` (probability mass function) is required for evaluating non-explicit properties.

In general, properties can be invoked by the more general function `eval_property()`, which is useful for accessing entries that are not known to `distionary`. For example, the user-defined property `g` can be accessed in this way.

```{r}
eval_property(my_normal, "g")
```

To use `eval_property()` to evaluate a function, arguments can be specified in the `...` argument.

```{r}
eval_property(my_normal, "another_representation", 1:4)
```

The `eval_property()` function is also useful for programmatically accessing distribution properties.

```{r}
properties <- c("mean", "variance")
lapply(properties, \(x) eval_property(my_normal, x))
```

## Distribution Attributes

Distribution metadata are stored as attributes, and are those arguments prefixed by `.` in the `distribution()` function.

- `.vtype`, as seen in the example starting this vignette, is used to specify the variable type, such as `"discrete"` or `"continuous"`.
- `.name` allows you to give the distribution a name.
- `.parameters` allows you to specify values for the distribution's parameters as a list, if applicable. This version of `distionary` does not draw from these values, so it currently serves a similar function as the distribution's name.

```{r}
my_distribution <- distribution(
  cdf = pnorm,
  density = dnorm,
  .vtype = "continuous",
  .name = "Special",
  .parameters = list(theta = 1.7, mat = diag(2), hello = "hi")
)
# Inspect
my_distribution
```

Retrieve the variable type:

```{r}
vtype(my_distribution)
```

Retrieve the parameters:

```{r}
parameters(my_distribution)
```

You can also reset the parameters with `<-`. Lets give this distribution scalar parameters.

```{r}
parameters(my_distribution)[["mat"]] <- 1
parameters(my_distribution)[["hello"]] <- NULL
# Inspect the distribution
my_distribution
```

Retrieve the distribution's name with `pretty_name()`.

```{r}
pretty_name(my_distribution)
```

Names are "pretty" because they can also include the parameters in a compact way, when the parameters are scalars (which they now are):

```{r}
pretty_name(my_distribution, param_digits = 2)
```

## Example of a New Parametric Family

If you want to create your own parametric family, such as a distribution whose density decays linearly from `x=0` to `x=a`, you can do this by making a function that accepts the distribution parameter as an input, and outputs the distribution. Note that the CDF and density are specified, as required for continuous distributions, and metadata are specified with the arguments starting with `.`.

```{r}
dst_linear <- function(a) {
  # It helps to check that the parameter is valid.
  checkmate::assert_number(a, lower = 0)
  # We'll create some representations outside of the `distribution()`
  # call for separation of concerns. Note that care is taken to ensure
  # proper behaviour outside of the range of the distribution.
  density <- function(x) {
    d <- 2 / a^2 * (a - x)
    d[x < 0 | x > a] <- 0
    d
  }
  cdf <- function(x) {
    p <- x / a^2 * (2 * a - x)
    p[x < 0] <- 0
    p[x > a] <- 1
    p
  }
  # Create the distribution.
  distribution(
    density = density,
    cdf = cdf,
    range = c(0, a),
    .vtype = "continuous",
    .name = "Linear",
    .parameters = list(a = a)
  )
}
```

Notice how the distribution metadata gets printed with the distribution.

```{r}
dst_linear(3)
```

Here are the density plots of different members of this family.

```{r}
plot(dst_linear(1), to = 4, col = "purple")
plot(dst_linear(2), add = TRUE, col = "red")
plot(dst_linear(4), add = TRUE, col = "blue")
legend(
  "topright",
  legend = c("a = 1", "a = 2", "a = 4"),
  col = c("purple", "red", "blue"),
  lty = 1
)
```

As usual, properties can be evaluated, even if they are not specified. Consider, for example, calculating return periods of three different Linear distributions.

```{r}
enframe_return(
  model1 = dst_linear(1),
  model2 = dst_linear(2),
  model3 = dst_linear(4),
  at = c(2, 5, 10, 20, 50, 100),
  arg_name = "return_period"
)
```

## Network of Properties

The key innovation in `distionary` is its network-based approach to probability distributions. Different distributional representations (CDF, PDF/PMF, quantile function, survival function, etc.) are interconnected through a network of mathematical relationships.

The diagram below shows how `distionary` derives distribution properties from other properties. An arrow from one property to another indicates that the first property can be calculated from the second. For example, `quantile` can be calculated from `cdf`, and `mean` can be calculated from `density`. This network structure is what allows you to define a custom distribution with just one or two representations (like `cdf` and `density` in the example above) and still have access to all other properties.

<!-- 
Network diagram generated using man/figures/network_diagram.R script.
The script uses DiagrammeR and Graphviz DOT language to create the diagram.
To regenerate: Rscript man/figures/network_diagram.R
-->

```{r, echo=FALSE, out.width="100%", fig.align="center"}
knitr::include_graphics("network_diagram.png")
```

## Making your own properties

Some properties are easy to make yourself. Here is an example of a function that calculates interquartile range.

```{r}
# Make a function that takes a distribution as input, and returns the
# interquartile range.
iqr <- function(distribution) {
  diff(eval_quantile(distribution, at = c(0.25, 0.75)))
}
```

Apply the function to a Linear distribution as defined before.

```{r}
iqr(dst_linear(2))
```

For properties that are not handled by `distionary` (e.g., extreme value index, or moment generating function), one option is to build these properties into your own distribution. A future version of `distionary` will make user-defined properties easier to work with.

## Limitations

The `distionary` package is still young. Lots of helpful features are still missing, and your patience is appreciated as more features are developed. Some examples follow.

| #  | Limitation | Explanation |
|----|------------|------|
| 1. | When a user-specified distribution isn't continuous (e.g., is discrete), only the properties specified in `distribution()` can be accessed in this version. | Specifying the set of possible outcomes for discrete distributions requires special attention, particularly when there are infinitely many of them. |
| 2. | Representations are not checked for accuracy in this version. | While this is done in the test suite for the built-in distributions, additional levels of detail are required when accepting a foreign distribution. |
| 3. | Typos when specifying distribution property names (e.g. `densty` instead of `density`) or variable type (e.g., `"contnous"` instead of `"continuous"`) will not trigger an error. | The package does not assume that distribution properties and types are limited, allowing for flexibility. |
| 4. | It is currently assumed that the distributional representations are specified in a way that remains valid beyond the range of the distribution. | A future version aims to use the specified range to automatically implement appropriate behavior. |

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R, R/new_distribution.R
\name{distribution}
\alias{distribution}
\alias{is_distribution}
\alias{is.distribution}
\title{Build a Distribution Object}
\usage{
distribution(..., .vtype = NULL, .name = NULL, .parameters = list())

is_distribution(object)

is.distribution(object)
}
\arguments{
\item{...}{Name-value pairs for defining the distribution.}

\item{.vtype}{The variable type, typically "discrete" or "continuous".
Can be any character vector of length 1, but is converted to
lowercase with \code{tolower()} for compliance with known types.}

\item{.name}{A name to give to the distribution.
Can be any character vector of length 1.}

\item{.parameters}{A named list with one entry per distribution parameter,
each of which can be any data type. In this version of distionary,
these parameters are only stored for the benefit of the user to know
what distribution they are working with; the code never looks at these
parameters to inform its calculations. This is anticipated to change in
a future version of distionary.}

\item{object}{Object to be tested}
}
\value{
A distribution object.
}
\description{
Make a distribution object by specifying properties
(e.g., cdf, density, mean, etc.).
Some properties, if not included, will be calculated based
on other properties that are included (e.g., quantile from cdf;
variance from standard deviation).
A list of these representations can be found in the details.
}
\details{
Currently, the CDF (\code{cdf}) is required to be specified, along with the PMF
(\code{pmf}) for discrete distributions and density (\code{density}) for continuous
distributions. Otherwise, the full extent of distribution properties will
not be accessible.

A distributional representation is a function that fully describes the
distribution. Besides \code{cdf}, \code{density}, and \code{pmf}, other options
understood by \code{distionary} include:
\itemize{
\item \code{survival}: the survival function, or one minus the cdf.
\item \code{hazard}: the hazard function, for continuous variables only.
\item \code{chf}: the cumulative hazard function, for continuous variables only.
\item \code{quantile}: the quantile function, or left-inverse of the cdf.
\item \code{realise} or \code{realize}: a function that takes an integer and generates
a vector of that many random draws from the distribution.
\item \code{odds}: for discrete variables, the probability odds function
(pmf / (1 - pmf))
\item \code{return}: the quantiles associated with the provided return periods,
where events are exceedances.
}

All functions should be vectorized.

Other properties that are understood by \code{distionary} include:
\itemize{
\item \code{mean}, \code{stdev}, \code{variance}, \code{skewness}, \code{median} are self-explanatory.
\item \code{kurtosis_exc} and \code{kurtosis} are the distribution's excess
kurtosis and regular kurtosis.
\item \code{range}: A vector of the minimum and maximum value of a distribution's
support.
}
}
\examples{
linear <- distribution(
  density = function(x) {
    d <- 2 * (1 - x)
    d[x < 0 | x > 1] <- 0
    d
  },
  cdf = function(x) {
    p <- 2 * x * (1 - x / 2)
    p[x < 0] <- 0
    p[x > 1] <- 1
    p
  },
  .vtype = "continuous",
  .name = "My Linear",
  .parameters = list(could = "include", anything = data.frame(x = 1:10))
)

# Inspect
linear

# Plot
plot(linear)
}
\concept{Distribution Construction}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pretty_name.R
\name{pretty_name}
\alias{pretty_name}
\title{Distribution name}
\usage{
pretty_name(distribution, param_digits = 0)
}
\arguments{
\item{distribution}{Distribution object.}

\item{param_digits}{How many significant digits to include when displaying
the parameters? \code{0} if you don't want to display parameters. Length 1
vector.}
}
\value{
A character containing the distribution's name, possibly
followed by parameters in brackets.
}
\description{
Print the name of a distribution, possibly with parameters.
}
\examples{
d <- dst_norm(0.3552, 1.1453)
pretty_name(d)
pretty_name(d, 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_null.R
\name{dst_null}
\alias{dst_null}
\title{Null Distribution}
\usage{
dst_null()
}
\value{
A Null distribution.
}
\description{
Sometimes it's convenient to work with a distribution object that is
akin to a missing value. This is especially true when programmatically
outputting distributions, such as when a distribution fails to fit to
data. This function makes such a distribution object. It always evaluates
to \code{NA}.
}
\examples{
x <- dst_null()
mean(x)
eval_pmf(x, at = 1:10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_geom.R
\name{dst_geom}
\alias{dst_geom}
\title{Geometric Distribution}
\usage{
dst_geom(prob)
}
\arguments{
\item{prob}{Probability of success in each trial; single numeric
between 0 and 1.}
}
\value{
A Geometric distribution.
}
\description{
Makes a Geometric distribution, corresponding to the number of failures
in a sequence of independent trials before observing a success.
}
\examples{
d <- dst_geom(0.4)

# This version of the Geometric distribution does not count the success.
range(d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_gp.R
\name{dst_gp}
\alias{dst_gp}
\title{Generalised Pareto Distribution}
\usage{
dst_gp(scale, shape)
}
\arguments{
\item{scale}{Scale parameter; single positive numeric.}

\item{shape}{Shape parameter; single positive numeric.
This is also the extreme value index, so that \code{shape > 0} is heavy
tailed, and \code{shape < 0} is short-tailed.}
}
\value{
A Generalised Pareto Distribution.
}
\description{
Makes a Generalized Pareto (GP) distribution, corresponding to the
limiting distribution of excesses over a threshold.
}
\examples{
# Short-tailed example
short <- dst_gp(1, -1)
range(short)
mean(short)

# Heavy-tailed example
heavy <- dst_gp(1, 1)
range(heavy)
mean(heavy)

# Light-tailed example (a Gumbel distribution)
light <- dst_gp(1, 0)
range(light)
mean(light)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_lp3.R
\name{dst_lp3}
\alias{dst_lp3}
\title{Log Pearson Type III distribution}
\usage{
dst_lp3(meanlog, sdlog, skew)
}
\arguments{
\item{meanlog}{Mean of the log of the random variable; single numeric.}

\item{sdlog}{Standard deviation of the log of the random variable;
single positive numeric.}

\item{skew}{Skewness of the log of the random variable;
single numeric.}
}
\value{
A Log Pearson Type III distribution.
}
\description{
Makes a Log Pearson Type III distribution, which is the
distribution of the exponential of a random variable following
a Pearson Type III distribution.
}
\examples{
dst_lp3(0, 1, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.dst}
\alias{plot.dst}
\title{Plot a Distribution}
\usage{
\method{plot}{dst}(
  x,
  what = c("density", "pmf", "cdf", "survival", "quantile", "hazard", "chf"),
  ...
)
}
\arguments{
\item{x}{Distribution object}

\item{what}{Name of the representation to plot.}

\item{...}{Other arguments to pass to the \code{graphics::curve} function,
or \code{graphics::plot} in the case of the PMF.}
}
\value{
This function is run for its graphics byproduct, and therefore
returns the original distribution, invisibly.
}
\description{
Plot a distribution's representation.
}
\examples{
d <- dst_norm(0, 1)
plot(d, from = -4, to = 4)
plot(d, "cdf", n = 1000)
plot(d, "survival")
plot(d, "quantile")
plot(d, "hazard")
plot(d, "chf")

p <- dst_pois(4)
plot(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_pois.R
\name{dst_pois}
\alias{dst_pois}
\title{Poisson Distribution}
\usage{
dst_pois(lambda)
}
\arguments{
\item{lambda}{Mean of the Poisson distribution; single positive numeric.}
}
\value{
A Poisson distribution.
}
\description{
Makes a Poisson distribution.
}
\examples{
dst_pois(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval_chf.R
\name{eval_chf}
\alias{eval_chf}
\alias{enframe_chf}
\title{Cumulative Hazard Function}
\usage{
eval_chf(distribution, at)

enframe_chf(..., at, arg_name = ".arg", fn_prefix = "chf", sep = "_")
}
\arguments{
\item{distribution, ...}{A distribution, or possibly multiple
distributions in the case of \code{...}.}

\item{at}{Vector of values to evaluate the representation at.}

\item{arg_name}{For \code{enframe_}, name of the column containing
the function arguments. Length 1 character vector.}

\item{fn_prefix}{For \code{enframe_}, name of the function to
appear in the column(s). Length 1 character vector.}

\item{sep}{When \code{enframe}'ing more than one distribution, the
character that will be separating the \code{fn_name} and the distribution name.
Length 1 character vector.}
}
\value{
The evaluated representation in vector form (for \code{eval_})
with length matching the length of \code{at}, and data frame
or tibble form (for \code{enframe_}) with number of rows matching the
length of \code{at}. The \code{at} input occupies the first column,
named \code{.arg} by default, or the specification in \code{arg_name};
the evaluated representations for each distribution in \code{...}
go in the subsequent columns (one column per distribution). For a
single distribution, this column is named according to the
representation by default (cdf, survival, quantile, etc.),
or the value in \code{fn_prefix}. For multiple distributions, unnamed
distributions are auto-named, and columns are named
\verb{<fn_prefix><sep><distribution_name>} (e.g., \code{cdf_distribution1}).
}
\description{
Access a distribution's cumulative hazard function (chf).
}
\examples{
d <- dst_unif(0, 4)
eval_chf(d, at = 0:4)
enframe_chf(d, at = 0:4)
}
\seealso{
Other distributional representations: 
\code{\link{eval_cdf}()},
\code{\link{eval_density}()},
\code{\link{eval_hazard}()},
\code{\link{eval_odds}()},
\code{\link{eval_pmf}()},
\code{\link{eval_quantile}()},
\code{\link{eval_return}()},
\code{\link{eval_survival}()}
}
\concept{distributional representations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval_quantile.R
\name{eval_quantile}
\alias{eval_quantile}
\alias{enframe_quantile}
\title{Distribution Quantiles}
\usage{
eval_quantile(distribution, at)

enframe_quantile(..., at, arg_name = ".arg", fn_prefix = "quantile", sep = "_")
}
\arguments{
\item{distribution, ...}{A distribution, or possibly multiple
distributions in the case of \code{...}.}

\item{at}{Vector of values to evaluate the representation at.}

\item{arg_name}{For \code{enframe_}, name of the column containing
the function arguments. Length 1 character vector.}

\item{fn_prefix}{For \code{enframe_}, name of the function to
appear in the column(s). Length 1 character vector.}

\item{sep}{When \code{enframe}'ing more than one distribution, the
character that will be separating the \code{fn_name} and the distribution name.
Length 1 character vector.}
}
\value{
The evaluated representation in vector form (for \code{eval_})
with length matching the length of \code{at}, and data frame
or tibble form (for \code{enframe_}) with number of rows matching the
length of \code{at}. The \code{at} input occupies the first column,
named \code{.arg} by default, or the specification in \code{arg_name};
the evaluated representations for each distribution in \code{...}
go in the subsequent columns (one column per distribution). For a
single distribution, this column is named according to the
representation by default (cdf, survival, quantile, etc.),
or the value in \code{fn_prefix}. For multiple distributions, unnamed
distributions are auto-named, and columns are named
\verb{<fn_prefix><sep><distribution_name>} (e.g., \code{cdf_distribution1}).
}
\description{
Access a distribution's quantiles.
}
\details{
When a quantile function does not exist, an algorithm is
deployed that calculates the left inverse of the CDF. This algorithm
works by progressively cutting the specified range in half, moving
into the left or right half depending on where the solution is.
The algorithm is not currently fast and is subject to improvement,
and is a simple idea that has been passed around on the internet
here and there. Tolerance is less than 1e-9, unless the maximum
number of iterations (200) is reached.

The algorithm is not new, and is rather simple.
The algorithm works by progressively
cutting an initially wide range in half, moving into the left or right
half depending on where the solution is. I found the idea on Stack
Overflow somewhere, but unfortunately cannot find the location anymore.
}
\examples{
d <- dst_unif(0, 4)
eval_quantile(d, at = 1:9 / 10)
enframe_quantile(d, at = 1:9 / 10)
}
\seealso{
Other distributional representations: 
\code{\link{eval_cdf}()},
\code{\link{eval_chf}()},
\code{\link{eval_density}()},
\code{\link{eval_hazard}()},
\code{\link{eval_odds}()},
\code{\link{eval_pmf}()},
\code{\link{eval_return}()},
\code{\link{eval_survival}()}
}
\concept{distributional representations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range.r
\name{range.dst}
\alias{range.dst}
\title{Range of Distribution}
\usage{
\method{range}{dst}(distribution, ...)
}
\arguments{
\item{distribution}{Distribution to compute range from.}

\item{...}{Not used; vestige of the \code{base::range()} S3 generic.}
}
\value{
Vector of length two, containing the minimum and maximum
values of a distribution.
}
\description{
Range returns a vector of length two, with the minimum and maximum
values of the (support of the) distribution.
}
\details{
If there are no methods for the distribution's class,
the range is calculated
using \code{eval_quantile()} at 0 and at 1.
}
\examples{
a <- dst_gp(1, 0.5)
b <- dst_unif(0, 1)
c <- dst_norm(3, 4)
range(a)
range(b)
range(c)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_binom.R
\name{dst_binom}
\alias{dst_binom}
\title{Binomial Distribution}
\usage{
dst_binom(size, prob)
}
\arguments{
\item{size}{Number of trials; single positive integer.}

\item{prob}{Success probability of each trial; single numeric
between 0 and 1.}
}
\value{
A binomial distribution.
}
\description{
Makes a Binomial distribution, representing the number of successes
in a fixed number of independent trials.
}
\examples{
dst_binom(10, 0.6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_f.R
\name{dst_f}
\alias{dst_f}
\title{F Distribution}
\usage{
dst_f(df1, df2)
}
\arguments{
\item{df1, df2}{Degrees of freedom of the numerator and denominator,
both single positive numerics.}
}
\value{
An F distribution.
}
\description{
Makes an F distribution.
}
\examples{
dst_f(2, 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdq_gp.R
\name{pgp}
\alias{pgp}
\alias{qgp}
\alias{dgp}
\title{Representations of the Generalized Pareto Distribution}
\usage{
pgp(q, scale, shape, lower.tail = TRUE)

qgp(p, scale, shape)

dgp(x, scale, shape)
}
\arguments{
\item{scale}{Vector of scale parameters; positive numeric.}

\item{shape}{Vector of shape parameters; positive numeric.}

\item{lower.tail}{Single logical. If \code{TRUE}, cdf (default);
if \code{FALSE}, survival function.}

\item{p}{Vector of probabilities.}

\item{x, q}{Vector of quantiles.}
}
\value{
Vector of evaluated GP distribution, with length
equal to the recycled lengths of \code{q}/\code{x}/\code{p}, \code{scale}, and \code{shape}.
}
\description{
Representations of the Generalized Pareto Distribution
}
\examples{
pgp(1:10, 1, 1)
dgp(1:10, 2, 0)
qgp(1:9 / 10, 10, -2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_weibull.R
\name{dst_weibull}
\alias{dst_weibull}
\title{Weibull Distribution}
\usage{
dst_weibull(shape, scale)
}
\arguments{
\item{shape}{Shape parameter; single positive numeric.}

\item{scale}{Scale parameter; single positive numeric.}
}
\value{
A Weibull distribution.
}
\description{
Makes a Weibull distribution.
}
\examples{
dst_weibull(1, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_empirical.R
\name{dst_empirical}
\alias{dst_empirical}
\title{Empirical Distribution}
\usage{
dst_empirical(
  y,
  weights = 1,
  data = NULL,
  na_action_y = c("null", "drop", "fail"),
  na_action_w = c("null", "drop", "fail")
)
}
\arguments{
\item{y}{<\code{data-masking}> Numeric vector representing the potential
outcomes of the distribution.}

\item{weights}{<\code{data-masking}> Numeric vector of weights corresponding to
to the outcomes \code{y}. These will be scaled so that they add up to 1.}

\item{data}{Optionally, a data frame to compute \code{y} and \code{weights} from.
\code{NULL} if data are not coming from a data frame (the default).}

\item{na_action_y, na_action_w}{What should be done with \code{NA} entries in
\code{y} and \code{w}eights? Character vector of length 1: one of \code{"fail"},
\code{"null"} (default), or \code{"drop"}. See details.}
}
\value{
A finite distribution. If only one outcome, returns a degenerate
distribution. Returns a Null distribution if \code{NA} values are present
and \code{"null"} is specified as an NA action.
}
\description{
An empirical distribution is a non-parametric way to
estimate a distribution using data. By default,
it assigns equal probability to all observations
(this can be overridden with the \code{weights} argument).
Identical to \code{\link[=dst_finite]{dst_finite()}} with NA handling and with weights not needing
to add to 1.
}
\details{
\code{y} and \code{weights} are recycled to have the same length, but only
if one of them has length 1 (via \code{vctrs::vec_recycle_common()}).

\code{na_action_y} and \code{na_action_w} specify the NA action for \code{y} and \code{weights}.
Options are, in order of precedence:
\itemize{
\item \code{"fail"}: Throw an error in the presence of \code{NA}.
\item \code{"null"}: Return a Null distribution (\code{dst_null()}) in the presence
of \code{NA}.
\item \code{"drop"} (the default for \code{na_action_w}): Remove outcome-weight pairs
having an \code{NA} value in the specified vector.
}
}
\examples{
t <- -2:7
dst_empirical(t)

# Using a data frame
df <- data.frame(time = c(NA, NA, t))
dst_empirical(time * 60, data = df)  # Null, since `NA` in `time`.

# Drop NA `time` values.
dst_empirical(time * 60, data = df, na_action_y = "drop")

# Weights explicit. Zero-weight outcomes ("-120") are gone.
df$w <- c(1, 1, 0:9)
dst_empirical(time * 60, w, data = df, na_action_y = "drop")

# "Null" takes precedence over "drop".
df$w <- c(NA, NA, 0:9)
df$time[1] <- -3
df$time[12] <- NA
dst_empirical(time, w, data = df, na_action_w = "null", na_action_y = "drop")
dst_empirical(time, w, data = df, na_action_w = "drop", na_action_y = "null")
dst_empirical(time, w, data = df, na_action_w = "drop", na_action_y = "drop")
}
\seealso{
\code{\link[=dst_finite]{dst_finite()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdq_gev.R
\name{pgev}
\alias{pgev}
\alias{qgev}
\alias{dgev}
\title{Representations of the Generalized Extreme Value Distribution}
\usage{
pgev(q, location, scale, shape)

qgev(p, location, scale, shape)

dgev(x, location, scale, shape)
}
\arguments{
\item{location}{Location parameter; numeric vector.}

\item{scale}{Scale parameter; positive numeric vector.}

\item{shape}{Shape parameter; numeric vector.
This is also the extreme value index,
so that \code{shape > 0} is heavy tailed, and \code{shape < 0} is short-tailed.}

\item{p}{Vector of probabilities.}

\item{x, q}{Vector of quantiles.}
}
\value{
Vector of evaluated GEV distribution, with length
equal to the recycled lengths of \code{q}/\code{x}/\code{p}, \code{location}, \code{scale}, and
\code{shape}.
}
\description{
Representations of the Generalized Extreme Value Distribution
}
\examples{
pgev(1:10, 0, 1, 1)
dgev(1:10, 1:10, 2, 0)
qgev(1:9 / 10, 2, 10, -2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_degenerate.R
\name{dst_degenerate}
\alias{dst_degenerate}
\title{Degenerate Distribution}
\usage{
dst_degenerate(location)
}
\arguments{
\item{location}{Outcome of the distribution; single positive numeric.}
}
\value{
A degenerate distribution
}
\description{
A degenerate distribution assigns a 100\% probability to one outcome.
}
\examples{
d <- dst_degenerate(5)
realise(d)
variance(d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate_weights.R
\name{aggregate_weights}
\alias{aggregate_weights}
\title{Aggregate discrete values}
\usage{
aggregate_weights(y, weights, sum_to_one = FALSE)
}
\arguments{
\item{y}{Vector of outcomes.}

\item{weights}{Vector of weights, one for each of \code{y}.
These need not sum to one, but must not be negative and non-NA.}

\item{sum_to_one}{Logical; should the weights be normalized
to sum to 1? Default is FALSE.}
}
\value{
Data frame with the following columns:
\itemize{
\item \code{y}: Increasing vector of unique values of \code{y} that have positive weight.
\item \code{weight}: Weights corresponding to each outcome.
}
}
\description{
Aggregates discrete values together with their weights
into a data frame or tibble.
}
\details{
For a vector of outcomes \code{y} with a matching vector of \code{weights},
\code{aggregate_weights()} provides a single non-zero, non-NA
weight per unique value of \code{y}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{parameters}
\alias{parameters}
\alias{parameters<-}
\title{Parameters of a Distribution}
\usage{
parameters(distribution)

parameters(distribution) <- value
}
\arguments{
\item{distribution}{Distribution.}

\item{value}{A list of named parameter values, or \code{NULL}.}
}
\value{
A list of the distribution parameters. More specifically,
returns the \code{"parameters"} entry of the list making up the
probability distribution.
}
\description{
Get or set the parameters of a distribution, if applicable. See
details.
}
\details{
If a distribution is made by specifying parameter values
(e.g., mean and variance for a Normal distribution; shape parameters
for a Beta distribution), it is useful to keep track of what
these parameters are. This is done by adding \code{parameters}
to the list of objects defining the distribution; for instance,
\code{distribution(parameters = c(shape1 = 1.4, shape2 = 3.4))}.
Note that no checks are made to ensure the parameters are valid.
It's important to note that, in this version of distionary,
manually changing the parameters after the distribution has been
created will not change the functionality of the distribution,
because the parameters are never referred to when making calculations.
}
\examples{
a <- dst_beta(1, 2)
parameters(a)

b <- distribution(mean = 5)
parameters(b)
parameters(b) <- list(t = 7)
parameters(b)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vtype.R
\name{vtype}
\alias{vtype}
\title{Variable Type of a Distribution}
\usage{
vtype(distribution)
}
\arguments{
\item{distribution}{Distribution object.}
}
\value{
Single character with the variable type.
}
\description{
Retrieve the variable type of a distribution, such as
"continuous" or "discrete".
}
\examples{
vtype(dst_beta(1, 2))
vtype(dst_bern(0.4))
vtype(distribution())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_unif.R
\name{dst_unif}
\alias{dst_unif}
\title{Uniform Distribution}
\usage{
dst_unif(min, max)
}
\arguments{
\item{min, max}{Minimum and maximum of the distribution.
Single numerics.}
}
\value{
A Uniform distribution.
}
\description{
Makes a Uniform distribution.
}
\examples{
dst_unif(0, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_bern.R
\name{dst_bern}
\alias{dst_bern}
\title{Bernoulli Distribution}
\usage{
dst_bern(prob)
}
\arguments{
\item{prob}{Probability of success; single numeric between 0 and 1.}
}
\value{
A Bernoulli distribution.
}
\description{
Makes a Bernoulli distribution, representing the outcome of a
single trial with a given success probability.
}
\examples{
dst_bern(0.3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_finite.R
\name{dst_finite}
\alias{dst_finite}
\title{Finite Distribution}
\usage{
dst_finite(outcomes, probs)
}
\arguments{
\item{outcomes}{Numeric vector representing the potential
outcomes of the distribution.}

\item{probs}{Numeric vector of probabilities corresponding to the outcomes
in \code{outcomes}. Must not be negative and must sum to 1.}
}
\value{
A distribution with finite outcomes.
}
\description{
Makes a finite distribution, which is a distribution with a finite number
of possible outcomes.
}
\examples{
dst_finite(2:5, probs = 1:4 / 10)
}
\seealso{
\code{\link[=dst_empirical]{dst_empirical()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_exp.R
\name{dst_exp}
\alias{dst_exp}
\title{Exponential Distribution}
\usage{
dst_exp(rate)
}
\arguments{
\item{rate}{Rate parameter; single positive numeric.}
}
\value{
An Exponential distribution.
}
\description{
Makes an Exponential distribution.
}
\examples{
dst_exp(1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_t.R
\name{dst_t}
\alias{dst_t}
\title{Student t Distribution}
\usage{
dst_t(df)
}
\arguments{
\item{df}{Degrees of freedom; single positive numeric.}
}
\value{
A Student t distribution.
}
\description{
Makes a Student t distribution.
}
\examples{
dst_t(3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_gamma.R
\name{dst_gamma}
\alias{dst_gamma}
\title{Gamma Distribution}
\usage{
dst_gamma(shape, rate)
}
\arguments{
\item{shape}{Shape parameter; single positive numeric.}

\item{rate}{Rate parameter; single positive numeric.}
}
\value{
A Gamma distribution.
}
\description{
Makes a Gamma distribution.
}
\examples{
dst_gamma(2, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/median.R
\name{median.dst}
\alias{median.dst}
\title{Median of a Distribution}
\usage{
\method{median}{dst}(x, ...)
}
\arguments{
\item{x}{Distribution to calculate median from.}

\item{...}{Not used.}
}
\value{
Median of a distribution; single numeric.
}
\description{
Finds the median of a distribution.
}
\details{
Median is calculated as the 0.5-quantile when not found in the
distribution. So, when the median is
non-unique, the smallest of the possibilities is taken.
}
\examples{
d <- dst_gamma(3, 3)
median(d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kurtosis.R, R/mean.R, R/skewness.R, R/stdev.R,
%   R/variance.R
\name{kurtosis}
\alias{kurtosis}
\alias{kurtosis_exc}
\alias{mean.dst}
\alias{skewness}
\alias{stdev}
\alias{variance}
\title{Moments of a Distribution}
\usage{
kurtosis(distribution)

kurtosis_exc(distribution)

\method{mean}{dst}(x, ...)

skewness(distribution)

stdev(distribution)

variance(distribution)
}
\arguments{
\item{x, distribution}{Distribution to evaluate.}

\item{...}{When calculating the mean via integration of the quantile
function, arguments passed to \code{stats::integrate()}.}
}
\value{
A single numeric.
}
\description{
Get common moment-related quantities of a
distribution: \code{mean}, \code{variance}, standard deviation (\code{stdev}),
\code{skewness}, and \code{kurtosis} or excess kurtosis (\code{kurtosis_exc}).
If these quantities are not supplied in the
distribution's definition, a numerical algorithm may be used.
}
\details{
If there is no method associated with a subclass of
\code{x}, then moments are calculated using
\code{stats::integrate()} from the density function.
}
\note{
Beware that if a quantity is being calculated numerically
for a non-continuous (e.g., discrete) distribution, the calculation
could be highly approximate. An upcoming version of distionary will
resolve this issue.
}
\examples{
a <- dst_gp(1, 0.5)
b <- dst_unif(0, 1)
c <- dst_norm(3, 4)
mean(a)
variance(b)
kurtosis(c)
kurtosis_exc(c)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eval_property.R
\name{eval_property}
\alias{eval_property}
\title{Evaluate a distribution}
\usage{
eval_property(distribution, entry, ...)
}
\arguments{
\item{distribution}{Distribution object.}

\item{entry}{Name of the property, such as "cdf" or
"mean". Length 1 character vector.}

\item{...}{If the property is a function, arguments to the
function go here. Need not be named; inserted in the order they
appear.}
}
\value{
The distribution's property, evaluated. If cannot be
evaluated, returns \code{NULL}.
}
\description{
Evaluate a distribution property. The distribution itself
is first searched for the property, and if it can't
be found, will attempt to calculate the property
from other entries.
}
\examples{
d <- distribution(
  cdf = function(x) {
    (x > 0) * pmin(x^2, 1)
  },
  g = 9.81,
  .vtype = "continuous"
)
eval_property(d, "g")
eval_property(d, "quantile", 1:9 / 10)
eval_property(d, "mean")
eval_property(d, "realise", 10)
eval_property(d, "foofy")
eval_property(d, "foofy", 1:10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_lnorm.R
\name{dst_lnorm}
\alias{dst_lnorm}
\title{Log Normal Distribution}
\usage{
dst_lnorm(meanlog, sdlog)
}
\arguments{
\item{meanlog}{Mean of the log of the random variable;
single numeric.}

\item{sdlog}{Standard deviation of the log of the random variable;
single positive numeric.}
}
\value{
A Log Normal distribution.
}
\description{
Makes a Log Normal distribution, which is the distribution of
the exponential of a Normally distributed random variable.
}
\examples{
dst_lnorm(0, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prob_left_right.R
\name{prob_left}
\alias{prob_left}
\alias{prob_right}
\title{Find the probability left or right of a number}
\usage{
prob_left(distribution, of, inclusive)

prob_right(distribution, of, inclusive)
}
\arguments{
\item{distribution}{Distribution to find probabilities of.}

\item{of}{Find the probability to the left or right \emph{of} this number.
Could be a vector.}

\item{inclusive}{Should \code{of} be included in the probability calculation?
Logical.}
}
\value{
A vector of probabilities.
}
\description{
Probability to the left or right of a number, inclusive or not.
\code{prob_left()} is a more general cdf defined using either \code{<} or \code{<=}, and
\code{prob_right()} is a more general survival function defined using either
\code{>} or \code{>=}.
}
\examples{
d <- dst_pois(5)
prob_left(d, of = 3, inclusive = TRUE)
prob_left(d, of = 3, inclusive = FALSE)
prob_right(d, of = 0:3, inclusive = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_norm.R
\name{dst_norm}
\alias{dst_norm}
\title{Normal (Gaussian) Distribution}
\usage{
dst_norm(mean, sd)
}
\arguments{
\item{mean}{Mean of the distribution. Single numeric.}

\item{sd}{Standard deviation of the distribution.
Single positive numeric.}
}
\value{
A Normal distribution.
}
\description{
Makes a Normal (Gaussian) distribution.
}
\examples{
dst_norm(0, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_cauchy.R
\name{dst_cauchy}
\alias{dst_cauchy}
\title{Cauchy Distribution}
\usage{
dst_cauchy(location, scale)
}
\arguments{
\item{location}{Location parameter; single numeric.}

\item{scale}{Scale parameter; single positive numeric.}
}
\value{
A Cauchy distribution.
}
\description{
Makes a Cauchy distribution.
}
\examples{
d <- dst_cauchy(0, 1)

# Moments do not exist for the Cauchy distribution.
mean(d)
variance(d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_pearson3.R
\name{dst_pearson3}
\alias{dst_pearson3}
\title{Pearson Type III distribution}
\usage{
dst_pearson3(location, scale, shape)
}
\arguments{
\item{location}{Location parameter, specifying how to shift the
Gamma distribution; single numeric.}

\item{scale}{Scale parameter of the Gamma distribution;
single positive numeric.}

\item{shape}{Shape parameter of the Gamma distribution;
single positive numeric.}
}
\value{
A Pearson Type III distribution.
}
\description{
Makes a Pearson Type III distribution, which is a Gamma distribution,
but shifted.
}
\examples{
dst_pearson3(1, 1, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_beta.R
\name{dst_beta}
\alias{dst_beta}
\title{Beta Distribution}
\usage{
dst_beta(shape1, shape2)
}
\arguments{
\item{shape1, shape2}{Shape parameters of the distribution;
single positive numerics.}
}
\value{
A Beta distribution.
}
\description{
Makes a Beta distribution.
}
\examples{
dst_beta(2, 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distionary-package.R
\docType{package}
\name{distionary-package}
\alias{distionary}
\alias{distionary-package}
\title{distionary: Create and Evaluate Probability Distributions}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Create and evaluate probability distribution objects from a variety of families or define custom distributions. Automatically compute distributional properties, even when they have not been specified. This package supports statistical modeling and simulations, and forms the core of the probaverse suite of R packages.
}
\section{Overview}{

The \pkg{distionary} package provides a comprehensive framework for working
with probability distributions in R. With \pkg{distionary}, you can:
\enumerate{
\item Specify probability distributions from common families or create
custom distributions.
\item Evaluate distributional properties and representations.
\item Access distributional calculations even when they're not directly
specified.
}

The main purpose of \pkg{distionary} is to implement a distribution object
that powers the wider \href{https://probaverse.com}{probaverse ecosystem} for
making probability distributions that are representative of your data.
}

\section{Creating Distributions}{

Use the \code{dst_*()} family of functions to create distributions from
common families:
\itemize{
\item \code{\link{dst_norm}()}, \code{\link{dst_exp}()},
\code{\link{dst_unif}()}, etc. are some continuous distributions.
\item \code{\link{dst_pois}()}, \code{\link{dst_binom}()},
\code{\link{dst_geom}()}, etc. are some discrete distributions.
\item \code{\link{dst_empirical}()} is useful for creating a non-parametric
distribution from data.
}
You can also make your own distribution using the
\code{\link{distribution}()} function, which allows you to specify
any combination of distributional representations and properties. For this
version of \pkg{distionary}, the CDF and density/PMF are required
in order to access all functionality.
}

\section{Evaluating Distributions}{

A distribution's \emph{representations} are functions that fully describe the
distribution. They can be accessed with the \code{eval_*()} functions.
For example, \code{\link{eval_cdf}()} and \code{\link{eval_quantile}()}
invoke the distribution's cumulative distribution function (CDF)
and quantile function.

Other properties of the distribution can be calculated by functions of the
property's name, such as \code{\link{mean}()} and \code{\link{range}()}.
}

\section{Random Samples}{

Generate random samples from a distribution using
\code{\link{realise}()}.
}

\section{Getting Started}{

New users should start with the package vignettes:
\itemize{
\item \code{vignette("specify", package = "distionary")} -
Learn how to specify distributions.
\item \code{vignette("evaluate", package = "distionary")} -
Learn how to evaluate distributions.
}
}

\examples{
# Create a Poisson distribution.
poisson <- dst_pois(lambda = 1.5)
poisson

# Evaluate the probability mass function.
eval_pmf(poisson, at = 0:4)
plot(poisson)

# Get distribution properties.
mean(poisson)
variance(poisson)

# Create a continuous distribution (Normal).
normal <- dst_norm(mean = 0, sd = 1)

# Evaluate quantiles.
eval_quantile(normal, at = c(0.025, 0.5, 0.975))

# Create a custom distribution.
my_dist <- distribution(
  density = function(x) ifelse(x >= 0 & x <= 1, 2 * (1 - x), 0),
  cdf = function(x) ifelse(x >= 0 & x <= 1, 1 - (1 - x)^2, 0),
  .vtype = "continuous",
  .name = "Linear"
)
plot(my_dist)
plot(my_dist, "cdf")

# Even without specifying all properties, they can still be computed.
mean(my_dist)

}
\seealso{
Useful links:
\itemize{
  \item \url{https://distionary.probaverse.com/}
  \item \url{https://github.com/probaverse/distionary}
  \item Report bugs at \url{https://github.com/probaverse/distionary/issues}
}

}
\author{
\strong{Maintainer}: Vincenzo Coia \email{vincenzo.coia@gmail.com} [copyright holder]

Other contributors:
\itemize{
  \item Amogh Joshi [contributor]
  \item Shuyi Tan [contributor]
  \item Zhipeng Zhu [contributor]
  \item olivroy (GitHub contributor) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/realise.R
\name{realise}
\alias{realise}
\alias{realize}
\title{Generate a Sample from a Distribution}
\usage{
realise(distribution, n = 1)

realize(distribution, n = 1)
}
\arguments{
\item{distribution}{Distribution object.}

\item{n}{Number of observations to generate.}
}
\value{
Vector of independent values drawn from the inputted distribution.
}
\description{
Draw \code{n} independent observations from a distribution.
}
\note{
\code{realise()} and \code{realize()} are aliases and do the same thing.
}
\examples{
d <- dst_pois(5)
set.seed(2)
realise(d, n = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_nbinom.R
\name{dst_nbinom}
\alias{dst_nbinom}
\title{Negative binomial Distribution}
\usage{
dst_nbinom(size, prob)
}
\arguments{
\item{size}{Number of successful trials; single positive numeric.}

\item{prob}{Probability of a successful trial; single numeric
between 0 and 1.}
}
\value{
A Negative Binomial distribution.
}
\description{
Makes a Negative Binomial distribution, corresponding to the number
of failures in a sequence of independent trials until a given number
of successes are observed.
}
\examples{
d <- dst_nbinom(10, 0.5)

# This version of the Negative Binomial distribution does not count
# the successes.
range(d)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_hyper.R
\name{dst_hyper}
\alias{dst_hyper}
\title{Hypergeometric Distribution}
\usage{
dst_hyper(m, n, k)
}
\arguments{
\item{m}{The number of red balls in the urn; single positive integer.}

\item{n}{The number of non-red balls in the urn; single positive integer.}

\item{k}{the number of balls drawn from the urn (between 0 and \code{m + n});
single positive integer.}
}
\value{
A Hypergeometric distribution.
}
\description{
Creates a Hypergeometric distribution. The parameterization used here
is the same as for \code{stats::phyper()}, where the outcome can be thought
of as the number of red balls drawn from an urn of coloured balls,
using a scoop that holds a fixed number of balls.
}
\examples{
dst_hyper(15, 50, 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dst_chisq.R
\name{dst_chisq}
\alias{dst_chisq}
\title{Chi-Squared Distribution}
\usage{
dst_chisq(df)
}
\arguments{
\item{df}{degrees of freedom parameter; single positive numeric.}
}
\value{
A Chi-Squared distribution
}
\description{
Makes a Chi-Squared distribution.
}
\examples{
dst_chisq(3)
}
